////////////////////////////////////////////////////////////////////////////////
// Fractions 1.2
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit Fractions;

{$I HIT_SYMBOLS.INC}

//==============================================================================
{$IFDEF DEBUG}
  {$ASSERTIONS ON}     //$C+
  {$DEBUGINFO ON}      //$D+
  {$DEFINITIONINFO ON} //$YD
  {$LOCALSYMBOLS ON}   //$L+
  {$OPTIMIZATION OFF}  //$O-
  {$STACKFRAMES ON}    //$W+
{$ELSE}
  {$ASSERTIONS OFF}     //$C-
  {$DEBUGINFO OFF}      //$D-
  {$DEFINITIONINFO OFF} //$Y-
  {$LOCALSYMBOLS OFF}   //$L-
  {$OPTIMIZATION ON}    //$O+
  {$STACKFRAMES OFF}    //$W-
{$ENDIF}
//------------------------------------------------------------------------------
{$ALIGN ON}           //$A+
{$BOOLEVAL OFF}       //$B-  !Must never be modified
{$DENYPACKAGEUNIT ON}
{$IMPORTEDDATA ON}    //$G+
{$IOCHECKS OFF}       //$I-
{$LONGSTRINGS ON}     //$H+
{$OVERFLOWCHECKS OFF} //$Q-
{$RANGECHECKS OFF}    //$R-
{$WRITEABLECONST OFF} //$J-
//==============================================================================

interface
uses Integers, Classes;

const
  UnitVersionStr = '1.2';
  UnitVersionInt = 102;

type
  //----- Continued Fraction Expansion coefficient
  TCFEFunc = function(Index: Word): Longword;

  //----- Abstract ancestor
  TAbstractFraction = class(THITObject)
  end;

  //----- Fraction
  TFraction = class(TAbstractFraction)
  private
    // Fields
    fN: TInteger; //Numerator
    fD: TInteger; //Denominator, always > 0
    fNormalized: Boolean;
    // Methods
    procedure BuildFromCFELoc(const CFE: array of Longword; Rep: Word);
    function  GetDen: TInteger;
    function  GetNeg: Boolean;
    function  GetNum: TInteger;
    function  GetSign: Integer;
    function  IsIntegerQ: Boolean;
    procedure Normalize;
    procedure SetNeg(OnOff: Boolean);
  public
    // Constructors and destructor
    constructor Create; overload;
    constructor Create(A: TFraction); overload;
    constructor Create(N: Int64); overload;
    constructor Create(N, D: Int64); overload;
    constructor Create(N: TInteger); overload;
    constructor Create(N, D: TInteger); overload;
    destructor  Destroy; override;
    // Methods
    procedure BuildFromCFE(CFE: TCFEFunc; Last: Word); overload;
    procedure BuildFromCFE(const CFE: array of Longword; Repetition: Word); overload;
    procedure BuildFromCFE(const CFE: TDynLongArray; Repetition: Word); overload;
    function  Clone: TFraction;
    procedure Copy(A: TFraction);
    function  Is0: Boolean;
    function  Is1: Boolean;
    function  IsMinus1: Boolean;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure Negate;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure Set0;
    procedure Set1;
    procedure SetMinus1;
    procedure SplitAbs(Q, R: TInteger);
    // Properties
    property Den: TInteger read GetDen;
    property IsInteger: Boolean read IsIntegerQ;
    property Neg: Boolean read GetNeg write SetNeg;
    property Num: TInteger read GetNum;
    property Sign: Integer read GetSign;
  end;

  //----- Fraction vector
  TFractionVector = class(THITObject)
  private
    // Fields
    fCell: Pointer; //Pointer to a TFraction array
    fSize: Longint;
    // Methods
    function  GetCellPtr(I: Longint): TFraction;
    procedure SetSize(Value: Longint);
  public
    // Class method
    class function MaxSize: Longint;
    // Destructor
    destructor Destroy; override;
    // Methods
    procedure AddCell(A: TFraction = nil);
    function  Clone: TFractionVector;
    procedure Copy(V: TFractionVector);
    procedure Fill(A: TFraction);
    function  IsNull: Boolean;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure Negate;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure SetNull;
    procedure Sort(AbsValues: Boolean = false);
    procedure SwpCells(I, J: Longint);
    // Properties
    property Cell[I: Longint]: TFraction read GetCellPtr; default;
    property Size: Longint read fSize write SetSize;
  end;

  //----- Complex fraction
  TXFraction = class(TAbstractFraction)
  private
    // Fields
    fR: TFraction; //Real part
    fI: TFraction; //Imaginary part
  public
    // Constructors and destructor
    constructor Create; overload;
    constructor Create(R, I: Int64); overload;
    constructor Create(R, I: TFraction); overload;
    constructor Create(F: TXFraction); overload;
    destructor  Destroy; override;
    // Methods
    function  Clone: TXFraction;
    procedure Conjugate; //(R,I) := (R,-I)
    procedure Copy(F: TXFraction);
    function  Is0: Boolean; //Self = (0,0) ?
    function  Is1: Boolean; //Self = (1,0) ?
    function  IsI: Boolean; //Self = (0,1) ?
    function  IsUnit: Boolean; //R+I = 1 ? (i.e., Self is a unit ?)
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure MulI;
    procedure MulMinusI;
    procedure Negate;
    procedure Norm(F: TFraction);
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure Set0; //Self := (0,0)
    procedure Set1; //Self := (1,0)
    procedure SetIm(I: Int64); overload;     //Self := (unmodified,I)
    procedure SetIm(I: TFraction); overload; //Self := (unmodified,I)
    procedure SetRe(R: Int64); overload;     //Self := (R,unmodified)
    procedure SetRe(R: TFraction); overload; //Self := (R,unmodified)
    // Properties
    property Im: TFraction read fI;
    property Re: TFraction read fR;
  end;

  //----- Complex fraction vector
  TXFractionVector = class(THITObject)
  private
    // Fields
    fCell: Pointer; //Pointer to a TXFraction array
    fSize: Longint;
    // Methods
    function  GetCellPtr(I: Longint): TXFraction;
    procedure SetSize(Value: Longint);
  public
    // Class method
    class function MaxSize: Longint;
    // Destructor
    destructor Destroy; override;
    // Methods
    procedure AddCell(A: TXFraction = nil);
    function  Clone: TXFractionVector;
    procedure Copy(V: TXFractionVector);
    procedure Fill(A: TXFraction);
    function  IsNull: Boolean;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure Negate;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure SetNull;
    procedure SwpCells(I, J: Longint);
    // Properties
    property Cell[I: Longint]: TXFraction read GetCellPtr; default;
    property Size: Longint read fSize write SetSize;
  end;

procedure FAbs(A: TFraction);
procedure FAdd(A: TFraction; B: Int64); overload;
procedure FAdd(A: TFraction; B: TInteger); overload;
procedure FAdd(A, B: TFraction); overload;
procedure FAdd(A, B: TXFraction); overload;
function  FCmp(A: TFraction; N, D: Int64): Integer; overload;
function  FCmp(A, B: TFraction): Integer; overload;
function  FCmpAbs(A: TFraction; N, D: Int64): Integer; overload;
function  FCmpAbs(A, B: TFraction): Integer; overload;
procedure FDec(A: TFraction);
function  FDecimalStr(A: TFraction; Blocks: Byte; Places: Longint): AnsiString;
procedure FDiv(A: TFraction; B: Int64); overload;
procedure FDiv(A: TFraction; B: TInteger); overload;
procedure FDiv(A, B: TFraction); overload;
procedure FDiv(A: TXFraction; B: TFraction); overload;
procedure FDiv(A, B: TXFraction); overload;
procedure FDiv2(A: TFraction); overload;
procedure FDiv2(A: TFraction; k: Longint); overload;
procedure FDiv2(A: TXFraction); overload;
procedure FDiv2(A: TXFraction; k: Longint); overload;
function  FEqu(A, B: TFraction): Boolean; overload;
function  FEqu(A, B: TXFraction): Boolean; overload;
function  FEquAbs(A, B: TFraction): Boolean;
procedure FInc(A: TFraction);
procedure FInv(A: TFraction); overload;
procedure FInv(A: TXFraction); overload;
procedure FMul(A: TFraction; B: Int64); overload;
procedure FMul(A: TFraction; B: TInteger); overload;
procedure FMul(A, B: TFraction); overload;
procedure FMul(A: TXFraction; B: TFraction); overload;
procedure FMul(A, B: TXFraction); overload;
procedure FMul2(A: TFraction); overload;
procedure FMul2(A: TFraction; k: Longint); overload;
procedure FMul2(A: TXFraction); overload;
procedure FMul2(A: TXFraction; k: Longint); overload;
procedure FPow(A: TFraction; E: Longint); overload;
procedure FPow(A: TXFraction; E: TInteger); overload;
procedure FRnd(A: TFraction; BitCount: Longint; Signed: Boolean = false);
procedure FSet(A: TFraction; N: Int64); overload;
procedure FSet(A: TFraction; N, D: Int64); overload;
procedure FSet(A: TFraction; N: TInteger); overload;
procedure FSet(A: TFraction; N, D: TInteger); overload;
procedure FSet(A, B: TFraction); overload;
procedure FSet(A, B: TXFraction); overload;
procedure FSet(A: TXFraction; R, I: Int64); overload;
procedure FSet(A: TXFraction; R, I: TFraction); overload;
procedure FSqr(A: TFraction); overload;
procedure FSqr(A: TXFraction); overload;
procedure FSqrRoot(A: TFraction); overload;
procedure FSqrRoot(A: TXFraction); overload;
procedure FSub(A: TFraction; B: Int64); overload;
procedure FSub(A: TFraction; B: TInteger); overload;
procedure FSub(A, B: TFraction); overload;
procedure FSub(A, B: TXFraction); overload;
procedure FSwp(A, B: TFraction); overload;
procedure FSwp(A, B: TXFraction); overload;

//----- Some CFE functions (Continued Fraction Expansion)
function CFEEuler(Index: Word): Longword;
function CFEGoldenRatio(Index: Word): Longword;
function CFETan1(Index: Word): Longword;

implementation
//==============================================================================
uses SysUtils;

////////////////////////////////////////////////////////////////////////////////
// Resource string
////////////////////////////////////////////////////////////////////////////////

resourcestring
  sFRA0001 = '(Value < 0) or (Value > TInteger.MaxBitSize)';

////////////////////////////////////////////////////////////////////////////////
// Consts, Types
////////////////////////////////////////////////////////////////////////////////

type
  TFracHeaderFile = array[0..12] of AnsiChar;
  TXFracHeaderFile = array[0..13] of AnsiChar;
  TFracVectorHeaderFile = array[0..19] of AnsiChar;
  TXFracVectorHeaderFile = array[0..20] of AnsiChar;
  TFracCmpFunction = function(A, B: TFraction): Integer;
  TVCell = array[0..MaxListSize-1] of TFraction;
  TXVCell = array[0..MaxListSize-1] of TXFraction;

const
  FracHeaderFile : TFracHeaderFile = 'HIT Fraction ';
  XFracHeaderFile : TXFracHeaderFile = 'HIT XFraction ';
  FracVectorHeaderFile : TFracVectorHeaderFile = 'HIT Fraction Vector ';
  XFracVectorHeaderFile : TXFracVectorHeaderFile = 'HIT XFraction Vector ';

////////////////////////////////////////////////////////////////////////////////
// CFE Routines
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Compute a coefficient of the CFE of e (Euler constant)
// -> Index, any
// <- Result = Index-th coefficient of the CFE
//==============================================================================
function CFEEuler(Index: Word): Longword;
begin
  //----- Integer part
  if Index = 0 then Result := 2
  else
  //----- Fractional part
  if (Index mod 3) <> 2 then Result := 1
  else
  Result := (Succ(Index) div 3) shl 1;
end;

//==============================================================================
// Compute a coefficient of the CFE of the Golden Ratio (1,618...)
// -> Index, any
// <- Result = Index-th coefficient of the CFE
//------------------------------------------------------------------------------
// The Golden Ratio is the positive root of X - X - 1 = 0
//==============================================================================
function CFEGoldenRatio(Index: Word): Longword;
begin
  Result := 1;
end;

//==============================================================================
// Compute a coefficient of the CFE of Tan(1)
// -> Index, any
// <- Result = Index-th coefficient of the CFE
//==============================================================================
function CFETan1(Index: Word): Longword;
begin
  if Odd(Index) then Result := Index else Result := 1;
end;

////////////////////////////////////////////////////////////////////////////////
// Routines
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Global
// A := |A|
//==============================================================================
procedure FAbs(A: TFraction);
begin
  with A do
  begin
    if not fNormalized then Normalize;
    IAbs(fN);
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A + B
//==============================================================================
procedure FAdd(A: TFraction; B: Int64);
  var T : TInteger;
begin
  with A do
  begin
    if not fNormalized then Normalize;
    T := fD.Clone;
    try
      IMul(T,B);
      IAdd(fN,T);
      Normalize;
    finally
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A + B
//==============================================================================
procedure FAdd(A: TFraction; B: TInteger);
  var T : TInteger;
begin
  with A do
  begin
    if not fNormalized then Normalize;
    T := fD.Clone;
    try
      IMul(T,B);
      IAdd(fN,T);
      Normalize;
    finally
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A + B
//==============================================================================
procedure FAdd(A, B: TFraction);
  var G, T : TInteger;
begin
  if A = B then
  begin
    FMul2(A);
    Exit;
  end;

  if not A.fNormalized then A.Normalize;
  if not B.fNormalized then B.Normalize;

  if not B.Is0 then
    if A.Is0 then FSet(A,B)
    else
    begin
      G := nil;
      T := nil;
      try
        G := A.fD.Clone;
        if IGCD(G,B.fD) then
        begin
          //----- GCD = 1 (with random inputs, proba = 6/PI ~ 0.6)
          T := A.fD.Clone;
          IMul(T,B.fN);
          IMul(A.fN,B.fD);
          IAdd(A.fN,T);
          IMul(A.fD,B.fD);
        end
        else
        begin
          IDiv(A.fD,G);
          T := B.fD.Clone;
          IDiv(T,G);
          IMul(A.fN,T);
          ISet(T,B.fN);
          IMul(T,A.fD);
          IAdd(A.fN,T);
          ISet(T,B.fD);
          if not IGCD(G,A.fN) then
          begin
            IDiv(A.fN,G);
            IDiv(T,G);
          end;
          IMul(A.fD,T);
        end;
        //----- Normalize
        if A.fD.Neg then
        begin
          A.fN.Negate;
          IAbs(A.fD);
        end;
      finally
        G.Free;
        T.Free;
      end;
    end;
end;

//==============================================================================
// Global/Overloaded
// A := A + B
//==============================================================================
procedure FAdd(A, B: TXFraction);
begin
  FAdd(A.fR,B.fR);
  FAdd(A.fI,B.fI);
end;

//==============================================================================
// Global
// A := A - 1
//==============================================================================
procedure FDec(A: TFraction);
begin
  with A do
  begin
    if not fNormalized then Normalize;
    ISub(fN,fD);
    if fN.Is0 then fD.Set1; //Normalize
  end;
end;

//==============================================================================
// Global/Overloaded
// Result := TRUE if A = N/D, FALSE otherwise
//==============================================================================
function FCmp(A: TFraction; N, D: Int64): Integer;
  var S, T : TInteger;
begin
  //----- Check
  if D <= 0 then HITRaiseInvalidArg('FCmp','D <= 0');

  if not A.fNormalized then A.Normalize;
  S := nil;
  T := nil;
  try
    S := A.fN.Clone;
    IMul(S,D);
    T := A.fD.Clone;
    IMul(T,N);
    Result := ICmp(S,T);
  finally
    S.Free;
    T.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// Result := TRUE if A = B, FALSE otherwise
//==============================================================================
function FCmp(A, B: TFraction): Integer;
  var S, T : TInteger;
begin
  if A = B then Result := 0
  else
  begin
    if not A.fNormalized then A.Normalize;
    if not B.fNormalized then B.Normalize;
    S := nil;
    T := nil;
    try
      S := A.fN.Clone;
      IMul(S,B.fD);
      T := A.fD.Clone;
      IMul(T,B.fN);
      Result := ICmp(S,T);
    finally
      S.Free;
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// Result := TRUE if |A| = |N/D|, FALSE otherwise
//==============================================================================
function FCmpAbs(A: TFraction; N, D: Int64): Integer;
  var S, T : TInteger;
begin
  //----- Check
  if D <= 0 then HITRaiseInvalidArg('FCmpAbs','D <= 0');

  if not A.fNormalized then A.Normalize;
  S := nil;
  T := nil;
  try
    S := A.fN.Clone;
    IMul(S,D);
    T := A.fD.Clone;
    IMul(T,N);
    Result := ICmpAbs(S,T);
  finally
    S.Free;
    T.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// Result := TRUE if |A| = |B|, FALSE otherwise
//==============================================================================
function FCmpAbs(A, B: TFraction): Integer;
  var S, T : TInteger;
begin
  if A = B then Result := 0
  else
  begin
    if not A.fNormalized then A.Normalize;
    if not B.fNormalized then B.Normalize;
    S := nil;
    T := nil;
    try
      S := A.fN.Clone;
      IMul(S,B.fD);
      T := A.fD.Clone;
      IMul(T,B.fN);
      Result := ICmpAbs(S,T);
    finally
      S.Free;
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global
// Return a "decimal" string
//==============================================================================
function FDecimalStr(A: TFraction; Blocks: Byte; Places: Longint): AnsiString;
  const
    PowOf10 : array[1..8] of Longword = (
     10,100,1000,10000,100000,1000000,10000000,100000000);
  var
    Q, R, E, D : TInteger;
    i, j       : Longint;
    S          : AnsiString;
begin
  if A.Is0 then
  begin
    with HITStrFormat do
    if Signed then
      if SpaceAsPlus then Result := ' 0'
      else Result := '+0'
    else Result := '0';
    Result := Result + DecimalSeparator + '0';
    Exit;
  end;
  Q := nil;
  R := nil;
  E := nil;
  D := nil;
  try
    Q := TInteger.Create;
    R := TInteger.Create;
    A.SplitAbs(Q,R);
    Result := IStr(Q,10,Blocks,false,true) + DecimalSeparator;
    if A.Neg then Result[1] := '-';
    if A.fD.Is1 or R.Is0 then
    begin
      Result := Result + '0';
      Exit;
    end;
    E := TInteger.Create;
    D := A.fD.Clone;
    i := R.BitSize + A.fD.BitSize;
    if (Places = 0) or (Places > i) then Places := i;
    j := 0;
    for i := 1 to (Places + 8) div 9 do
    begin
      IMul(R,1000000000);
      ISet(Q,D);
      IRemDiv(R,Q);
      IMul(E,1000000000); //On 1st iteration, E = 0. Don't mind.
      IAdd(E,Q);
      Inc(j,9);
    end;
    //----- Suppress decimals (if need be)
    //      Here, Places <= j < (Places + 9)
    if j > Places then
    begin
      IDiv(E,PowOf10[j-Places]);
      j := Places;
    end;
    S := IStr(E,10,Blocks,false,false);
    Dec(j,Length(S));
    if j > 0 then Result := Result + StringOfChar('0',j) + S
             else Result := Result + S;
  finally
    Q.Free;
    R.Free;
    E.Free;
    D.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A / B
//==============================================================================
procedure FDiv(A: TFraction; B: Int64);
begin
  //----- Check
  if B = 0 then HITRaiseDivByZero('FDiv');

  with A do
  begin
    if not fNormalized then Normalize;
    if not fN.Is0 then
    begin
      IMul(fD,B);
      Normalize;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A / B
//==============================================================================
procedure FDiv(A: TFraction; B: TInteger);
  var G, H : TInteger;
begin
  //----- Check
  if B.Is0 then HITRaiseDivByZero('FDiv');

  if not A.fNormalized then A.Normalize;
  if not A.fN.Is0 then
  begin
    G := A.fN.Clone;
    try
      IGCD(G,B);
      if G.Is1 then IMul(A.fD,B)
      else
      begin
        IDiv(A.fN,G);
        H := B.Clone;
        try
          IDiv(H,G);
          IMul(A.fD,H);
        finally
          H.Free;
        end;
      end;
      //----- Normalize
      if A.fD.Neg then
      begin
        A.fN.Negate;
        IAbs(A.fD);
      end;
    finally
      G.Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A / B
//==============================================================================
procedure FDiv(A, B: TFraction);
  var
    G, H, T  : TInteger;
    Selector : Integer;
begin
  if A = B then
  begin
    A.Set1;
    Exit;
  end;

  if not A.fNormalized then A.Normalize;
  if not B.fNormalized then B.Normalize;

  if not A.Is0 then
    if B.Is0 then HITRaiseDivByZero('FDiv')
    else
    begin
      G := nil;
      H := nil;
      T := nil;
      try
        Selector := 0;
        G := A.fN.Clone;
        if IGCD(G,B.fN) then Inc(Selector);
        H := A.fD.Clone;
        if IGCD(H,B.fD) then Inc(Selector,2);
        case Selector of
          //----- G > 1 and H > 1
          0: begin
               //----- A.N = (A.N / G) * (B.D / H)
               IDiv(A.fN,G);
               T := B.fD.Clone;
               IDiv(T,H);
               IMul(A.fN,T);
               //----- A.D = (A.D / H) * (B.N / G)
               IDiv(A.fD,H);
               ISet(T,B.fN);
               IDiv(T,G);
               IMul(A.fD,T);
             end;
          //----- G = 1 and H > 1
          1: begin
               //----- A.N = A.N * (B.D / H)
               T := B.fD.Clone;
               IDiv(T,H);
               IMul(A.fN,T);
               //----- A.D = (A.D / H) * B.N
               IDiv(A.fD,H);
               IMul(A.fD,B.fN);
             end;
          //----- G > 1 and H = 1
          2: begin
               //----- A.N = (A.N / G) * B.D
               IDiv(A.fN,G);
               IMul(A.fN,B.fD);
               //----- A.D =  A.D * (B.N / G)
               T := B.fN.Clone;
               IDiv(T,G);
               IMul(A.fD,T);
             end;
          //3:
          else
             //----- G = 1 and H = 1
             begin
               IMul(A.fN,B.fD);
               IMul(A.fD,B.fN);
             end;
        end;
        //----- Normalize
        if A.fD.Neg then
        begin
          A.fN.Negate;
          IAbs(A.fD);
        end;
      finally
        G.Free;
        H.Free;
        T.Free;
      end;
    end;
end;

//==============================================================================
// Global/Overloaded
// A := A / B
//==============================================================================
procedure FDiv(A: TXFraction; B: TFraction);
begin
  FDiv(A.fR,B);
  FDiv(A.fI,B);
end;

//==============================================================================
// Global/Overloaded
// A := A div B
//==============================================================================
procedure FDiv(A, B: TXFraction);
  var N : TFraction;
begin
  //----- Check
  if B.Is0 then HITRaiseDivByZero('FDiv');

  if A = B then
  begin
    A.Set1;
    Exit;
  end;

  N := TFraction.Create;
  try
    //----- A := A * ~B
    B.Conjugate;
    FMul(A,B);
    B.Conjugate;
    //----- A := (Ar/N, Ai/N)
    B.Norm(N);
    FDiv(A.fR,N);
    FDiv(A.fI,N);
  finally
    N.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A / 2
//==============================================================================
procedure FDiv2(A: TFraction);
begin
  with A do
  begin
    if not fNormalized then Normalize;
    if fN.Odd then IShl(fD) else IShr(fN);
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A / 2**k
//==============================================================================
procedure FDiv2(A: TFraction; k: Longint);
begin
  //----- Check
  if k < 0 then HITRaiseInvalidArg('FDiv2','k < 0');

  with A do
  begin
    if not fNormalized then Normalize;
    IShl(fD,k);
    if not fN.Odd then Normalize;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A / 2
//==============================================================================
procedure FDiv2(A: TXFraction);
begin
  FDiv2(A.Re);
  FDiv2(A.Im);
end;

//==============================================================================
// Global/Overloaded
// A := A / 2**k
//==============================================================================
procedure FDiv2(A: TXFraction; k: Longint);
begin
  FDiv2(A.Re,k);
  FDiv2(A.Im,k);
end;

//==============================================================================
// Global/Overloaded
// Result := A = B
//==============================================================================
function FEqu(A, B: TFraction): Boolean;
begin
  if A = B then Result := true
  else
  begin
    if not A.fNormalized then A.Normalize;
    if not B.fNormalized then B.Normalize;
    Result := IEqu(A.fN,B.fN) and IEquAbs(A.fD,B.fD);
  end;
end;

//==============================================================================
// Global/Overloaded
// Result := A = B
//==============================================================================
function FEqu(A, B: TXFraction): Boolean;
begin
  Result := FEqu(A.fR,B.fR) and FEqu(A.fI,B.fI)
end;

//==============================================================================
// Global
// Result := |A| = |B|
//==============================================================================
function FEquAbs(A, B: TFraction): Boolean;
begin
  if A = B then Result := true
  else
  begin
    if not A.fNormalized then A.Normalize;
    if not B.fNormalized then B.Normalize;
    Result := IEquAbs(A.fN,B.fN) and IEqu(A.fD,B.fD);
  end;
end;

//==============================================================================
// Global
// A := A + 1
//==============================================================================
procedure FInc(A: TFraction);
begin
  with A do
  begin
    if not fNormalized then Normalize;
    IAdd(fN,fD);
    if fN.Is0 then fD.Set1; //Normalize
  end;
end;

//==============================================================================
// Global/Overloaded
// A := 1/A
//==============================================================================
procedure FInv(A: TFraction);
begin
  with A do
  begin
    if not fNormalized then Normalize;
    if fN.Is0 then HITRaiseDivByZero('FInv');

    ISwp(fN,fD);
    if fD.Neg then
    begin
      fN.Negate;
      IAbs(fD);
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := 1/A = ~A/|A|
//==============================================================================
procedure FInv(A: TXFraction);
  var N : TFraction;
begin
  //----- Check
  if A.Is0 then HITRaiseDivByZero('FInv');

  N := TFraction.Create;
  try
    A.Norm(N);
    A.Conjugate;
    FDiv(A.Re,N);
    FDiv(A.Im,N);
  finally
    N.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A * B
//==============================================================================
procedure FMul(A: TFraction; B: Int64);
begin
  with A do
  begin
    if not fNormalized then Normalize;
    IMul(fN,B);
    Normalize;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A * B
//==============================================================================
procedure FMul(A: TFraction; B: TInteger);
begin
  with A do
  begin
    if not fNormalized then Normalize;
    IMul(fN,B);
    Normalize;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A * B
//==============================================================================
procedure FMul(A, B: TFraction);
  var
    G, H, T  : TInteger;
    Selector : Integer;
begin
  if A = B then
  begin
    FSqr(A);
    Exit;
  end;

  if not A.fNormalized then A.Normalize;
  if not B.fNormalized then B.Normalize;

  if not A.Is0 then
    if B.Is0 then A.Set0
    else
    begin
      G := nil;
      H := nil;
      T := nil;
      try
        Selector := 0;
        G := A.fN.Clone;
        if IGCD(G,B.fD) then Inc(Selector);
        H := A.fD.Clone;
        if IGCD(H,B.fN) then Inc(Selector,2);
        case Selector of
          //----- G > 1 and H > 1
          0: begin
               //----- A.N = (A.N / G) * (B.N / H)
               IDiv(A.fN,G);
               T := B.fN.Clone;
               IDiv(T,H);
               IMul(A.fN,T);
               //----- A.D = (A.D / H) * (B.D / G)
               IDiv(A.fD,H);
               ISet(T,B.fD);
               IDiv(T,G);
               IMul(A.fD,T);
             end;
          //----- G = 1 and H > 1
          1: begin
               //----- A.N = A.N * (B.N / H)
               T := B.fN.Clone;
               IDiv(T,H);
               IMul(A.fN,T);
               //----- A.D = (A.D / h) * B.D
               IDiv(A.fD,H);
               IMul(A.fD,B.fD);
             end;
          //----- G > 1 and H = 1
          2: begin
               //----- A.N = (A.N / G) * B.N
               IDiv(A.fN,G);
               IMul(A.fN,B.fN);
               //----- A.D =  A.D * (B.D / G)
               T := B.fD.Clone;
               IDiv(T,G);
               IMul(A.fD,T);
             end;
          //3:
          else
             //----- G = 1 and H = 1
             begin
               IMul(A.fN,B.fN);
               IMul(A.fD,B.fD);
             end;
        end;
        //----- Normalize
        if A.fD.Neg then
        begin
          A.fN.Negate;
          A.fD.Negate;
        end;
      finally
        G.Free;
        H.Free;
        T.Free;
      end;
    end;
end;

//==============================================================================
// Global/Overloaded
// A := A * B
//==============================================================================
procedure FMul(A: TXFraction; B: TFraction);
begin
  FMul(A.fR,B);
  FMul(A.fI,B);
end;

//==============================================================================
// Global/Overloaded
// A := A * B = (Ar*Br - Ai*Bi, Ar*Bi + Ai*Br)
//==============================================================================
procedure FMul(A, B: TXFraction);
  var S, T : TFraction;
begin
  if A = B then FSqr(A)
  else
  with A do
  begin
    S := nil;
    T := nil;
    try
      T := fR.Clone;
      S := fI.Clone;
      FMul(fR,B.fR);
      FMul(S,B.fI);
      FSub(fR,S);
      FMul(fI,B.fR);
      FMul(T,B.fI);
      FAdd(fI,T);
    finally
      S.Free;
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A * 2
//==============================================================================
procedure FMul2(A: TFraction);
begin
  with A do
  begin
    if not fNormalized then Normalize;
    if fD.Odd then IShl(fN) else IShr(fD);
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A * 2**k
//==============================================================================
procedure FMul2(A: TFraction; k: Longint);
begin
  //----- Check
  if k < 0 then HITRaiseInvalidArg('FMul2','k < 0');

  with A do
  begin
    if not fNormalized then Normalize;
    IShl(fN,k);
    if not fD.Odd then Normalize;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A * 2
//==============================================================================
procedure FMul2(A: TXFraction);
begin
  FMul2(A.Re);
  FMul2(A.Im);
end;

//==============================================================================
// Global/Overloaded
// A := A * 2**k
//==============================================================================
procedure FMul2(A: TXFraction; k: Longint);
begin
  FMul2(A.Re,k);
  FMul2(A.Im,k);
end;

//==============================================================================
// Global/Overloaded
// A := A**E
//==============================================================================
procedure FPow(A: TFraction; E: Longint);
begin
  with A do
  if E = 0 then Set1
  else
  begin
    if not fNormalized then Normalize;

    if not Is0 then
    begin
      if E < 0 then
      begin
        E := -E;
        FInv(A);
      end;
      IPow(fN,E);
      IPow(fD,E);
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A^E
//==============================================================================
procedure FPow(A: TXFraction; E: TInteger);
  var
    T : TXFraction;
    i : Longint;
begin
  if E.Is0 then A.Set1
  else
  if not E.Is1 then
  begin
    if A.Is0 then Exit;
    if E.Neg then FInv(A);
    T := A.Clone;
    try
      //----- Repeated squaring exponentiation
      for i := E.BitSize-2 downto 0 do
      begin
        FSqr(A);
        if E.Bit[i] then FMul(A,T);
      end;
    finally
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global
// A := Random fraction
//==============================================================================
procedure FRnd(A: TFraction; BitCount: Longint; Signed: Boolean = false);
begin
  with A do
  begin
    IRnd(fN,BitCount,false,Signed);
    IRnd(fD,BitCount,false);
    if fD.Is0 then fD.Set1;
    Normalize;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := N
//==============================================================================
procedure FSet(A: TFraction; N: TInteger);
begin
  with A do
  begin
    ISet(fN,N);
    fD.Set1;
    fNormalized := true;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := N/D
//==============================================================================
procedure FSet(A: TFraction; N, D: TInteger);
begin
  //----- Check
  if D.Is0 then HITRaiseInvalidArg('FSet','D = 0');

  with A do
  begin
    ISet(fN,N);
    ISet(fD,D);
    Normalize;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := N
//==============================================================================
procedure FSet(A: TFraction; N: Int64);
begin
  with A do
  begin
    ISet(fN,N);
    fD.Set1;
    fNormalized := true;
  end;
end;

//==============================================================================
// Gloabal/Overloaded
// A := N/D
//==============================================================================
procedure FSet(A: TFraction; N, D: Int64);
begin
  //----- Check
  if D = 0 then HITRaiseInvalidArg('FSet','D = 0');

  with A do
  begin
    ISet(fN,N);
    ISet(fD,D);
    Normalize;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := B
//==============================================================================
procedure FSet(A, B: TFraction);
begin
  with A do
  begin
    ISet(fN,B.fN);
    ISet(fD,B.fD);
    fNormalized := B.fNormalized;
    if not fNormalized then Normalize;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := B
//==============================================================================
procedure FSet(A: TXFraction; R, I: Int64);
begin
  FSet(A.fR,R);
  FSet(A.fI,I);
end;

//==============================================================================
// Global/Overloaded
// A := B
//==============================================================================
procedure FSet(A: TXFraction; R, I: TFraction);
begin
  if Assigned(R) then FSet(A.fR,R); //Allow calls like "FSet(A,nil,I)"
  if Assigned(I) then FSet(A.fI,I);
end;

//==============================================================================
// Global/Overloaded
// A := B
//==============================================================================
procedure FSet(A, B: TXFraction);
begin
  if A <> B then
  begin
    FSet(A.fR,B.fR);
    FSet(A.fI,B.fI);
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A
//==============================================================================
procedure FSqr(A: TFraction);
begin
  with A do
  begin
    if not fNormalized then Normalize;
    ISqr(fN);
    ISqr(fD);
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A = (r-i, 2ri)
//==============================================================================
procedure FSqr(A: TXFraction);
  var S, T : TFraction;
begin
  S := nil;
  T := nil;
  with A do
  try
    S := fR.Clone;
    T := fI.Clone;
    FMul(fI,fR);
    FMul2(fI);
    FAdd(fR,T);
    FSub(S,T);
    FMul(fR,S);
  finally
    S.Free;
    T.Free;
  end;
end;

//------------------------------------------------------------------------------
// Compute a rational approximation of Sqrt(N) using Continued Fraction stuff
// -> N >= 0
// <- X > 0, such that X-NY = Result * K
// <- Y > 0, such that X-NY = Result * K
// <- Result = 1 (or 0 if N is a perfect square)
//------------------------------------------------------------------------------
function RationalApprox(N, X, Y: TInteger): Integer;
  const MAXCOUNT = 1000;
  var
    B, C, P    : array[0..2] of TInteger;
    i, j, k, t : Byte;
    Count      : Longint;
begin
  if N.Neg then HITRaiseInvalidArg('RationalApprox','N < 0');

  //----- Initialization ("0" is equivalent to "nil")
  FillChar(B,SizeOf(B),0);
  FillChar(C,SizeOf(C),0);
  FillChar(P,SizeOf(P),0);
  try
    for i := 0 to 2 do
    begin
      B[i] := TInteger.Create;
      C[i] := TInteger.Create;
      P[i] := TInteger.Create;
    end;
    //----- X := N^(1/2)
    ISet(X,N);
    ISqrRoot(X);
    //----- Initialize the B, C and P arrays
    ISet(C[1],X);
    ISqr(C[1]);
    //----- N perfect square ?
    if IEqu(C[1],N) then
    begin
      Y.Set1;
      Result := 0;
    end
    else
    begin
      ISet(B[1],X);  //B[1] := N^(1/2)
      C[0].Set1;     //C[0] := 1
      ISubr(C[1],N); //C[1] := N-(N^(1/2))
      P[0].Set1;     //P[0] := 1
      ISet(P[1],X);  //P[1] := N^(1/2)
      //----- Initialize the array indices
      i := 2; //i = (k-2) mod 3
      j := 0; //j = (k-1) mod 3
      k := 1;
      //----- Result of the Pell equation
      Result := -1;
      Count := MAXCOUNT;
      while (not C[k].Is1) and (Count > 0) do
      begin
        Dec(Count);
        //----- Rotation of the indices (avoids numerous "ISet").
        t := i;
        i := j;
        j := k;
        k := t;
        //----- Result of the Pell equation (1)
        Result := -Result;
        //----- Ak (stored in Bi) := (N^(1/2) + Bj) div Cj
        ISet(B[i],X);
        IAdd(B[i],B[j]);
        IDiv(B[i],C[j]);
        //----- Bk := Ak*Cj - Bj
        ISet(B[k],B[i]);
        IMul(B[k],C[j]);
        ISub(B[k],B[j]);
        //----- Ck := Ak*(Bj-Bk) + Ci
        ISet(C[k],B[j]);
        ISub(C[k],B[k]);
        IMul(C[k],B[i]);
        IAdd(C[k],C[i]);
        //----- Pk := Ak*Pj + Pi
        ISet(P[k],B[i]);
        IMul(P[k],P[j]);
        IAdd(P[k],P[i]);
      end;
      //----- Solution (X,Y)
      ISwp(X,P[k]); //Faster than "ISet(X,P[k])"
      ISet(Y,X);
      ISqr(Y);
      if C[k].Is1 then ISub(Y,Result)
      else
      begin
        ISqr(C[k]);
        if Result = -1 then IAdd(Y,C[k]) else ISub(Y,C[k]);
      end;
      IDiv(Y,N);
      ISqrRoot(Y);
    end;
  finally
    //----- Free the integers
    for i := 0 to 2 do
    begin
      B[i].Free;
      C[i].Free;
      P[i].Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A^(1/2)
// The resulting fraction R is such that R <= A
//==============================================================================
procedure FSqrRoot(A: TFraction);
  var X, Y : TInteger;
begin
  with A do
  begin
    if Neg then HITRaiseInvalidArg('FSqrRoot','A < 0');

    if not fNormalized then Normalize;
    IMul(fN,fD);
    X := nil;
    Y := nil;
    try
      X := TInteger.Create;
      Y := TInteger.Create;
      if RationalApprox(fN,X,Y) <> 0 then IMul(fD,Y);
      ISet(fN,X);
      Normalize;
    finally
      X.Free;
      Y.Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A^(1/2)
//==============================================================================
procedure FSqrRoot(A: TXFraction);
  var
    T              : TFraction;
    ReTest, ImTest : Boolean;
begin
  if A.Is0 then Exit;

  T := TFraction.Create;
  with A do
  try
    ReTest := fR.Sign <= 0;
    ImTest := fI.Neg;
    Norm(T);
    FSqrRoot(T);
    FAbs(fR);
    FAdd(fR,T);
    FDiv2(fR);
    FSqrRoot(fR);
    FDiv(fI,fR);
    FDiv2(fI);
    if ReTest then
    begin
      FSwp(fR,fI);
      if ImTest then Negate;
    end;
  finally
    T.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A - B
//==============================================================================
procedure FSub(A: TFraction; B: Int64);
  var T : TInteger;
begin
  with A do
  begin
    if not fNormalized then Normalize;
    T := fD.Clone;
    try
      IMul(T,B);
      ISub(fN,T);
      Normalize;
    finally
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A - B
//==============================================================================
procedure FSub(A: TFraction; B: TInteger);
  var T : TInteger;
begin
  with A do
  begin
    if not fNormalized then Normalize;
    T := fD.Clone;
    try
      IMul(T,B);
      ISub(fN,T);
      Normalize;
    finally
      T.Free;
    end;
  end;
end;

//==============================================================================
// Global/Overloaded
// A := A - B
//==============================================================================
procedure FSub(A, B: TFraction);
  var G, T : TInteger;
begin
  if A = B then
  begin
    A.Set0;
    Exit;
  end;
  if not A.fNormalized then A.Normalize;
  if not B.fNormalized then B.Normalize;

  if not B.Is0 then
    if A.Is0 then
    begin
      FSet(A,B);
      A.Negate;
    end
    else
    begin
      G := nil;
      T := nil;
      try
        G := A.fD.Clone;
        if IGCD(G,B.fD) then
        begin
          //----- GCD = 1 (for random inputs proba = 6/PI ~ 0.6)
          T := A.fD.Clone;
          IMul(T,B.fN);
          IMul(A.fN,B.fD);
          ISub(A.fN,T);
          IMul(A.fD,B.fD);
        end
        else
        begin
          IDiv(A.fD,G);
          T := B.fD.Clone;
          IDiv(T,G);
          IMul(A.fN,T);
          ISet(T,B.fN);
          IMul(T,A.fD);
          ISub(A.fN,T);
          ISet(T,B.fD);
          if not IGCD(G,A.fN) then
          begin
            IDiv(A.fN,G);
            IDiv(T,G);
          end;
          IMul(A.fD,T);
        end;
        //----- Normalize
        if A.fD.Neg then
        begin
          A.fN.Negate;
          A.fD.Negate;
        end;
      finally
        G.Free;
        T.Free;
      end;
    end;
end;

//==============================================================================
// Global/Overloaded
// A := A - B
//==============================================================================
procedure FSub(A, B: TXFraction);
begin
  FSub(A.fR,B.fR);
  FSub(A.fI,B.fI);
end;

//==============================================================================
// Global/Overloaded
// A <-> B
//==============================================================================
procedure FSwp(A, B: TFraction);
  var t : Boolean;
begin
  ISwp(A.fN,B.fN);
  ISwp(A.fD,B.fD);
  t := A.fNormalized; A.fNormalized := B.fNormalized; B.fNormalized := t;
end;

//==============================================================================
// Global/Overloaded
// Swap the values of A and B
//==============================================================================
procedure FSwp(A, B: TXFraction);
begin
  FSwp(A.fR,B.fR);
  FSwp(A.fI,B.fI);
end;

////////////////////////////////////////////////////////////////////////////////
// TFraction class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Public/Overloaded
// Create a fraction from the Last+1 coefficients of a CFE
// CFE ~ Continued Fraction Expansion
//==============================================================================
procedure TFraction.BuildFromCFE(CFE: TCFEFunc; Last: Word);
  var
    X : TInteger;
    i : Longword;
begin
  X := TInteger.Create;
  try
    //----- Initialize
    fN.Set1; //N := 1
    fD.Set0; //D := 0
    //----- Iterate
    for i := Last downto 0 do
    begin
      ISet(X,fN);
      ISwp(fN,fD);
      IMul(X,CFE(i));
      IAdd(fN,X);
    end;
    Normalize;
  finally
    X.Free;
  end;
end;

//------------------------------------------------------------------------------
// Private
// Create a fraction from the coefficients of a CFE
// CFE ~ Continued Fraction Expansion
//------------------------------------------------------------------------------
// If CFE = [a0, a1, a2, ..., aN] then
//   [a0]              is the integer part
//   [a1, a2, ..., aN] is the periodic part
// Example:
// With [3,1,1,1,1,6], that is the CFE of Sqrt(13),
// the call
//    "BuildFromCFE([3,1,1,1,1,6],3)"
// is equivalent to
//    "BuildFromCFE([3,1,1,1,1,6,1,1,1,1,6,1,1,1,1,6],1)"
//------------------------------------------------------------------------------
procedure TFraction.BuildFromCFELoc(const CFE: array of Longword; Rep: Word);
  var
    X    : TInteger;
    i, j : Longword;
begin
  if Length(CFE) = 0 then
  begin
    Set0;
    Exit;
  end;

  X := TInteger.Create;
  try
    //----- Initialize
    fN.Set1; //N := 1
    fD.Set0; //D := 0
    //----- Iterate
    for j := 1 to Rep do
      for i := High(CFE) downto 1 do
      begin
        ISet(X,fN);
        ISwp(fN,fD);
        IMul(X,CFE[i]);
        IAdd(fN,X);
      end;
    ISet(X,fN);
    ISwp(fN,fD);
    IMul(X,CFE[0]);
    IAdd(fN,X);
    fNormalized := true;
  finally
    X.Free;
  end;
end;

//==============================================================================
// Public/Overloaded
// Create a fraction from the coefficients of a CFE
// CFE ~ Continued Fraction Expansion
//==============================================================================
procedure TFraction.BuildFromCFE(const CFE: array of Longword; Repetition: Word);
begin
  BuildFromCFELoc(CFE,Repetition);
end;

//==============================================================================
// Public/Overloaded
// Create a fraction from the coefficients of a CFE
// CFE ~ Continued Fraction Expansion
//==============================================================================
procedure TFraction.BuildFromCFE(const CFE: TDynLongArray; Repetition: Word);
begin
  BuildFromCFELoc(CFE,Repetition);
end;

//==============================================================================
// Public
// Create a fraction equal to Self and sets Result with it
//==============================================================================
function TFraction.Clone: TFraction;
begin
  Result := TFraction.Create(Self);
end;

//==============================================================================
// Public
// Self := A
//==============================================================================
procedure TFraction.Copy(A: TFraction);
begin
  ISet(fN,A.fN);
  ISet(fD,A.fD);
  fNormalized := A.fNormalized;
end;

//==============================================================================
// Public
// Create Self equal to 0
//==============================================================================
constructor TFraction.Create;
begin
  inherited Create;
  fN := TInteger.Create;
  fD := TInteger.Create(1);
  fNormalized := true;
end;

//==============================================================================
// Public
// Create Self equal to A
//==============================================================================
constructor TFraction.Create(A: TFraction);
begin
  inherited Create;
  fN := A.fN.Clone;
  fD := A.fD.Clone;
  fNormalized := A.fNormalized;
end;

//==============================================================================
// Public
// Create Self equal to N
//==============================================================================
constructor TFraction.Create(N: Int64);
begin
  inherited Create;
  fN := TInteger.Create(N);
  fD := TInteger.Create(1);
  fNormalized := true;
end;

//==============================================================================
// Public
// Create Self equal to N/D
//==============================================================================
constructor TFraction.Create(N, D: Int64);
begin
  inherited Create;
  fN := TInteger.Create(N);
  fD := TInteger.Create(D);
  Normalize;
end;

//==============================================================================
// Public
// Create Self equal to N
//==============================================================================
constructor TFraction.Create(N: TInteger);
begin
  inherited Create;
  fN := N.Clone;
  fD := TInteger.Create(1);
  fNormalized := true;
end;

//==============================================================================
// Public
// Create Self equal to N/D
//==============================================================================
constructor TFraction.Create(N, D: TInteger);
begin
  inherited Create;
  fN := N.Clone;
  fD := D.Clone;
  Normalize;
end;

//==============================================================================
// Public
// Free Self
//==============================================================================
destructor TFraction.Destroy;
begin
  fN.Free;
  fD.Free;
  inherited Destroy;
end;

//------------------------------------------------------------------------------
// Private
// Returns a pointer on fD
// fNormalized is set to FALSE because fD might be modified
//------------------------------------------------------------------------------
function TFraction.GetDen: TInteger;
begin
  fNormalized := false;
  Result := fD;
end;

//------------------------------------------------------------------------------
// Private
// Result := TRUE if Self is negative, FALSE otherwise
//------------------------------------------------------------------------------
function TFraction.GetNeg: Boolean;
begin
  if not fNormalized then Normalize;
  Result := fN.Neg;
end;

//------------------------------------------------------------------------------
// Private
// Returns a pointer on fN
// fNormalized is set to FALSE because fN might be modified
//------------------------------------------------------------------------------
function TFraction.GetNum: TInteger;
begin
  fNormalized := false;
  Result := fN;
end;

//------------------------------------------------------------------------------
// Private
// Result := -1 if Self < 0, 0 if Self = 0, 1 if Self > 0
//------------------------------------------------------------------------------
function TFraction.GetSign: Integer;
begin
  if not fNormalized then Normalize;
  Result := fN.Sign;
end;

//==============================================================================
// Public
// Result := TRUE if Self = 0, FALSE otherwise
//==============================================================================
function TFraction.Is0: Boolean;
begin
  if not fNormalized then Normalize;
  Result := fN.Is0;
end;

//==============================================================================
// Public
// Result := TRUE if Self = 1, FALSE otherwise
//==============================================================================
function TFraction.Is1: Boolean;
begin
  if not fNormalized then Normalize;
  Result := fN.Is1 and fD.Is1;
end;

//==============================================================================
// Public
// Result := TRUE if Self = -1, FALSE otherwise
//==============================================================================
function TFraction.IsMinus1: Boolean;
begin
  if not fNormalized then Normalize;
  Result := fN.IsMinus1 and fD.Is1;
end;

//------------------------------------------------------------------------------
// Private
// Result := TRUE if Self is an integer, FALSE otherwise
//------------------------------------------------------------------------------
function TFraction.IsIntegerQ: Boolean;
begin
  if not fNormalized then Normalize;
  Result := fD.Is1;
end;

//==============================================================================
// Public/Virtual
//==============================================================================
procedure TFraction.LoadFromFile(const FileName: string);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------------------------
  var
    F : TFileStream;
    H : TFracHeaderFile;
begin
  F := TFileStream.Create(FileName,HITLOADFLAGS);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> FracHeaderFile then Error;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public/Virtual
//==============================================================================
procedure TFraction.LoadFromStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    fN.LoadFromStream(Stream);
    fD.LoadFromStream(Stream);
    Stream.ReadBuffer(fNormalized,1);
  except
    Stream.Position := pos;
    Set0;
    HITRaiseStreamError(ClassName + '.LoadFromStream');
  end;
end;

//==============================================================================
// Public
// Self := -Self
//==============================================================================
procedure TFraction.Negate;
begin
  if not fNormalized then Normalize;
  fN.Negate;
end;

//------------------------------------------------------------------------------
// Private
// Reduce Self
//------------------------------------------------------------------------------
procedure TFraction.Normalize;
  var
    sgn : Integer;
    G   : TInteger;
begin
  sgn := fD.Sign;
  if sgn = 0 then HITRaiseDivByZero(ClassName + '.Normalize (private)');

  //----- Correct sign
  if sgn < 0 then
  begin
    fN.Negate;
    IAbs(fD);
  end;

  //----- N = 0 ?
  if fN.Is0 then fD.Set1
  else
  begin
    G := fN.Clone;
    try
      if not IGCD(G,fD) then //G cannot equal 0
      begin
        IDiv(fN,G);
        IDiv(fD,G);
      end;
    finally
      G.Free;
    end;
  end;
  fNormalized := true;
end;

//==============================================================================
// Public/Virtual
//==============================================================================
procedure TFraction.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,HITSAVEFLAGS);
  try
    F.WriteBuffer(FracHeaderFile,SizeOf(FracHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public/Virtual
//==============================================================================
procedure TFraction.SaveToStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    fN.SaveToStream(Stream);
    fD.SaveToStream(Stream);
    Stream.WriteBuffer(fNormalized,1);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
// Public
// Self := 0
//==============================================================================
procedure TFraction.Set0;
begin
  fN.Set0;
  fD.Set1;
  fNormalized := true;
end;

//==============================================================================
// Public
// Self := 1
//==============================================================================
procedure TFraction.Set1;
begin
  fN.Set1;
  fD.Set1;
  fNormalized := true;
end;

//==============================================================================
// Public
// Self := -1
//==============================================================================
procedure TFraction.SetMinus1;
begin
  fN.SetMinus1;
  fD.Set1;
  fNormalized := true;
end;

//------------------------------------------------------------------------------
// Private
// Set Self negative if OnOff and (Self <> 0), otherwise set Self positive
//------------------------------------------------------------------------------
procedure TFraction.SetNeg(OnOff: Boolean);
begin
  if not fNormalized then Normalize;
  fN.Neg := OnOff;
end;

//==============================================================================
// Public
// Q = |fN| div fD, R = |fN| rem fD,
// ! The returned Q and R are always nonnegative
//==============================================================================
procedure TFraction.SplitAbs(Q, R: TInteger);
begin
  if not fNormalized then Normalize;
  ISet(R,fN);
  R.Neg := false;
  ISet(Q,fD);
  IRemDiv(R,Q);
end;

////////////////////////////////////////////////////////////////////////////////
// TFractionVector class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Public
// Return the maximal size of a vector (highly theoretical!)
//==============================================================================
class function TFractionVector.MaxSize: Longint;
begin
  Result := MaxListSize;
end;

//==============================================================================
// Public
// Add a new cell
//==============================================================================
procedure TFractionVector.AddCell(A: TFraction = nil);
begin
  //----- Check
  if Longword(fSize) >= MaxListSize then
    HITRaiseSizeError(ClassName + '.AddCell');

  //----- Add a new cell
  ReallocMem(fCell,(fSize+1) shl 2);
  if Assigned(A) then TVCEll(fCell^)[fSize] := A.Clone
                 else TVCEll(fCell^)[fSize] := TFracTion.Create;
  Inc(fSize);
end;

//==============================================================================
// Public
// Create Result and set it equal to Self
//==============================================================================
function TFractionVector.Clone: TFractionVector;
begin
  Result := TFractionVector.Create;
  Result.Copy(Self);
end;

//==============================================================================
// Public
// Set Self equal to V
//==============================================================================
procedure TFractionVector.Copy(V: TFractionVector);
  var i : Longint;
begin
  if V <> Self then
  begin
    SetSize(V.fSize);
    for i := 0 to fSize-1 do FSet(TVCell(fCell^)[i],TVCell(V.fCell^)[i]);
  end;
end;

//==============================================================================
// Public
//==============================================================================
destructor TFractionVector.Destroy;
begin
  SetSize(0);
  inherited Destroy;
end;

//==============================================================================
// Public
// Fill all cells with A
//==============================================================================
procedure TFractionVector.Fill(A: TFraction);
  var i : Longint;
begin
  for i := 0 to fSize-1 do FSet(TVCell(fCell^)[i],A);
end;

//------------------------------------------------------------------------------
// Private
//------------------------------------------------------------------------------
function TFractionVector.GetCellPtr(I: Longint): TFraction;
begin
  //----- Check
  if Longword(I) >= Longword(fSize) then
    HITRaiseIndexError(ClassName + '.Cell');

  Result := TVCell(fCell^)[I];
end;

//==============================================================================
// Public
// Result := (Cell[i] = 0 for all i) or (Self is empty)
//==============================================================================
function TFractionVector.IsNull: Boolean;
  var i : Longint;
begin
  for i := 0 to fSize-1 do
    if not TVCell(fCell^)[i].Is0 then
    begin
      Result := false;
      Exit;
    end;
  Result := true;
end;

//==============================================================================
// Public
// Load Self from a HIT file
//==============================================================================
procedure TFractionVector.LoadFromFile(const FileName: string);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------------------------
  var
    F : TFileStream;
    H : TFracVectorHeaderFile;
begin
  F := TFileStream.Create(FileName,HITLOADFLAGS);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> FracVectorHeaderFile then Error;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public
// Load Self from a stream
//==============================================================================
procedure TFractionVector.LoadFromStream(Stream: TStream);
  var pos, temp, i : Longint; //! temp must be 4-byte sized
begin
  pos := Stream.Position;
  try
    //----- Size
    Stream.ReadBuffer(temp,4);
    //----- Fractions
    SetSize(temp);
    for i := 0 to fSize-1 do TVCell(fCell^)[i].LoadFromStream(Stream);
  except
    Stream.Position := pos;
    SetSize(0);
    HITRaiseStreamError(ClassName + '.LoadFromStream');
  end;
end;

//==============================================================================
// Public
// Set Self equal to -Self
//==============================================================================
procedure TFractionVector.Negate;
  var i : Longint;
begin
  for i := 0 to fSize-1 do TVCell(fCell^)[i].Negate;
end;

//==============================================================================
// Public
// Save Self into a HIT file
//==============================================================================
procedure TFractionVector.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,HITSAVEFLAGS);
  try
    F.WriteBuffer(FracVectorHeaderFile,SizeOf(FracVectorHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public
// Save Self into a stream
//==============================================================================
procedure TFractionVector.SaveToStream(Stream: TStream);
  var pos, i : Longint; //! temp must be 4-byte sized
begin
  pos := Stream.Position;
  try
    //----- Size and sign
    Stream.WriteBuffer(fSize,4);
    //----- Fractions
    for i := 0 to fSize-1 do TVCell(fCell^)[i].SaveToStream(Stream);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
// Public
// Set all cells to 0
//==============================================================================
procedure TFractionVector.SetNull;
  var i : Longint;
begin
  for i := 0 to fSize-1 do TVCell(fCell^)[i].Set0;
end;

//------------------------------------------------------------------------------
// Private
//------------------------------------------------------------------------------
procedure TFractionVector.SetSize(Value: Longint);
  var i : Longint;
begin
  //----- Check
  if Longword(Value) > MaxListSize then HITRaiseSizeError(ClassName + '.Size');

  if Value > fSize then
  begin
    ReallocMem(fCell,Value shl 2);
    for i := fSize to Value-1 do TVCell(fCell^)[i] := TFraction.Create;
    fSize := Value;
  end
  else
  if Value < fSize then
  begin
    for i := Value to fSize-1 do TVCell(fCell^)[i].Free;
    ReallocMem(fCell,Value shl 2);
    fSize := Value;
  end;
end;

//------------------------------------------------------------------------------
// Local
// Sort the vector V[a]..V[z]
//------------------------------------------------------------------------------
procedure QSort(V: TFractionVector; a, z: Longint; Cmp: TFracCmpFunction);
  var
    i, j, k : Longint;
    X       : TFraction;
begin
  with V do
  repeat
    i := a;
    j := z;
    k := (i + j) shr 1;
    X := TVCell(fCell^)[k]; //Just copy the pointer
    repeat
      while Cmp(TVCell(fCell^)[i],X) < 0 do Inc(i);
      while Cmp(TVCell(fCell^)[j],X) > 0 do Dec(j);
      if i <= j then
      begin
        FSwp(TVCell(fCell^)[i],TVCell(fCell^)[j]);
        //----- The value pointed by X may have changed!
        if k = i then
        begin
          k := j;
          X := TVCell(fCell^)[k];
        end
        else
        if k = j then
        begin
          k := i;
          X := TVCell(fCell^)[k];
        end;
        Inc(i);
        Dec(j);
      end;
    until i > j;
    if a < j then QSort(V,a,j,Cmp);
    a := i;
  until i >= z;
end;

//==============================================================================
// Public
// Sort the vector
//==============================================================================
procedure TFractionVector.Sort(AbsValues: Boolean = false);
begin
  if fSize > 1 then
    if AbsValues then QSort(Self,0,fSize-1,FCmpAbs)
                 else QSort(Self,0,fSize-1,FCmp);
end;

//==============================================================================
// Public
// Exchange Cell[I] with Cell[J]
//==============================================================================
procedure TFractionVector.SwpCells(I, J: Longint);
begin
  //----- Check
  if (Longword(I) >= Longword(fSize)) or (Longword(J) >= Longword(fSize)) then
    HITRaiseIndexError(ClassName + '.SwpCells');

  if I <> J then FSwp(TVCell(fCell^)[I],TVCell(fCell^)[J]);
end;

////////////////////////////////////////////////////////////////////////////////
// TXFraction class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Public
// Create and set Result equal to Self
//==============================================================================
function TXFraction.Clone: TXFraction;
begin
  Result := TXFraction.Create(Self);
end;

//==============================================================================
// Public
// Conjugate the number
//==============================================================================
procedure TXFraction.Conjugate;
begin
  fI.Negate;
end;

//==============================================================================
// Public
// Self := Z
//==============================================================================
procedure TXFraction.Copy(F: TXFraction);
begin
  if Self <> F then
  begin
    FSet(fR,F.fR);
    FSet(fI,F.fI);
  end;
end;

//==============================================================================
// Public/Overloaded
// Create an instance of TXFraction
//==============================================================================
constructor TXFraction.Create(R, I: Int64);
begin
  inherited Create;
  fR := TFraction.Create(R);
  fI := TFraction.Create(I);
end;

//==============================================================================
// Public/Overloaded
// Create an instance of TXFraction
//==============================================================================
constructor TXFraction.Create(R, I: TFraction);
begin
  inherited Create;
  fR := R.Clone;
  fI := I.Clone;
end;

//==============================================================================
// Public/Overloaded
// Create an instance of TXFraction
//==============================================================================
constructor TXFraction.Create(F: TXFraction);
begin
  inherited Create;
  fR := F.fR.Clone;
  fI := F.fI.Clone;
end;

//==============================================================================
// Public/Overloaded
// Create an instance of TXFraction
//==============================================================================
constructor TXFraction.Create;
begin
  inherited Create;
  fR := TFraction.Create;
  fI := TFraction.Create;
end;

//==============================================================================
// Public
// Destroy the instance
//==============================================================================
destructor TXFraction.Destroy;
begin
  fR.Free;
  fI.Free;
  inherited Destroy;
end;

//==============================================================================
// Public
// Return "TRUE" if and only if the number is equal to (0,0)
//==============================================================================
function TXFraction.Is0: Boolean;
begin
  Result := fR.Is0 and fI.Is0;
end;

//==============================================================================
// Public
// Return "TRUE" if and only if the number is equal to (1,0)
//==============================================================================
function TXFraction.Is1: Boolean;
begin
  Result := fR.Is1 and fI.Is0;
end;

//==============================================================================
// Public
// Return "TRUE" if and only if the number is equal to (0,1)
//==============================================================================
function TXFraction.IsI: Boolean;
begin
  Result := fR.Is0 and fI.Is1;
end;

//==============================================================================
// Public
// Return "TRUE" if and only if the number is a unit of Z[i]
// Units = {(1,0), (-1,0), (0,1), (0,-1)}
//==============================================================================
function TXFraction.IsUnit: Boolean;
begin
  Result := (fR.Is0 and (fI.Is1 or fI.IsMinus1)) or
            ((fR.Is1 or fR.IsMinus1) and fI.Is0);
end;

//==============================================================================
// Public
// Load the number from a file
//==============================================================================
procedure TXFraction.LoadFromFile(const FileName: string);
  var
    F : TFileStream;
    H : TXFracHeaderFile;
begin
  F := TFileStream.Create(FileName,HITLOADFLAGS);
  try
    //----- Header
    F.ReadBuffer(H,SizeOf(H));
    if (H <> XFracHeaderFile) then
      HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
    //----- Values
    LoadFromStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public
// Load the number from a stream
//==============================================================================
procedure TXFraction.LoadFromStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    fR.LoadFromStream(Stream);
    fI.LoadFromStream(Stream);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
// Public
// Set Self equal to Self * i
//==============================================================================
procedure TXFraction.MulI;
begin
  FSwp(fR,fI);
  fR.Negate;
end;

//==============================================================================
// Public
// Set Self equal to Self * -i
//==============================================================================
procedure TXFraction.MulMinusI;
begin
  FSwp(fR,fI);
  fI.Negate;
end;

//==============================================================================
// Public
// Negate the number
//==============================================================================
procedure TXFraction.Negate;
begin
  fR.Negate;
  fI.Negate;
end;

//==============================================================================
// Public
// Calculate the norm of the number
// F := R + I
//==============================================================================
procedure TXFraction.Norm(F: TFraction);
  var T : TFraction;
begin
  T := fI.Clone;
  try
    FSet(F,fR);
    FSqr(F);
    FSqr(T);
    FAdd(F,T);
  finally
    T.Free;
  end;
end;

//==============================================================================
// Public
// Save the number to a file
//==============================================================================
procedure TXFraction.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,HITSAVEFLAGS);
  try
    //----- Header
    F.WriteBuffer(XFracHeaderFile,SizeOf(TXFracHeaderFile));
    //----- Values
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public
// Save the fraction to a stream
//==============================================================================
procedure TXFraction.SaveToStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    fR.SaveToStream(Stream);
    fI.SaveToStream(Stream);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
// Public
// Self := (0,0)
//==============================================================================
procedure TXFraction.Set0;
begin
  fR.Set0;
  fI.Set0;
end;

//==============================================================================
// Public
// Self := (1,0)
//==============================================================================
procedure TXFraction.Set1;
begin
  fR.Set1;
  fI.Set0;
end;

//==============================================================================
// Public
// Self.Im := I
//==============================================================================
procedure TXFraction.SetIm(I: Int64);
begin
  FSet(fI,I);
end;

//==============================================================================
// Public
// Self.Im := I
//==============================================================================
procedure TXFraction.SetIm(I: TFraction);
begin
  FSet(fI,I);
end;

//==============================================================================
// Public
// Self.Re := R
//==============================================================================
procedure TXFraction.SetRe(R: Int64);
begin
  FSet(fR,R);
end;

//==============================================================================
// Public
// Self.Re := R
//==============================================================================
procedure TXFraction.SetRe(R: TFraction);
begin
  FSet(fR,R);
end;

////////////////////////////////////////////////////////////////////////////////
// TXFractionVector class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Public
// Return the maximal size of a vector (highly theoretical!)
//==============================================================================
class function TXFractionVector.MaxSize: Longint;
begin
  Result := MaxListSize;
end;

//==============================================================================
// Public
// Add a new cell
//==============================================================================
procedure TXFractionVector.AddCell(A: TXFraction = nil);
begin
  //----- Check
  if Longword(fSize) >= MaxListSize then
    HITRaiseSizeError(ClassName + '.AddCell');

  //----- Add a new cell
  ReallocMem(fCell,(fSize+1) shl 2);
  if Assigned(A) then TXVCEll(fCell^)[fSize] := A.Clone
                 else TXVCEll(fCell^)[fSize] := TXFracTion.Create;
  Inc(fSize);
end;

//==============================================================================
// Public
// Create Result and set it equal to Self
//==============================================================================
function TXFractionVector.Clone: TXFractionVector;
begin
  Result := TXFractionVector.Create;
  Result.Copy(Self);
end;

//==============================================================================
// Public
// Set Self equal to V
//==============================================================================
procedure TXFractionVector.Copy(V: TXFractionVector);
  var i : Longint;
begin
  if V <> Self then
  begin
    SetSize(V.fSize);
    for i := 0 to fSize-1 do FSet(TXVCEll(fCell^)[i],TXVCEll(V.fCell^)[i]);
  end;
end;

//==============================================================================
// Public
//==============================================================================
destructor TXFractionVector.Destroy;
begin
  SetSize(0);
  inherited Destroy;
end;

//==============================================================================
// Public
// Fill all cells with A
//==============================================================================
procedure TXFractionVector.Fill(A: TXFraction);
  var i : Longint;
begin
  for i := 0 to fSize-1 do FSet(TXVCEll(fCell^)[i],A);
end;

//------------------------------------------------------------------------------
// Private
//------------------------------------------------------------------------------
function TXFractionVector.GetCellPtr(I: Longint): TXFraction;
begin
  //----- Check
  if Longword(I) >= Longword(fSize) then
    HITRaiseIndexError(ClassName + '.Cell');

  Result := TXVCEll(fCell^)[I];
end;

//==============================================================================
// Public
// Result := (Cell[i] = 0 for all i) or (Self is empty)
//==============================================================================
function TXFractionVector.IsNull: Boolean;
  var i : Longint;
begin
  for i := 0 to fSize-1 do
    if not TXVCEll(fCell^)[i].Is0 then
    begin
      Result := false;
      Exit;
    end;
  Result := true;
end;

//==============================================================================
// Public
// Load Self from a HIT file
//==============================================================================
procedure TXFractionVector.LoadFromFile(const FileName: string);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------------------------
  var
    F : TFileStream;
    H : TXFracVectorHeaderFile;
begin
  F := TFileStream.Create(FileName,HITLOADFLAGS);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> XFracVectorHeaderFile then Error;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public
// Load Self from a stream
//==============================================================================
procedure TXFractionVector.LoadFromStream(Stream: TStream);
  var pos, temp, i : Longint; //! temp must be 4-byte sized
begin
  pos := Stream.Position;
  try
    //----- Size
    Stream.ReadBuffer(temp,4);
    //----- Fractions
    SetSize(temp);
    for i := 0 to fSize-1 do TXVCEll(fCell^)[i].LoadFromStream(Stream);
  except
    Stream.Position := pos;
    SetSize(0);
    HITRaiseStreamError(ClassName + '.LoadFromStream');
  end;
end;

//==============================================================================
// Public
// Set Self equal to -Self
//==============================================================================
procedure TXFractionVector.Negate;
  var i : Longint;
begin
  for i := 0 to fSize-1 do TXVCEll(fCell^)[i].Negate;
end;

//==============================================================================
// Public
// Save Self into a HIT file
//==============================================================================
procedure TXFractionVector.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,HITSAVEFLAGS);
  try
    F.WriteBuffer(XFracVectorHeaderFile,SizeOf(XFracVectorHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public
// Save Self into a stream
//==============================================================================
procedure TXFractionVector.SaveToStream(Stream: TStream);
  var pos, i : Longint; //! temp must be 4-byte sized
begin
  pos := Stream.Position;
  try
    //----- Size and sign
    Stream.WriteBuffer(fSize,4);
    //----- Fractions
    for i := 0 to fSize-1 do TXVCEll(fCell^)[i].SaveToStream(Stream);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
// Public
// Set all cells to 0
//==============================================================================
procedure TXFractionVector.SetNull;
  var i : Longint;
begin
  for i := 0 to fSize-1 do TXVCEll(fCell^)[i].Set0;
end;

//------------------------------------------------------------------------------
// Private
//------------------------------------------------------------------------------
procedure TXFractionVector.SetSize(Value: Longint);
  var i : Longint;
begin
  //----- Check
  if Longword(Value) > MaxListSize then HITRaiseSizeError(ClassName + '.Size');

  if Value > fSize then
  begin
    ReallocMem(fCell,Value shl 2);
    for i := fSize to Value-1 do TXVCEll(fCell^)[i] := TXFraction.Create;
    fSize := Value;
  end
  else
  if Value < fSize then
  begin
    for i := Value to fSize-1 do TXVCEll(fCell^)[i].Free;
    ReallocMem(fCell,Value shl 2);
    fSize := Value;
  end;
end;

//==============================================================================
// Public
// Exchange Cell[I] with Cell[J]
//==============================================================================
procedure TXFractionVector.SwpCells(I, J: Longint);
begin
  //----- Check
  if (Longword(I) >= Longword(fSize)) or (Longword(J) >= Longword(fSize)) then
    HITRaiseIndexError(ClassName + '.SwpCells');

  if I <> J then FSwp(TXVCEll(fCell^)[I],TXVCEll(fCell^)[J]);
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

