////////////////////////////////////////////////////////////////////////////////
// RIPEMD-128 and RIPEMD-160 Hash Functions
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
//------------------------------------------------------------------------------
// RIPEMD-128 and RIPEMD-160 were designed in 1996 by Hans Dobbertin, Antoon
// Bosselaers and Bart Preneel.
////////////////////////////////////////////////////////////////////////////////
unit HIT_Hash;

{$I HIT_SYMBOLS.INC}

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers, GF2, Classes;

procedure HITByteSwap(P: TGF2Poly); overload;
procedure HITByteSwap(A: TInteger); overload;

procedure GF2H128(P: TGF2Poly); overload;
procedure GF2H128(P: TGF2Poly; Deg: Longint; Index: Longword); overload;
procedure GF2H128(P: TGF2Poly; const S: AnsiString); overload;
procedure GF2H128(P: TGF2Poly; Stream: TStream); overload;
procedure GF2H160(P: TGF2Poly); overload;
procedure GF2H160(P: TGF2Poly; Deg: Longint; Index: Longword); overload;
procedure GF2H160(P: TGF2Poly; const S: AnsiString); overload;
procedure GF2H160(P: TGF2Poly; Stream: TStream); overload;

procedure IH128(A: TInteger); overload;
procedure IH128(A: TInteger; BitCount: Longint; Index: Longword); overload;
procedure IH128(A: TInteger; const S: AnsiString); overload;
procedure IH128(A: TInteger; Stream: TStream); overload;
procedure IH160(A: TInteger); overload;
procedure IH160(A: TInteger; BitCount: Longint; Index: Longword); overload;
procedure IH160(A: TInteger; const S: AnsiString); overload;
procedure IH160(A: TInteger; Stream: TStream); overload;

implementation
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// Resource strings
////////////////////////////////////////////////////////////////////////////////

resourcestring
  sHSH0001 = '(BitCount < 0) or (BitCount > TInteger.MaxBitSize)';
  sHSH0002 = '(k < 0) or (k > TGF2Poly.MaxDegree)';

////////////////////////////////////////////////////////////////////////////////
// Types, Consts
////////////////////////////////////////////////////////////////////////////////

type
  TMD128 = array[0..3] of Longword;
  TMD160 = array[0..4] of Longword;
  THWork = record
    Buffer   : TMD160;
    CountLo  : Longword;
    CountHi  : Longword;
    RemBytes : Longint;
    case Integer of
      0: (BData : array[0..63] of Byte);
      1: (LData : array[0..15] of Longword);
  end;

const
  //----- Standard Initialization Vector
  SIV : TMD160 = ($67452301,$efcdab89,$98badcfe,$10325476,$c3d2e1f0);


////////////////////////////////////////////////////////////////////////////////
// RIPEMD-128 hash function
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Local
// Compress function of RIPEMD-128
// With lookup tables and loops, this routine could be more compact but, here,
// it is the running time that is important.
//------------------------------------------------------------------------------
procedure RMD128Mix(var Work : THWork);
  const
    K2 = $5a827999;
    K3 = $6ed9eba1;
    K4 = $8f1bbcdc;
    K5 = $50a28be6;
    K6 = $5c4dd124;
    K7 = $6d703ef3;
  var a1, b1, c1, d1, a2, b2, c2, d2: Longword;
begin
  with Work do
  begin
    //----- Initialization
    a1 := Buffer[0];
    b1 := Buffer[1];
    c1 := Buffer[2];
    d1 := Buffer[3];
    a2 := a1;
    b2 := b1;
    c2 := c1;
    d2 := d1;

    //===== 4 rounds to the "left"
    //----- L1
    Inc(a1,(b1 xor c1 xor d1) + LData[ 0]); asm rol a1,11 end;
    Inc(d1,(a1 xor b1 xor c1) + LData[ 1]); asm rol d1,14 end;
    Inc(c1,(d1 xor a1 xor b1) + LData[ 2]); asm rol c1,15 end;
    Inc(b1,(c1 xor d1 xor a1) + LData[ 3]); asm rol b1,12 end;
    Inc(a1,(b1 xor c1 xor d1) + LData[ 4]); asm rol a1, 5 end;
    Inc(d1,(a1 xor b1 xor c1) + LData[ 5]); asm rol d1, 8 end;
    Inc(c1,(d1 xor a1 xor b1) + LData[ 6]); asm rol c1, 7 end;
    Inc(b1,(c1 xor d1 xor a1) + LData[ 7]); asm rol b1, 9 end;
    Inc(a1,(b1 xor c1 xor d1) + LData[ 8]); asm rol a1,11 end;
    Inc(d1,(a1 xor b1 xor c1) + LData[ 9]); asm rol d1,13 end;
    Inc(c1,(d1 xor a1 xor b1) + LData[10]); asm rol c1,14 end;
    Inc(b1,(c1 xor d1 xor a1) + LData[11]); asm rol b1,15 end;
    Inc(a1,(b1 xor c1 xor d1) + LData[12]); asm rol a1, 6 end;
    Inc(d1,(a1 xor b1 xor c1) + LData[13]); asm rol d1, 7 end;
    Inc(c1,(d1 xor a1 xor b1) + LData[14]); asm rol c1, 9 end;
    Inc(b1,(c1 xor d1 xor a1) + LData[15]); asm rol b1, 8 end;
    //----- L2
    Inc(a1,((b1 and (c1 xor d1)) xor d1) + LData[ 7] + K2); asm rol a1, 7 end;
    Inc(d1,((a1 and (b1 xor c1)) xor c1) + LData[ 4] + K2); asm rol d1, 6 end;
    Inc(c1,((d1 and (a1 xor b1)) xor b1) + LData[13] + K2); asm rol c1, 8 end;
    Inc(b1,((c1 and (d1 xor a1)) xor a1) + LData[ 1] + K2); asm rol b1,13 end;
    Inc(a1,((b1 and (c1 xor d1)) xor d1) + LData[10] + K2); asm rol a1,11 end;
    Inc(d1,((a1 and (b1 xor c1)) xor c1) + LData[ 6] + K2); asm rol d1, 9 end;
    Inc(c1,((d1 and (a1 xor b1)) xor b1) + LData[15] + K2); asm rol c1, 7 end;
    Inc(b1,((c1 and (d1 xor a1)) xor a1) + LData[ 3] + K2); asm rol b1,15 end;
    Inc(a1,((b1 and (c1 xor d1)) xor d1) + LData[12] + K2); asm rol a1, 7 end;
    Inc(d1,((a1 and (b1 xor c1)) xor c1) + LData[ 0] + K2); asm rol d1,12 end;
    Inc(c1,((d1 and (a1 xor b1)) xor b1) + LData[ 9] + K2); asm rol c1,15 end;
    Inc(b1,((c1 and (d1 xor a1)) xor a1) + LData[ 5] + K2); asm rol b1, 9 end;
    Inc(a1,((b1 and (c1 xor d1)) xor d1) + LData[ 2] + K2); asm rol a1,11 end;
    Inc(d1,((a1 and (b1 xor c1)) xor c1) + LData[14] + K2); asm rol d1, 7 end;
    Inc(c1,((d1 and (a1 xor b1)) xor b1) + LData[11] + K2); asm rol c1,13 end;
    Inc(b1,((c1 and (d1 xor a1)) xor a1) + LData[ 8] + K2); asm rol b1,12 end;
    //----- L3
    Inc(a1,((b1 or not c1) xor d1) + LData[ 3] + K3); asm rol a1,11 end;
    Inc(d1,((a1 or not b1) xor c1) + LData[10] + K3); asm rol d1,13 end;
    Inc(c1,((d1 or not a1) xor b1) + LData[14] + K3); asm rol c1, 6 end;
    Inc(b1,((c1 or not d1) xor a1) + LData[ 4] + K3); asm rol b1, 7 end;
    Inc(a1,((b1 or not c1) xor d1) + LData[ 9] + K3); asm rol a1,14 end;
    Inc(d1,((a1 or not b1) xor c1) + LData[15] + K3); asm rol d1, 9 end;
    Inc(c1,((d1 or not a1) xor b1) + LData[ 8] + K3); asm rol c1,13 end;
    Inc(b1,((c1 or not d1) xor a1) + LData[ 1] + K3); asm rol b1,15 end;
    Inc(a1,((b1 or not c1) xor d1) + LData[ 2] + K3); asm rol a1,14 end;
    Inc(d1,((a1 or not b1) xor c1) + LData[ 7] + K3); asm rol d1, 8 end;
    Inc(c1,((d1 or not a1) xor b1) + LData[ 0] + K3); asm rol c1,13 end;
    Inc(b1,((c1 or not d1) xor a1) + LData[ 6] + K3); asm rol b1, 6 end;
    Inc(a1,((b1 or not c1) xor d1) + LData[13] + K3); asm rol a1, 5 end;
    Inc(d1,((a1 or not b1) xor c1) + LData[11] + K3); asm rol d1,12 end;
    Inc(c1,((d1 or not a1) xor b1) + LData[ 5] + K3); asm rol c1, 7 end;
    Inc(b1,((c1 or not d1) xor a1) + LData[12] + K3); asm rol b1, 5 end;
    //----- L4
    Inc(a1,(((b1 xor c1) and d1) xor c1) + LData[ 1] + K4); asm rol a1,11 end;
    Inc(d1,(((a1 xor b1) and c1) xor b1) + LData[ 9] + K4); asm rol d1,12 end;
    Inc(c1,(((d1 xor a1) and b1) xor a1) + LData[11] + K4); asm rol c1,14 end;
    Inc(b1,(((c1 xor d1) and a1) xor d1) + LData[10] + K4); asm rol b1,15 end;
    Inc(a1,(((b1 xor c1) and d1) xor c1) + LData[ 0] + K4); asm rol a1,14 end;
    Inc(d1,(((a1 xor b1) and c1) xor b1) + LData[ 8] + K4); asm rol d1,15 end;
    Inc(c1,(((d1 xor a1) and b1) xor a1) + LData[12] + K4); asm rol c1, 9 end;
    Inc(b1,(((c1 xor d1) and a1) xor d1) + LData[ 4] + K4); asm rol b1, 8 end;
    Inc(a1,(((b1 xor c1) and d1) xor c1) + LData[13] + K4); asm rol a1, 9 end;
    Inc(d1,(((a1 xor b1) and c1) xor b1) + LData[ 3] + K4); asm rol d1,14 end;
    Inc(c1,(((d1 xor a1) and b1) xor a1) + LData[ 7] + K4); asm rol c1, 5 end;
    Inc(b1,(((c1 xor d1) and a1) xor d1) + LData[15] + K4); asm rol b1, 6 end;
    Inc(a1,(((b1 xor c1) and d1) xor c1) + LData[14] + K4); asm rol a1, 8 end;
    Inc(d1,(((a1 xor b1) and c1) xor b1) + LData[ 5] + K4); asm rol d1, 6 end;
    Inc(c1,(((d1 xor a1) and b1) xor a1) + LData[ 6] + K4); asm rol c1, 5 end;
    Inc(b1,(((c1 xor d1) and a1) xor d1) + LData[ 2] + K4); asm rol b1,12 end;

    //===== 4 rounds "to the right"
    //----- R1
    Inc(a2,(((b2 xor c2) and d2) xor c2) + LData[ 5] + K5); asm rol a2, 8 end;
    Inc(d2,(((a2 xor b2) and c2) xor b2) + LData[14] + K5); asm rol d2, 9 end;
    Inc(c2,(((d2 xor a2) and b2) xor a2) + LData[ 7] + K5); asm rol c2, 9 end;
    Inc(b2,(((c2 xor d2) and a2) xor d2) + LData[ 0] + K5); asm rol b2,11 end;
    Inc(a2,(((b2 xor c2) and d2) xor c2) + LData[ 9] + K5); asm rol a2,13 end;
    Inc(d2,(((a2 xor b2) and c2) xor b2) + LData[ 2] + K5); asm rol d2,15 end;
    Inc(c2,(((d2 xor a2) and b2) xor a2) + LData[11] + K5); asm rol c2,15 end;
    Inc(b2,(((c2 xor d2) and a2) xor d2) + LData[ 4] + K5); asm rol b2, 5 end;
    Inc(a2,(((b2 xor c2) and d2) xor c2) + LData[13] + K5); asm rol a2, 7 end;
    Inc(d2,(((a2 xor b2) and c2) xor b2) + LData[ 6] + K5); asm rol d2, 7 end;
    Inc(c2,(((d2 xor a2) and b2) xor a2) + LData[15] + K5); asm rol c2, 8 end;
    Inc(b2,(((c2 xor d2) and a2) xor d2) + LData[ 8] + K5); asm rol b2,11 end;
    Inc(a2,(((b2 xor c2) and d2) xor c2) + LData[ 1] + K5); asm rol a2,14 end;
    Inc(d2,(((a2 xor b2) and c2) xor b2) + LData[10] + K5); asm rol d2,14 end;
    Inc(c2,(((d2 xor a2) and b2) xor a2) + LData[ 3] + K5); asm rol c2,12 end;
    Inc(b2,(((c2 xor d2) and a2) xor d2) + LData[12] + K5); asm rol b2, 6 end;
    //----- R2
    Inc(a2,((b2 or not c2) xor d2) + LData[ 6] + K6); asm rol a2, 9 end;
    Inc(d2,((a2 or not b2) xor c2) + LData[11] + K6); asm rol d2,13 end;
    Inc(c2,((d2 or not a2) xor b2) + LData[ 3] + K6); asm rol c2,15 end;
    Inc(b2,((c2 or not d2) xor a2) + LData[ 7] + K6); asm rol b2, 7 end;
    Inc(a2,((b2 or not c2) xor d2) + LData[ 0] + K6); asm rol a2,12 end;
    Inc(d2,((a2 or not b2) xor c2) + LData[13] + K6); asm rol d2, 8 end;
    Inc(c2,((d2 or not a2) xor b2) + LData[ 5] + K6); asm rol c2, 9 end;
    Inc(b2,((c2 or not d2) xor a2) + LData[10] + K6); asm rol b2,11 end;
    Inc(a2,((b2 or not c2) xor d2) + LData[14] + K6); asm rol a2, 7 end;
    Inc(d2,((a2 or not b2) xor c2) + LData[15] + K6); asm rol d2, 7 end;
    Inc(c2,((d2 or not a2) xor b2) + LData[ 8] + K6); asm rol c2,12 end;
    Inc(b2,((c2 or not d2) xor a2) + LData[12] + K6); asm rol b2, 7 end;
    Inc(a2,((b2 or not c2) xor d2) + LData[ 4] + K6); asm rol a2, 6 end;
    Inc(d2,((a2 or not b2) xor c2) + LData[ 9] + K6); asm rol d2,15 end;
    Inc(c2,((d2 or not a2) xor b2) + LData[ 1] + K6); asm rol c2,13 end;
    Inc(b2,((c2 or not d2) xor a2) + LData[ 2] + K6); asm rol b2,11 end;
    //----- R3
    Inc(a2,((b2 and (c2 xor d2)) xor d2) + LData[15] + K7); asm rol a2, 9 end;
    Inc(d2,((a2 and (b2 xor c2)) xor c2) + LData[ 5] + K7); asm rol d2, 7 end;
    Inc(c2,((d2 and (a2 xor b2)) xor b2) + LData[ 1] + K7); asm rol c2,15 end;
    Inc(b2,((c2 and (d2 xor a2)) xor a2) + LData[ 3] + K7); asm rol b2,11 end;
    Inc(a2,((b2 and (c2 xor d2)) xor d2) + LData[ 7] + K7); asm rol a2, 8 end;
    Inc(d2,((a2 and (b2 xor c2)) xor c2) + LData[14] + K7); asm rol d2, 6 end;
    Inc(c2,((d2 and (a2 xor b2)) xor b2) + LData[ 6] + K7); asm rol c2, 6 end;
    Inc(b2,((c2 and (d2 xor a2)) xor a2) + LData[ 9] + K7); asm rol b2,14 end;
    Inc(a2,((b2 and (c2 xor d2)) xor d2) + LData[11] + K7); asm rol a2,12 end;
    Inc(d2,((a2 and (b2 xor c2)) xor c2) + LData[ 8] + K7); asm rol d2,13 end;
    Inc(c2,((d2 and (a2 xor b2)) xor b2) + LData[12] + K7); asm rol c2, 5 end;
    Inc(b2,((c2 and (d2 xor a2)) xor a2) + LData[ 2] + K7); asm rol b2,14 end;
    Inc(a2,((b2 and (c2 xor d2)) xor d2) + LData[10] + K7); asm rol a2,13 end;
    Inc(d2,((a2 and (b2 xor c2)) xor c2) + LData[ 0] + K7); asm rol d2,13 end;
    Inc(c2,((d2 and (a2 xor b2)) xor b2) + LData[ 4] + K7); asm rol c2, 7 end;
    Inc(b2,((c2 and (d2 xor a2)) xor a2) + LData[13] + K7); asm rol b2, 5 end;
    //----- R4
    Inc(a2,(b2 xor c2 xor d2) + LData[ 8]); asm rol a2,15 end;
    Inc(d2,(a2 xor b2 xor c2) + LData[ 6]); asm rol d2, 5 end;
    Inc(c2,(d2 xor a2 xor b2) + LData[ 4]); asm rol c2, 8 end;
    Inc(b2,(c2 xor d2 xor a2) + LData[ 1]); asm rol b2,11 end;
    Inc(a2,(b2 xor c2 xor d2) + LData[ 3]); asm rol a2,14 end;
    Inc(d2,(a2 xor b2 xor c2) + LData[11]); asm rol d2,14 end;
    Inc(c2,(d2 xor a2 xor b2) + LData[15]); asm rol c2, 6 end;
    Inc(b2,(c2 xor d2 xor a2) + LData[ 0]); asm rol b2,14 end;
    Inc(a2,(b2 xor c2 xor d2) + LData[ 5]); asm rol a2, 6 end;
    Inc(d2,(a2 xor b2 xor c2) + LData[12]); asm rol d2, 9 end;
    Inc(c2,(d2 xor a2 xor b2) + LData[ 2]); asm rol c2,12 end;
    Inc(b2,(c2 xor d2 xor a2) + LData[13]); asm rol b2, 9 end;
    Inc(a2,(b2 xor c2 xor d2) + LData[ 9]); asm rol a2,12 end;
    Inc(d2,(a2 xor b2 xor c2) + LData[ 7]); asm rol d2, 5 end;
    Inc(c2,(d2 xor a2 xor b2) + LData[10]); asm rol c2,15 end;
    Inc(b2,(c2 xor d2 xor a2) + LData[14]); asm rol b2, 8 end;

    //----- Update Buffer
    Inc(d2,      Buffer[1] + c1);
    Buffer[1] := Buffer[2] + d1 + a2;
    Buffer[2] := Buffer[3] + a1 + b2;
    Buffer[3] := Buffer[0] + b1 + c2;
    Buffer[0] := d2;
  end;
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  //----- Clean up sensitive data
  // ASM Code is not really better than PASCAL code, but it avoids DELPHI
  // warning messages such that "Value assigned to ... never used".
  if HITCryptoLevel then
  asm
    xor  eax, eax
    mov  a1, eax
    mov  b1, eax
    mov  c1, eax
    mov  d1, eax
    mov  a2, eax
    mov  b2, eax
    mov  c2, eax
    mov  d2, eax
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//------------------------------------------------------------------------------
// Local
// Initialize Work (For both RIPEMD-128 and RIPEMD-160)
//------------------------------------------------------------------------------
procedure RMDInit(var Work: THWork);
begin
  //----- Set to zero
  FillChar(Work,SizeOf(Work),0);
  //----- Initialization Vector
  Move(SIV,Work.Buffer,SizeOf(Work.Buffer));
end;

//------------------------------------------------------------------------------
// Local
// Update Work depending on Buf and Size
//------------------------------------------------------------------------------
procedure RMD128Run(const Buf; Size: Longint; var Work: THWork);
  var
    BufByte : array[0..0] of Byte absolute Buf;
    i, k    : Longint;
begin
  //----- Update the bit count
  asm
    mov  eax, Size
    mov  edx, Work
    shl  eax, 3
    add  [edx].THWork.CountLo, eax
    adc  [edx].THWork.CountHi, 0
  end;
  //----- Split Buf in 64-byte (512-bit) chunks
  with Work do
  begin
    i := 0;
    while Size <> 0 do
    begin
      //----- Move the data from Buf to Data
      k := Size + RemBytes;
      if k > 64 then k := 64;
      Dec(k,RemBytes);
      Move(BufByte[i],BData[RemBytes],k);
      Inc(i,k);
      Inc(RemBytes,k);
      Dec(Size,k);
      //----- Hash
      if RemBytes = 64 then
      begin
        RMD128Mix(Work);
        RemBytes := 0;
      end;
    end;
  end;
end;

//------------------------------------------------------------------------------
// Local
// Called by RMD128Final and RMD160Final
// Swap the byte order of each long of X
//------------------------------------------------------------------------------
procedure MDByteSwap(var X: TMD160); register;
asm
     mov   ecx, 16
@00: mov   edx, [eax+ecx]
     bswap edx
     mov   [eax+ecx], edx
     sub   ecx, 4
     jnc   @00
end;

//------------------------------------------------------------------------------
// Local
// Finalize the computation of the message digest and flush Work
//------------------------------------------------------------------------------
function RMD128Final(var Work: THWork): TMD128;
begin
  with Work do
  try
    BData[RemBytes] := 128;
    Inc(RemBytes);
    if RemBytes > 56 then
    begin
      FillChar(BData[RemBytes],64-RemBytes,0);
      RMD128Mix(Work);
      RemBytes := 0;
    end;
    FillChar(BData[RemBytes],56-RemBytes,0);
    LData[14] := CountLo;
    LData[15] := CountHi;
    RMD128Mix(Work);
    //----- Message digest (reverse the byte order)
    //      The byte order is reversed so that, when converting integers
    //      to strings, the strings contain the bytes in the right order.
    //      Obviously, an integer is always displayed beginning at the
    //      left by the most significant byte.
    MDByteSwap(Buffer);
    Result[0] := Buffer[3];
    Result[1] := Buffer[2];
    Result[2] := Buffer[1];
    Result[3] := Buffer[0];
  finally
    //----- Clean up sensitive data
    FillChar(Work,SizeOf(Work),0);
  end;
end;

//==============================================================================
// Global/Overloaded
// P := RIPEMD-128 message digest of P
//==============================================================================
procedure GF2H128(P: TGF2Poly);
  var
    D  : TDynByteArray;
    W  : THWork;
    MD : TMD128;
begin
  P.GetCoefs(D);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    P.Set0;
    RMDInit(W);
    RMD128Run(D[0],Length(D),W);
    MD := RMD128Final(W);
    P.SetCoefs(MD,SizeOf(MD));
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(D[0],Length(D),0);
    end;
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// Compute a (Deg+1)-sized indexed message digest
//==============================================================================
procedure GF2H128(P: TGF2Poly; Deg: Longint; Index: Longword);
  var
    D          : TDynLongArray;
    W          : THWork;
    MD         : TMD128;
    i, j, L, S : Longint;
begin
  //----- Check
  if (Deg < -1) or (Deg > TGF2Poly.MaxDegree) then
    HITRaiseInvalidArg('GF2H128',LoadResString(@sHSH0002));

  //----- Handle special case
  if Deg = -1 then
  begin
    P.Set0;
    Exit;
  end;

  //----- Set D with the coefs of P
  P.GetCoefs(D);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    P.Set0;
    L := Length(D);
    SetLength(D,L+2);
    D[L] := 0;       //Block numbering
    D[L+1] := Index; //Hash function "index"
    //----- Size in bytes of D
    S := (L shl 2) + 8;
    j := 0;
    for i := (Deg shr 5) downto 0 do
    begin
      if j > 0 then Dec(j)
      else
      begin
        //----- Hash
        RMDInit(W);
        RMD128Run(D[0],S,W);
        MD := RMD128Final(W);
        //----- Change the block number 0,1,2,...
        Inc(D[L]);
        //----- Reset indice
        j := 3;
      end;
      //----- Set the block #i of P
      P.Block[i] := MD[j];
    end;
    //----- Keep only the degrees less than or equal to Deg
    GF2ModXk(P,Deg+1);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(D[0],SizeOf(D[0]) * Length(D),0);
    end;
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// P := RIPEMD-128 message digest of S
//==============================================================================
procedure GF2H128(P: TGF2Poly; const S: AnsiString);
  var
    W  : THWork;
    MD : TMD128;
begin
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    RMDInit(W);
    RMD128Run(S[1],Length(S),W);
    MD := RMD128Final(W);
    P.SetCoefs(MD,SizeOf(MD));
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then FillChar(MD,SizeOf(MD),0);
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// P := RIPEMD-128 message digest of Stream
//==============================================================================
procedure GF2H128(P: TGF2Poly; Stream: TStream);
  type
    TChunk = array[0..4095] of Byte;
    PChunk = ^TChunk;
  var
    Chunk : PChunk;
    W     : THWork;
    MD    : TMD128;
    Count : Longint;
begin
  GetMem(Chunk,SizeOf(TChunk));
  try
    RMDInit(W);
    repeat
      Count := Stream.Read(Chunk^,SizeOf(TChunk));
      if Count = 0 then Break;
      RMD128Run(Chunk^,Count,W);
    until false;
    MD := RMD128Final(W);
    P.SetCoefs(MD,SizeOf(MD));
  finally
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(Chunk^,SizeOf(TChunk),0);
    end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
    FreeMem(Chunk);
  end;
end;

//==============================================================================
// Global/Overloaded
// A := RIPEMD-128 message digest of A
//==============================================================================
procedure IH128(A: TInteger);
  var
    D   : TDynByteArray;
    W   : THWork;
    MD  : TMD128;
    sgn : Boolean;
begin
  A.GetDigits(D);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    sgn := A.Neg;
    A.Set0;
    RMDInit(W);
    RMD128Run(D[0],Length(D),W);
    MD := RMD128Final(W);
    A.SetDigits(MD,SizeOf(MD));
    if sgn then A.Neg := true;
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(D[0],Length(D),0);
    end;
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// Compute a BitCount-sized indexed message digest
//==============================================================================
procedure IH128(A: TInteger; BitCount: Longint; Index: Longword);
  var
    D          : TDynLongArray;
    W          : THWork;
    MD         : TMD128;
    i, j, L, S : Longint;
    sgn        : Boolean;
begin
  //----- Check
  if (BitCount < 0) or (BitCount > TInteger.MaxBitSize) then
    HITRaiseInvalidArg('IH128',LoadResString(@sHSH0001));

  //----- Handle special case
  if BitCount = 0 then
  begin
    A.Set0;
    Exit;
  end;

  //----- Set D with the "long" digits of A
  A.GetDigits(D);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    sgn := A.Neg;
    A.Set0;
    L := Length(D);
    SetLength(D,L+2);
    D[L] := 0;       //Block numbering
    D[L+1] := Index; //Hash function "index"
    //----- Size in bytes of D
    S := (L shl 2) + 8;
    j := 0;
    for i := Pred((BitCount+31) shr 5) downto 0 do
    begin
      if j > 0 then Dec(j)
      else
      begin
        //----- Hash
        RMDInit(W);
        RMD128Run(D[0],S,W);
        MD := RMD128Final(W);
        //----- Change the block number 0,1,2,...
        Inc(D[L]);
        //----- Reset indice
        j := 3;
      end;
      //----- Set the digit #i of A
      A.Long[i] := MD[j];
    end;
    //----- Keep only BitCount bits
    ICut(A,BitCount);
    if sgn then A.Neg := true;
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(D[0],SizeOf(D[0]) * Length(D),0);
    end;
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// A := RIPEMD-128 message digest of S
//==============================================================================
procedure IH128(A: TInteger; const S: AnsiString);
  var
    W  : THWork;
    MD : TMD128;
begin
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    RMDInit(W);
    RMD128Run(S[1],Length(S),W);
    MD := RMD128Final(W);
    A.SetDigits(MD,SizeOf(MD));
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then FillChar(MD,SizeOf(MD),0);
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// A := RIPEMD-128 message digest of Stream
//==============================================================================
procedure IH128(A: TInteger; Stream: TStream);
  type
    TChunk = array[0..4095] of Byte;
    PChunk = ^TChunk;
  var
    Chunk : PChunk;
    W     : THWork;
    MD    : TMD128;
    Count : Longint;
begin
  GetMem(Chunk,SizeOf(TChunk));
  try
    RMDInit(W);
    repeat
      Count := Stream.Read(Chunk^,SizeOf(TChunk));
      if Count = 0 then Break;
      RMD128Run(Chunk^,Count,W);
    until false;
    MD := RMD128Final(W);
    A.SetDigits(MD,SizeOf(MD));
  finally
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(Chunk^,SizeOf(TChunk),0);
    end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
    FreeMem(Chunk);
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// RIPEMD-160 hash function
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Local
// Compress function of RIPEMD-160
// With lookup tables and loops, this routine could be more compact but, here,
// it is the running time that is important.
//------------------------------------------------------------------------------
procedure RMD160Mix(var Work : THWork);
  const
    K2 = $5a827999;
    K3 = $6ed9eba1;
    K4 = $8f1bbcdc;
    K5 = $a953fd4e;
    K6 = $50a28be6;
    K7 = $5c4dd124;
    K8 = $6d703ef3;
    K9 = $7a6d76e9;
  var a1, b1, c1, d1, e1, a2, b2, c2, d2, e2: Longword;
begin
  with Work do
  begin
    //----- Initialization
    a1 := Buffer[0];
    b1 := Buffer[1];
    c1 := Buffer[2];
    d1 := Buffer[3];
    e1 := Buffer[4];
    a2 := a1;
    b2 := b1;
    c2 := c1;
    d2 := d1;
    e2 := e1;
    //===== 5 rounds "on the left"
    //----- Round L1
    Inc(a1,(b1 xor c1 xor d1) + LData[ 0]); asm rol a1,11 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,(a1 xor b1 xor c1) + LData[ 1]); asm rol e1,14 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,(e1 xor a1 xor b1) + LData[ 2]); asm rol d1,15 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,(d1 xor e1 xor a1) + LData[ 3]); asm rol c1,12 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,(c1 xor d1 xor e1) + LData[ 4]); asm rol b1, 5 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,(b1 xor c1 xor d1) + LData[ 5]); asm rol a1, 8 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,(a1 xor b1 xor c1) + LData[ 6]); asm rol e1, 7 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,(e1 xor a1 xor b1) + LData[ 7]); asm rol d1, 9 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,(d1 xor e1 xor a1) + LData[ 8]); asm rol c1,11 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,(c1 xor d1 xor e1) + LData[ 9]); asm rol b1,13 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,(b1 xor c1 xor d1) + LData[10]); asm rol a1,14 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,(a1 xor b1 xor c1) + LData[11]); asm rol e1,15 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,(e1 xor a1 xor b1) + LData[12]); asm rol d1, 6 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,(d1 xor e1 xor a1) + LData[13]); asm rol c1, 7 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,(c1 xor d1 xor e1) + LData[14]); asm rol b1, 9 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,(b1 xor c1 xor d1) + LData[15]); asm rol a1, 8 end;
    Inc(a1,e1); asm rol c1,10 end;
    //----- Round L2
    Inc(e1,((a1 and (b1 xor c1)) xor c1) + LData[ 7] + K2); asm rol e1, 7 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,((e1 and (a1 xor b1)) xor b1) + LData[ 4] + K2); asm rol d1, 6 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,((d1 and (e1 xor a1)) xor a1) + LData[13] + K2); asm rol c1, 8 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,((c1 and (d1 xor e1)) xor e1) + LData[ 1] + K2); asm rol b1,13 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,((b1 and (c1 xor d1)) xor d1) + LData[10] + K2); asm rol a1,11 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,((a1 and (b1 xor c1)) xor c1) + LData[ 6] + K2); asm rol e1, 9 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,((e1 and (a1 xor b1)) xor b1) + LData[15] + K2); asm rol d1, 7 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,((d1 and (e1 xor a1)) xor a1) + LData[ 3] + K2); asm rol c1,15 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,((c1 and (d1 xor e1)) xor e1) + LData[12] + K2); asm rol b1, 7 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,((b1 and (c1 xor d1)) xor d1) + LData[ 0] + K2); asm rol a1,12 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,((a1 and (b1 xor c1)) xor c1) + LData[ 9] + K2); asm rol e1,15 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,((e1 and (a1 xor b1)) xor b1) + LData[ 5] + K2); asm rol d1, 9 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,((d1 and (e1 xor a1)) xor a1) + LData[ 2] + K2); asm rol c1,11 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,((c1 and (d1 xor e1)) xor e1) + LData[14] + K2); asm rol b1, 7 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,((b1 and (c1 xor d1)) xor d1) + LData[11] + K2); asm rol a1,13 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,((a1 and (b1 xor c1)) xor c1) + LData[ 8] + K2); asm rol e1,12 end;
    Inc(e1,d1); asm rol b1,10 end;
    //----- Round L3
    Inc(d1,((e1 or not a1) xor b1) + LData[ 3] + K3); asm rol d1,11 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,((d1 or not e1) xor a1) + LData[10] + K3); asm rol c1,13 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,((c1 or not d1) xor e1) + LData[14] + K3); asm rol b1, 6 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,((b1 or not c1) xor d1) + LData[ 4] + K3); asm rol a1, 7 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,((a1 or not b1) xor c1) + LData[ 9] + K3); asm rol e1,14 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,((e1 or not a1) xor b1) + LData[15] + K3); asm rol d1, 9 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,((d1 or not e1) xor a1) + LData[ 8] + K3); asm rol c1,13 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,((c1 or not d1) xor e1) + LData[ 1] + K3); asm rol b1,15 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,((b1 or not c1) xor d1) + LData[ 2] + K3); asm rol a1,14 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,((a1 or not b1) xor c1) + LData[ 7] + K3); asm rol e1, 8 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,((e1 or not a1) xor b1) + LData[ 0] + K3); asm rol d1,13 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,((d1 or not e1) xor a1) + LData[ 6] + K3); asm rol c1, 6 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,((c1 or not d1) xor e1) + LData[13] + K3); asm rol b1, 5 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,((b1 or not c1) xor d1) + LData[11] + K3); asm rol a1,12 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,((a1 or not b1) xor c1) + LData[ 5] + K3); asm rol e1, 7 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,((e1 or not a1) xor b1) + LData[12] + K3); asm rol d1, 5 end;
    Inc(d1,c1); asm rol a1,10 end;
    //----- Round L4
    Inc(c1,(((d1 xor e1) and a1) xor e1) + LData[ 1] + K4); asm rol c1,11 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,(((c1 xor d1) and e1) xor d1) + LData[ 9] + K4); asm rol b1,12 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,(((b1 xor c1) and d1) xor c1) + LData[11] + K4); asm rol a1,14 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,(((a1 xor b1) and c1) xor b1) + LData[10] + K4); asm rol e1,15 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,(((e1 xor a1) and b1) xor a1) + LData[ 0] + K4); asm rol d1,14 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,(((d1 xor e1) and a1) xor e1) + LData[ 8] + K4); asm rol c1,15 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,(((c1 xor d1) and e1) xor d1) + LData[12] + K4); asm rol b1, 9 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,(((b1 xor c1) and d1) xor c1) + LData[ 4] + K4); asm rol a1, 8 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,(((a1 xor b1) and c1) xor b1) + LData[13] + K4); asm rol e1, 9 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,(((e1 xor a1) and b1) xor a1) + LData[ 3] + K4); asm rol d1,14 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,(((d1 xor e1) and a1) xor e1) + LData[ 7] + K4); asm rol c1, 5 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,(((c1 xor d1) and e1) xor d1) + LData[15] + K4); asm rol b1, 6 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,(((b1 xor c1) and d1) xor c1) + LData[14] + K4); asm rol a1, 8 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,(((a1 xor b1) and c1) xor b1) + LData[ 5] + K4); asm rol e1, 6 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,(((e1 xor a1) and b1) xor a1) + LData[ 6] + K4); asm rol d1, 5 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,(((d1 xor e1) and a1) xor e1) + LData[ 2] + K4); asm rol c1,12 end;
    Inc(c1,b1); asm rol e1,10 end;
    //----- Round L5
    Inc(b1,(c1 xor (d1 or not e1)) + LData[ 4] + K5); asm rol b1, 9 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,(b1 xor (c1 or not d1)) + LData[ 0] + K5); asm rol a1,15 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,(a1 xor (b1 or not c1)) + LData[ 5] + K5); asm rol e1, 5 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,(e1 xor (a1 or not b1)) + LData[ 9] + K5); asm rol d1,11 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,(d1 xor (e1 or not a1)) + LData[ 7] + K5); asm rol c1, 6 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,(c1 xor (d1 or not e1)) + LData[12] + K5); asm rol b1, 8 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,(b1 xor (c1 or not d1)) + LData[ 2] + K5); asm rol a1,13 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,(a1 xor (b1 or not c1)) + LData[10] + K5); asm rol e1,12 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,(e1 xor (a1 or not b1)) + LData[14] + K5); asm rol d1, 5 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,(d1 xor (e1 or not a1)) + LData[ 1] + K5); asm rol c1,12 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,(c1 xor (d1 or not e1)) + LData[ 3] + K5); asm rol b1,13 end;
    Inc(b1,a1); asm rol d1,10 end;
    Inc(a1,(b1 xor (c1 or not d1)) + LData[ 8] + K5); asm rol a1,14 end;
    Inc(a1,e1); asm rol c1,10 end;
    Inc(e1,(a1 xor (b1 or not c1)) + LData[11] + K5); asm rol e1,11 end;
    Inc(e1,d1); asm rol b1,10 end;
    Inc(d1,(e1 xor (a1 or not b1)) + LData[ 6] + K5); asm rol d1, 8 end;
    Inc(d1,c1); asm rol a1,10 end;
    Inc(c1,(d1 xor (e1 or not a1)) + LData[15] + K5); asm rol c1, 5 end;
    Inc(c1,b1); asm rol e1,10 end;
    Inc(b1,(c1 xor (d1 or not e1)) + LData[13] + K5); asm rol b1, 6 end;
    Inc(b1,a1); asm rol d1,10 end;
    //===== 5 rounds "on the right"
    //----- Round R1
    Inc(a2,(b2 xor (c2 or not d2)) + LData[ 5] + K6); asm rol a2, 8 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,(a2 xor (b2 or not c2)) + LData[14] + K6); asm rol e2, 9 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,(e2 xor (a2 or not b2)) + LData[ 7] + K6); asm rol d2, 9 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,(d2 xor (e2 or not a2)) + LData[ 0] + K6); asm rol c2,11 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,(c2 xor (d2 or not e2)) + LData[ 9] + K6); asm rol b2,13 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,(b2 xor (c2 or not d2)) + LData[ 2] + K6); asm rol a2,15 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,(a2 xor (b2 or not c2)) + LData[11] + K6); asm rol e2,15 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,(e2 xor (a2 or not b2)) + LData[ 4] + K6); asm rol d2, 5 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,(d2 xor (e2 or not a2)) + LData[13] + K6); asm rol c2, 7 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,(c2 xor (d2 or not e2)) + LData[ 6] + K6); asm rol b2, 7 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,(b2 xor (c2 or not d2)) + LData[15] + K6); asm rol a2, 8 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,(a2 xor (b2 or not c2)) + LData[ 8] + K6); asm rol e2,11 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,(e2 xor (a2 or not b2)) + LData[ 1] + K6); asm rol d2,14 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,(d2 xor (e2 or not a2)) + LData[10] + K6); asm rol c2,14 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,(c2 xor (d2 or not e2)) + LData[ 3] + K6); asm rol b2,12 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,(b2 xor (c2 or not d2)) + LData[12] + K6); asm rol a2, 6 end;
    Inc(a2,e2); asm rol c2,10 end;
    //----- Round R2
    Inc(e2,(((a2 xor b2) and c2) xor b2) + LData[ 6] + K7); asm rol e2, 9 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,(((e2 xor a2) and b2) xor a2) + LData[11] + K7); asm rol d2,13 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,(((d2 xor e2) and a2) xor e2) + LData[ 3] + K7); asm rol c2,15 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,(((c2 xor d2) and e2) xor d2) + LData[ 7] + K7); asm rol b2, 7 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,(((b2 xor c2) and d2) xor c2) + LData[ 0] + K7); asm rol a2,12 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,(((a2 xor b2) and c2) xor b2) + LData[13] + K7); asm rol e2, 8 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,(((e2 xor a2) and b2) xor a2) + LData[ 5] + K7); asm rol d2, 9 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,(((d2 xor e2) and a2) xor e2) + LData[10] + K7); asm rol c2,11 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,(((c2 xor d2) and e2) xor d2) + LData[14] + K7); asm rol b2, 7 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,(((b2 xor c2) and d2) xor c2) + LData[15] + K7); asm rol a2, 7 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,(((a2 xor b2) and c2) xor b2) + LData[ 8] + K7); asm rol e2,12 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,(((e2 xor a2) and b2) xor a2) + LData[12] + K7); asm rol d2, 7 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,(((d2 xor e2) and a2) xor e2) + LData[ 4] + K7); asm rol c2, 6 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,(((c2 xor d2) and e2) xor d2) + LData[ 9] + K7); asm rol b2,15 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,(((b2 xor c2) and d2) xor c2) + LData[ 1] + K7); asm rol a2,13 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,(((a2 xor b2) and c2) xor b2) + LData[ 2] + K7); asm rol e2,11 end;
    Inc(e2,d2); asm rol b2,10 end;
    //----- Round R3
    Inc(d2,((e2 or not a2) xor b2) + LData[15] + K8); asm rol d2, 9 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,((d2 or not e2) xor a2) + LData[ 5] + K8); asm rol c2, 7 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,((c2 or not d2) xor e2) + LData[ 1] + K8); asm rol b2,15 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,((b2 or not c2) xor d2) + LData[ 3] + K8); asm rol a2,11 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,((a2 or not b2) xor c2) + LData[ 7] + K8); asm rol e2, 8 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,((e2 or not a2) xor b2) + LData[14] + K8); asm rol d2, 6 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,((d2 or not e2) xor a2) + LData[ 6] + K8); asm rol c2, 6 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,((c2 or not d2) xor e2) + LData[ 9] + K8); asm rol b2,14 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,((b2 or not c2) xor d2) + LData[11] + K8); asm rol a2,12 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,((a2 or not b2) xor c2) + LData[ 8] + K8); asm rol e2,13 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,((e2 or not a2) xor b2) + LData[12] + K8); asm rol d2, 5 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,((d2 or not e2) xor a2) + LData[ 2] + K8); asm rol c2,14 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,((c2 or not d2) xor e2) + LData[10] + K8); asm rol b2,13 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,((b2 or not c2) xor d2) + LData[ 0] + K8); asm rol a2,13 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,((a2 or not b2) xor c2) + LData[ 4] + K8); asm rol e2, 7 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,((e2 or not a2) xor b2) + LData[13] + K8); asm rol d2, 5 end;
    Inc(d2,c2); asm rol a2,10 end;
    //----- Round R4
    Inc(c2,((d2 and (e2 xor a2)) xor a2) + LData[ 8] + K9); asm rol c2,15 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,((c2 and (d2 xor e2)) xor e2) + LData[ 6] + K9); asm rol b2, 5 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,((b2 and (c2 xor d2)) xor d2) + LData[ 4] + K9); asm rol a2, 8 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,((a2 and (b2 xor c2)) xor c2) + LData[ 1] + K9); asm rol e2,11 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,((e2 and (a2 xor b2)) xor b2) + LData[ 3] + K9); asm rol d2,14 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,((d2 and (e2 xor a2)) xor a2) + LData[11] + K9); asm rol c2,14 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,((c2 and (d2 xor e2)) xor e2) + LData[15] + K9); asm rol b2, 6 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,((b2 and (c2 xor d2)) xor d2) + LData[ 0] + K9); asm rol a2,14 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,((a2 and (b2 xor c2)) xor c2) + LData[ 5] + K9); asm rol e2, 6 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,((e2 and (a2 xor b2)) xor b2) + LData[12] + K9); asm rol d2, 9 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,((d2 and (e2 xor a2)) xor a2) + LData[ 2] + K9); asm rol c2,12 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,((c2 and (d2 xor e2)) xor e2) + LData[13] + K9); asm rol b2, 9 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,((b2 and (c2 xor d2)) xor d2) + LData[ 9] + K9); asm rol a2,12 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,((a2 and (b2 xor c2)) xor c2) + LData[ 7] + K9); asm rol e2, 5 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,((e2 and (a2 xor b2)) xor b2) + LData[10] + K9); asm rol d2,15 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,((d2 and (e2 xor a2)) xor a2) + LData[14] + K9); asm rol c2, 8 end;
    Inc(c2,b2); asm rol e2,10 end;
    //----- Round R5
    Inc(b2,(c2 xor d2 xor e2) + LData[12]); asm rol b2, 8 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,(b2 xor c2 xor d2) + LData[15]); asm rol a2, 5 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,(a2 xor b2 xor c2) + LData[10]); asm rol e2,12 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,(e2 xor a2 xor b2) + LData[ 4]); asm rol d2, 9 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,(d2 xor e2 xor a2) + LData[ 1]); asm rol c2,12 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,(c2 xor d2 xor e2) + LData[ 5]); asm rol b2, 5 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,(b2 xor c2 xor d2) + LData[ 8]); asm rol a2,14 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,(a2 xor b2 xor c2) + LData[ 7]); asm rol e2, 6 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,(e2 xor a2 xor b2) + LData[ 6]); asm rol d2, 8 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,(d2 xor e2 xor a2) + LData[ 2]); asm rol c2,13 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,(c2 xor d2 xor e2) + LData[13]); asm rol b2, 6 end;
    Inc(b2,a2); asm rol d2,10 end;
    Inc(a2,(b2 xor c2 xor d2) + LData[14]); asm rol a2, 5 end;
    Inc(a2,e2); asm rol c2,10 end;
    Inc(e2,(a2 xor b2 xor c2) + LData[ 0]); asm rol e2,15 end;
    Inc(e2,d2); asm rol b2,10 end;
    Inc(d2,(e2 xor a2 xor b2) + LData[ 3]); asm rol d2,13 end;
    Inc(d2,c2); asm rol a2,10 end;
    Inc(c2,(d2 xor e2 xor a2) + LData[ 9]); asm rol c2,11 end;
    Inc(c2,b2); asm rol e2,10 end;
    Inc(b2,(c2 xor d2 xor e2) + LData[11]); asm rol b2,11 end;
    Inc(b2,a2); asm rol d2,10 end;
    //----- Update the message digest
    Inc(d2,      Buffer[1] + c1);
    Buffer[1] := Buffer[2] + d1 + e2;
    Buffer[2] := Buffer[3] + e1 + a2;
    Buffer[3] := Buffer[4] + a1 + b2;
    Buffer[4] := Buffer[0] + b1 + c2;
    Buffer[0] := d2;
  end;
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  //----- Clean up sensitive data
  // ASM Code is not really better than PASCAL code, but it avoids DELPHI
  // warning messages such that "Value assigned to ... never used".
  if HITCryptoLevel then
  asm
    xor  eax, eax
    mov  a1, eax
    mov  b1, eax
    mov  c1, eax
    mov  d1, eax
    mov  e1, eax
    mov  a2, eax
    mov  b2, eax
    mov  c2, eax
    mov  d2, eax
    mov  e2, eax
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//------------------------------------------------------------------------------
// Local
// Update Work depending on Buf and Size
//------------------------------------------------------------------------------
procedure RMD160Run(const Buf; Size: Longint; var Work: THWork);
  var
    BufByte : array[0..0] of Byte absolute Buf;
    i, k    : Longint;
begin
  //----- Update the bit count
  asm
    mov  eax, Size
    mov  edx, Work
    shl  eax, 3
    add  [edx].THWork.CountLo, eax
    adc  [edx].THWork.CountHi, 0
  end;
  //----- Split Buf in 64-byte (512-bit) chunks
  with Work do
  begin
    i := 0;
    while Size <> 0 do
    begin
      //----- Move the data from Buf to Data
      k := Size + RemBytes;
      if k > 64 then k := 64;
      Dec(k,RemBytes);
      Move(BufByte[i],BData[RemBytes],k);
      Inc(i,k);
      Inc(RemBytes,k);
      Dec(Size,k);
      //----- Hash
      if RemBytes = 64 then
      begin
        RMD160Mix(Work);
        RemBytes := 0;
      end;
    end;
  end;
end;

//------------------------------------------------------------------------------
// Local
// Finalize the computation of the message digest and flush Work
//------------------------------------------------------------------------------
function RMD160Final(var Work: THWork): TMD160;
begin
  with Work do
  try
    BData[RemBytes] := 128;
    Inc(RemBytes);
    if RemBytes > 56 then
    begin
      FillChar(BData[RemBytes],64-RemBytes,0);
      RMD160Mix(Work);
      RemBytes := 0;
    end;
    FillChar(BData[RemBytes],56-RemBytes,0);
    LData[14] := CountLo;
    LData[15] := CountHi;
    RMD160Mix(Work);
    //----- Message digest (reverse the byte order)
    //      The byte order is reversed so that, when converting integers
    //      to strings, the strings contain the bytes in the right order.
    //      Obviously, an integer is always displayed beginning at the
    //      left by the most significant byte.
    MDByteSwap(Buffer);
    Result[0] := Buffer[4];
    Result[1] := Buffer[3];
    Result[2] := Buffer[2];
    Result[3] := Buffer[1];
    Result[4] := Buffer[0];
  finally
    //----- Clean up sensitive data
    FillChar(Work,SizeOf(Work),0);
  end;
end;

//==============================================================================
// Global/Overloaded
// P := RIPEMD-160 message digest of P
//==============================================================================
procedure GF2H160(P: TGF2Poly);
  var
    D  : TDynByteArray;
    W  : THWork;
    MD : TMD160;
begin
  P.GetCoefs(D);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    P.Set0;
    RMDInit(W);
    RMD160Run(D[0],Length(D),W);
    MD := RMD160Final(W);
    P.SetCoefs(MD,SizeOf(MD));
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(D[0],Length(D),0);
    end;
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// Compute a (Deg+1)-sized indexed message digest
//==============================================================================
procedure GF2H160(P: TGF2Poly; Deg: Longint; Index: Longword);
  var
    D          : TDynLongArray;
    W          : THWork;
    MD         : TMD160;
    i, j, L, S : Longint;
begin
  //----- Check
  if (Deg < -1) or (Deg > TGF2Poly.MaxDegree) then
    HITRaiseInvalidArg('GF2H160',LoadResString(@sHSH0002));

  //----- Handle special case
  if Deg = -1 then
  begin
    P.Set0;
    Exit;
  end;

  //----- Set D with the coefs of P
  P.GetCoefs(D);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    P.Set0;
    L := Length(D);
    SetLength(D,L+2);
    D[L] := 0;       //Block numbering
    D[L+1] := Index; //Hash function "index"
    //----- Size in bytes of D
    S := (L shl 2) + 8;
    j := 0;
    for i := (Deg shr 5) downto 0 do
    begin
      if j > 0 then Dec(j)
      else
      begin
        //----- Hash
        RMDInit(W);
        RMD160Run(D[0],S,W);
        MD := RMD160Final(W);
        //----- Change the block number 0,1,2,...
        Inc(D[L]);
        //----- Reset indice
        j := 4;
      end;
      //----- Set the block #i of P
      P.Block[i] := MD[j];
    end;
    //----- Keep only the degrees less than or equal to Deg
    GF2ModXk(P,Deg+1);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(D[0],SizeOf(D[0]) * Length(D),0);
    end;
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// P := RIPEMD-160 message digest of S
//==============================================================================
procedure GF2H160(P: TGF2Poly; const S: AnsiString);
  var
    W  : THWork;
    MD : TMD160;
begin
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    RMDInit(W);
    RMD160Run(S[1],Length(S),W);
    MD := RMD160Final(W);
    P.SetCoefs(MD,SizeOf(MD));
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then FillChar(MD,SizeOf(MD),0);
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// P := RIPEMD-160 message digest of Stream
//==============================================================================
procedure GF2H160(P: TGF2Poly; Stream: TStream);
  type
    TChunk = array[0..4095] of Byte;
    PChunk = ^TChunk;
  var
    Chunk : PChunk;
    W     : THWork;
    MD    : TMD160;
    Count : Longint;
begin
  GetMem(Chunk,SizeOf(TChunk));
  try
    RMDInit(W);
    repeat
      Count := Stream.Read(Chunk^,SizeOf(TChunk));
      if Count = 0 then Break;
      RMD160Run(Chunk^,Count,W);
    until false;
    MD := RMD160Final(W);
    P.SetCoefs(MD,SizeOf(MD));
  finally
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(Chunk^,SizeOf(TChunk),0);
    end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
    FreeMem(Chunk);
  end;
end;

//==============================================================================
// Global/Overloaded
// A := RIPEMD-160 message digest of A
//==============================================================================
procedure IH160(A: TInteger);
  var
    D   : TDynByteArray;
    W   : THWork;
    MD  : TMD160;
    sgn : Boolean;
begin
  A.GetDigits(D);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    sgn := A.Neg;
    A.Set0;
    RMDInit(W);
    RMD160Run(D[0],Length(D),W);
    MD := RMD160Final(W);
    A.SetDigits(MD,SizeOf(MD));
    if sgn then A.Neg := true;
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(D[0],Length(D),0);
    end;
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// Compute a BitCount-sized indexed message digest
//==============================================================================
procedure IH160(A: TInteger; BitCount: Longint; Index: Longword);
  var
    D          : TDynLongArray;
    W          : THWork;
    MD         : TMD160;
    i, j, L, S : Longint;
    sgn        : Boolean;
begin
  //----- Check
  if (BitCount < 0) or (BitCount > TInteger.MaxBitSize) then
    HITRaiseInvalidArg('IH160',LoadResString(@sHSH0001));

  //----- Handle special case
  if BitCount = 0 then
  begin
    A.Set0;
    Exit;
  end;

  //----- Set D with the "long" digits of A
  A.GetDigits(D);
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    sgn := A.Neg;
    A.Set0;
    L := Length(D);
    SetLength(D,L+2);
    D[L] := 0;       //Block numbering
    D[L+1] := Index; //Hash function "index"
    //----- Size in bytes of D
    S := (L shl 2) + 8;
    j := 0;
    for i := Pred((BitCount+31) shr 5) downto 0 do
    begin
      if j > 0 then Dec(j)
      else
      begin
        //----- Hash
        RMDInit(W);
        RMD160Run(D[0],S,W);
        MD := RMD160Final(W);
        //----- Change the block number 0,1,2,...
        Inc(D[L]);
        //----- Reset indice
        j := 4;
      end;
      //----- Set the digit #i of A
      A.Long[i] := MD[j];
    end;
    //----- Keep only BitCount bits
    ICut(A,BitCount);
    if sgn then A.Neg := true;
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(D[0],SizeOf(D[0]) * Length(D),0);
    end;
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// A := RIPEMD-160 message digest of S
//==============================================================================
procedure IH160(A: TInteger; const S: AnsiString);
  var
    W  : THWork;
    MD : TMD160;
begin
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  try
{$ENDIF} /////////////////////////////////////////////////////////////////////
    RMDInit(W);
    RMD160Run(S[1],Length(S),W);
    MD := RMD160Final(W);
    A.SetDigits(MD,SizeOf(MD));
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
  finally
    if HITCryptoLevel then FillChar(MD,SizeOf(MD),0);
  end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
end;

//==============================================================================
// Global/Overloaded
// A := RIPEMD-160 message digest of Stream
//==============================================================================
procedure IH160(A: TInteger; Stream: TStream);
  type
    TChunk = array[0..4095] of Byte;
    PChunk = ^TChunk;
  var
    Chunk : PChunk;
    W     : THWork;
    MD    : TMD160;
    Count : Longint;
begin
  GetMem(Chunk,SizeOf(TChunk));
  try
    RMDInit(W);
    repeat
      Count := Stream.Read(Chunk^,SizeOf(TChunk));
      if Count = 0 then Break;
      RMD160Run(Chunk^,Count,W);
    until false;
    MD := RMD160Final(W);
    A.SetDigits(MD,SizeOf(MD));
  finally
{$IFDEF CRYPTO_MEMORY_MANAGEMENT} ////////////////////////////////////////////
    if HITCryptoLevel then
    begin
      FillChar(MD,SizeOf(MD),0);
      FillChar(Chunk^,SizeOf(TChunk),0);
    end;
{$ENDIF} /////////////////////////////////////////////////////////////////////
    FreeMem(Chunk);
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Misc.
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Local
// Reverse the byte order of X from LSBF mode (on PC) to MSBF mode
// "LSBF" stands for "Least Significant Byte First"
// "MSBF" stands for "Most Significant Byte First"
//------------------------------------------------------------------------------
function ByteSwap(X: Longword): Longword; register;
asm
  bswap eax
end;

//==============================================================================
// Global
// Swaps the byte order from LSBF mode (on PC) to MSBF mode
//==============================================================================
procedure HITByteSwap(P: TGF2Poly);
  var i : Longint;
begin
  with P do for i := 0 to BlockCount-1 do Block[i] := ByteSwap(Block[i]);
end;

//==============================================================================
// Global
// Swaps the byte order from LSBF mode (on PC) to MSBF mode
//==============================================================================
procedure HITByteSwap(A: TInteger);
  var i : Longint;
begin
  with A do for i := 0 to LongSize-1 do Long[i] := ByteSwap(Long[i]);
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

