////////////////////////////////////////////////////////////////////////////////
// HIT Computation Threads
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit HIT_Threads;

{.$DEFINE DEBUG} //Debugging code

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Classes;

//------------------------------------------------------------------------------
// The following class should be the ancestor of all HIT computation threads
//------------------------------------------------------------------------------

type
  TAbstractHITThread = class(TThread)
  private
    fExceptMsg: string;
    fExceptName: string;
    fExceptRaised: Boolean;
  protected
    procedure Computation; virtual; abstract;
    procedure Execute; override;
    procedure LocalVarsFree;
  public
    property ExceptMsg: string read fExceptMsg;
    property ExceptName: string read fExceptName;
    property ExceptRaised: Boolean read fExceptRaised;
  end;

implementation
//==============================================================================
{$C-,G-,O+,Q-,R-}
//==============================================================================
uses SysUtils, Integers, GF2;

////////////////////////////////////////////////////////////////////////////////
// THITThread class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
//==============================================================================
procedure TAbstractHITThread.Execute;
begin
  try
    try

      // Each thread has its own random generator. If a thread makes use of
      // its random generator, this one must imperatively be initialized.

      HITTimeSeed;

      Computation;
    finally

      // Each thread has its own local variable banks in INTEGERS and in GF2.
      // These banks must be freed before the thread is destroyed, otherwise
      // they will stay in memory until the application terminates or until
      // the IUnitVarsFreeAll and/or the GF2UnitVarsFreeAll are called.

      LocalVarsFree;
    end;
  except

    // Need to catch exceptions (otherwise DELPHI crashes).

    if not (ExceptObject is EAbort) then
    begin
      fExceptRaised := true;
      fExceptName := ExceptObject.ClassName;
      if ExceptObject is Exception then
        fExceptMsg := Exception(ExceptObject).Message;
    end;
  end;
end;

//==============================================================================
// Free the local variables of the current thread (used by Integers and GF2)
// Each thread has its own local variables
//==============================================================================
procedure TAbstractHITThread.LocalVarsFree;
begin
  GF2UnitVarsFree;
  IUnitVarsFree;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

