////////////////////////////////////////////////////////////////////////////////
// Chinese Remainder Theorem (for Integers and GF2)
// Unit of the HIT package for DELPHI 4/5
// Version 1.01
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit ChineseRem;

{.$DEFINE DEBUG} //Debugging code

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers, GF2, HIT_Lists;

procedure GF2CRT(P, R1, R2, M1, M2: TGF2Poly); overload;
procedure GF2CRT(P: TGF2Poly; const R, M: array of TGF2Poly); overload;
procedure GF2CRT(P: TGF2Poly; R, M: TGF2PolyVector); overload;
procedure GF2CRT(P: TGF2Poly; R, M: TGF2List); overload;

procedure ICRT(X, R1, R2, M1, M2: TInteger); overload;
procedure ICRT(X: TInteger; const R, M: array of TInteger); overload;
procedure ICRT(X: TInteger; R, M: TIntegerVector); overload;
procedure ICRT(X: TInteger; R, M: TIntList); overload;

implementation
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// Resource strings
////////////////////////////////////////////////////////////////////////////////

resourcestring
  sCRT0010 = 'The R and M arrays should have the same size';
  sCRT0011 = 'The R and M arrays should not be empty';
  sCRT0012 = 'The remainders R should be nonnegative';
  sCRT0013 = 'The moduli M should be greater than 1';
  sCRT0014 = 'The moduli M should be pairwise relatively prime';
  sCRT0015 = 'The moduli M should be "greater" than 1';
  sCRT0016 = 'The moduli M should have pairwise a GCD equal to 1';
  sCRT0110 = 'The R and M lists should have the same size';
  sCRT0111 = 'The R and M lists should not be empty';
  sCRT0210 = 'The R and M vectors should have the same size';
  sCRT0211 = 'The R and M vectors should not be empty';

////////////////////////////////////////////////////////////////////////////////
// Polynomials over GF(2)
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Local
// Chinese Remainder Theorem. Garner's algorithm.
// -> R, remainders R[i]
// -> M, moduli M[i]
// <- P, solution of the equations system P = R[i] mod M[i]
//------------------------------------------------------------------------------
procedure GF2ChineseRem(P: TGF2Poly; const R, M: array of TGF2Poly);
  var
    T, U, Z : TGF2Poly;
    i       : Integer;
begin
  //===== Check
  //----- R and M should have the same size
  if High(R) <> High(M) then
    HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0010));
  //----- R and M should have at least one element
  if Length(R) = 0 then HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0011));
  //----- @P must be different from @R[i] and @M[i] (@P may be @R[0])
  if P = M[0] then HITRaiseDuplicatePtr('GF2CRT','P,R[],M[]');
  for i := 1 to High(R) do
  begin
    if (P = R[i]) or (P = M[i]) then HITRaiseDuplicatePtr('GF2CRT','P,R[],M[]');
  end;
  //----- M[i] should be "greater" than 1
  for i := 0 to High(R) do
    if M[i].Degree < 1 then
      HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0015));

  T := nil;
  U := nil;
  Z := nil;
  try
    T := TGF2Poly.Create(1); //Create T equal to 1
    U := TGF2Poly.Create;
    Z := TGF2Poly.Create;
    //----- P := R[0] mod M[0]
    GF2Set(P,R[0]);
    GF2Mod(P,M[0]);
    for i := 1 to High(R) do
    begin
      //----- T := T * M[i-1]
      GF2Mul(T,M[i-1]);
      //----- U := T^(-1) mod M[i]
      //      Test whether the moduli are pairwise relatively prime
      //      Failed inversion <-> GCD(U,M[i]) <> 1
      GF2Set(U,T);
      if not GF2PowMod(U,-1,M[i]) then
      begin
        P.Set0;
        HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0015));
      end;
      //----- Z := (R[i] + P) mod M[i]
      GF2Set(Z,R[i]);
      GF2Add(Z,P);
      GF2Mod(Z,M[i]);
      //----- P := P + ((U * Z) mod M[i]) * T
      GF2Mul(Z,U);
      GF2Mod(Z,M[i]);
      GF2Mul(Z,T);
      GF2Add(P,Z);
    end;
  finally
    T.Free;
    U.Free;
    Z.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// -> R1, R2, remainders
// -> M1, M2, moduli
// <- P, solution of the equations system P = R[i] mod M[i]
//==============================================================================
procedure GF2CRT(P, R1, R2, M1, M2: TGF2Poly);
begin
  GF2ChineseRem(P,[R1,R2],[M1,M2]);
end;

//==============================================================================
// Global/Overloaded
// -> R, remainders R[i]
// -> M, moduli M[i]
// <- P, solution of the equations system P = R[i] mod M[i]
//==============================================================================
procedure GF2CRT(P: TGF2Poly; const R, M: array of TGF2Poly);
begin
  GF2ChineseRem(P,R,M);
end;

//==============================================================================
// Global/Overloaded
// -> R, remainders R[i]
// -> M, moduli M[i]
// <- P, solution of the equations system P = R[i] mod M[i]
//==============================================================================
procedure GF2CRT(P: TGF2Poly; R, M: TGF2PolyVector);
  var
    T, U, Z : TGF2Poly;
    i       : Integer;
begin
  //===== Check
  //----- R and M should have the same size
  if R.Size <> M.Size then
    HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0210));
  //----- R and M should have at least one element
  if R.Size = 0 then HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0211));
  //----- @R must be different from @M
  if R = M then HITRaiseDuplicatePtr('GF2CRT','R,M');
  //----- M[i] should be "greater" than 1
  for i := 0 to R.Size-1 do
    if M[i].Degree < 1 then
      HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0015));

  T := nil;
  U := nil;
  Z := nil;
  try
    T := TGF2Poly.Create(1); //Create T equal to 1
    U := TGF2Poly.Create;
    Z := TGF2Poly.Create;
    //----- P := R[0] mod M[0]
    GF2Set(P,R[0]);
    GF2Mod(P,M[0]);
    for i := 1 to R.Size-1 do
    begin
      //----- T := T * M[i-1]
      GF2Mul(T,M[i-1]);
      //----- U := T^(-1) mod M[i]
      //      Test whether the moduli are pairwise relatively prime
      //      Failed inversion <-> GCD(U,M[i]) <> 1
      GF2Set(U,T);
      if not GF2PowMod(U,-1,M[i]) then
      begin
        P.Set0;
        HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0015));
      end;
      //----- Z := (R[i] + P) mod M[i]
      GF2Set(Z,R[i]);
      GF2Add(Z,P);
      GF2Mod(Z,M[i]);
      //----- P := P + ((U * Z) mod M[i]) * T
      GF2Mul(Z,U);
      GF2Mod(Z,M[i]);
      GF2Mul(Z,T);
      GF2Add(P,Z);
    end;
  finally
    T.Free;
    U.Free;
    Z.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// Chinese Remainder Theorem
// -> R, remainders R[i]
// -> M, moduli M[i]
// <- P, solution of the equations system P = R[i] mod M[i]
//==============================================================================
procedure GF2CRT(P: TGF2Poly; R, M: TGF2List);
  var
    T, U, Z : TGF2Poly;
    i       : Integer;
begin
  //===== Check
  //----- R and M should have the same size
  if R.Count <> M.Count then
    HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0110));
  //----- R and M should have at least one element
  if R.Count = 0 then HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0111));
  //----- @R must be different from @M
  if R = M then HITRaiseDuplicatePtr('GF2CRT','R,M');
  //----- M[i] should be "greater" than 1
  for i := 0 to R.Count-1 do
    if M[i].Degree < 1 then
      HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0015));

  T := nil;
  U := nil;
  Z := nil;
  try
    T := TGF2Poly.Create(1); //Create T equal to 1
    U := TGF2Poly.Create;
    Z := TGF2Poly.Create;
    //----- P := R[0] mod M[0]
    GF2Set(P,R[0]);
    GF2Mod(P,M[0]);
    for i := 1 to R.Count-1 do
    begin
      //----- T := T * M[i-1]
      GF2Mul(T,M[i-1]);
      //----- U := T^(-1) mod M[i]
      //      Test whether the moduli are pairwise relatively prime
      //      Failed inversion <-> GCD(U,M[i]) <> 1
      GF2Set(U,T);
      if not GF2PowMod(U,-1,M[i]) then
      begin
        P.Set0;
        HITRaiseInvalidArg('GF2CRT',LoadResString(@sCRT0015));
      end;
      //----- Z := (R[i] + P) mod M[i]
      GF2Set(Z,R[i]);
      GF2Add(Z,P);
      GF2Mod(Z,M[i]);
      //----- P := P + ((U * Z) mod M[i]) * T
      GF2Mul(Z,U);
      GF2Mod(Z,M[i]);
      GF2Mul(Z,T);
      GF2Add(P,Z);
    end;
  finally
    T.Free;
    U.Free;
    Z.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Integers
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Local
// Chinese Remainder Theorem. Garner's algorithm.
// -> R, remainders R[i]
// -> M, moduli M[i]
// <- X, solution of the equations system X = R[i] mod M[i]
//------------------------------------------------------------------------------
procedure IChineseRem(X: TInteger; const R, M: array of TInteger);
  var
    T, U, Z : TInteger;
    i       : Integer;
begin
  //===== Check
  //----- R and M should have the same size
  if High(R) <> High(M) then
    HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0010));
  //----- R and M should have at least one element
  if Length(R) = 0 then HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0011));
  //----- @X must be different from @R[i] and @M[i] (@X may be @R[0])
  if X = M[0] then HITRaiseDuplicatePtr('ICRT','X,R[],M[]');
  for i := 1 to High(R) do
  begin
    if (X = R[i]) or (X = M[i]) then HITRaiseDuplicatePtr('ICRT','X,R[],M[]');
  end;
  //----- R[i] should be positive and M[i] should be greater than 1
  for i := 0 to High(R) do
  begin
    if R[i].Neg then HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0012));
    if ICmp(M[i],2) < 0 then
      HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0013));
  end;

  T := nil;
  U := nil;
  Z := nil;
  try
    T := TInteger.Create(1);
    U := TInteger.Create;
    Z := TInteger.Create;
    //----- X := R[0] mod M[0]
    ISet(X,R[0]);
    IMod(X,M[0]);
    for i := 1 to High(R) do
    begin
      //----- T := T * M[i-1]
      IMul(T,M[i-1]);
      //----- U := T^(-1) mod M[i]
      //      Test whether the moduli are pairwise relatively prime
      //      Failed inversion <-> GCD(U,M[i]) <> 1
      ISet(U,T);
      if not IInvMod(U,M[i]) then
      begin
        X.Set0;
        HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0014));
      end;
      //----- Z := (R[i] - X) mod M[i]
      ISet(Z,R[i]);
      ISub(Z,X);
      IMod(Z,M[i]);
      //----- X := X + ((U * Z) mod M[i]) * T
      IMul(Z,U);
      IMod(Z,M[i]);
      IMul(Z,T);
      IAdd(X,Z);
    end;
  finally
    T.Free;
    U.Free;
    Z.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// -> R1, R2, remainders
// -> M1, M2, moduli
// <- X, solution of the equations system X = R[i] mod M[i]
//==============================================================================
procedure ICRT(X, R1, R2, M1, M2: TInteger);
begin
  IChineseRem(X,[R1,R2],[M1,M2]);
end;

//==============================================================================
// Global/Overloaded
// -> R, remainders R[i]
// -> M, moduli M[i]
// <- X, solution of the equations system X = R[i] mod M[i]
//==============================================================================
procedure ICRT(X: TInteger; const R, M: array of TInteger);
begin
  IChineseRem(X,R,M);
end;

//==============================================================================
// Global/Overloaded
// -> R, remainders R[i]
// -> M, moduli M[i]
// <- X, solution of the equations system X = R[i] mod M[i]
//==============================================================================
procedure ICRT(X: TInteger; R, M: TIntegerVector);
  var
    T, U, Z : TInteger;
    i       : Integer;
begin
  //===== Check
  //----- R and M should have the same size
  if R.Size <> M.Size then
    HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0210));
  //----- R and M should have at least one element
  if R.Size = 0 then HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0211));
  //----- @R should be different from @M
  if R = M then HITRaiseDuplicatePtr('ICRT','R,M');
  //----- R[i] should be positive and M[i] should be greater than 1
  for i := 0 to R.Size-1 do
  begin
    if R[i].Neg then HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0012));
    if ICmp(M[i],2) < 0 then
      HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0013));
  end;

  T := nil;
  U := nil;
  Z := nil;
  try
    T := TInteger.Create(1);
    U := TInteger.Create;
    Z := TInteger.Create;
    //----- X := R[0] mod M[0]
    ISet(X,R[0]);
    IMod(X,M[0]);
    for i := 1 to R.Size-1 do
    begin
      //----- T := T * M[i-1]
      IMul(T,M[i-1]);
      //----- U := T^(-1) mod M[i]
      //      Test whether the moduli are pairwise relatively prime
      //      Failed inversion <-> GCD(U,M[i]) <> 1
      ISet(U,T);
      if not IInvMod(U,M[i]) then
      begin
        X.Set0;
        HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0014));
      end;
      //----- Z := (R[i] - X) mod M[i]
      ISet(Z,R[i]);
      ISub(Z,X);
      IMod(Z,M[i]);
      //----- X := X + ((U * Z) mod M[i]) * T
      IMul(Z,U);
      IMod(Z,M[i]);
      IMul(Z,T);
      IAdd(X,Z);
    end;
  finally
    T.Free;
    U.Free;
    Z.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// Chinese Remainder Theorem
// -> R, remainders R[i]
// -> M, moduli M[i]
// <- X, solution of the equations system X = R[i] mod M[i]
//==============================================================================
procedure ICRT(X: TInteger; R, M: TIntList);
  var
    T, U, Z : TInteger;
    i       : Integer;
begin
  //===== Check
  //----- R and M should have the same size
  if R.Count <> M.Count then
    HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0110));
  //----- R and M should have at least one element
  if R.Count = 0 then HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0111));
  //----- @R should be different from @M
  if R = M then HITRaiseDuplicatePtr('ICRT','R,M');
  //----- R[i] should be positive and M[i] should be greater than 1
  for i := 0 to R.Count-1 do
  begin
    if R[i].Neg then HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0012));
    if ICmp(M[i],2) < 0 then
      HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0013));
  end;

  T := nil;
  U := nil;
  Z := nil;
  try
    T := TInteger.Create(1);
    U := TInteger.Create;
    Z := TInteger.Create;
    //----- X := R[0] mod M[0]
    ISet(X,R[0]);
    IMod(X,M[0]);
    for i := 1 to R.Count-1 do
    begin
      //----- T := T * M[i-1]
      IMul(T,M[i-1]);
      //----- U := T^(-1) mod M[i]
      //      Test whether the moduli are pairwise relatively prime
      //      Failed inversion <-> GCD(U,M[i]) <> 1
      ISet(U,T);
      if not IInvMod(U,M[i]) then
      begin
        X.Set0;
        HITRaiseInvalidArg('ICRT',LoadResString(@sCRT0014));
      end;
      //----- Z := (R[i] - X) mod M[i]
      ISet(Z,R[i]);
      ISub(Z,X);
      IMod(Z,M[i]);
      //----- X := X + ((U * Z) mod M[i]) * T
      IMul(Z,U);
      IMod(Z,M[i]);
      IMul(Z,T);
      IAdd(X,Z);
    end;
  finally
    T.Free;
    U.Free;
    Z.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

