////////////////////////////////////////////////////////////////////////////////
// Elliptic Curves over GF(2^(DE)) with Subfield GF(2^D)
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit GF2N_ECSubfields;

{.$DEFINE DEBUG} //Debugging code

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G-,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers, Integers_Primes, GF2;

//==============================================================================
// This unit allows to build Elliptic Curves over GF(2^(DE)) having a nearly
// prime order. It uses algorithm described in IEEE P1363 A.11.6 (I don't know
// whether this algorithm is patented or not).
// * See in GF2DemoMain.pas for an example of use.
//------------------------------------------------------------------------------
// WARNING
// Conventionally, polynomials that are elements of a field with a normal basis 
// are displayed with a reversed bit order, i.e., starting with low indice on 
// the left to high indice on the right.
// If you want (or have to) follow this convention, you should reverse the bit 
// order of polynomials with the GF2Reverse routine before creating strings.
//==============================================================================

function GetSubfieldCurve(A, B, SA, SB : TGF2Poly;
                          O, K         : TInteger;
                          Bound        : Longword;
                          gfm, gfd     : TGF2N): Boolean;

implementation

//------------------------------------------------------------------------------
// Get the order of a curve Y^2 + XY = X^3 + AX^2 + B over GF(2^d) for d < 17
// -> A, B, elements of GF(2^d), coefficients of the curve
// -> G = GF(2^d), Galois field of which field degree is less than 17
// <- Result := Order of the curve
// See P1363 A.11.4
//------------------------------------------------------------------------------
function SubfieldCurveOrder(A, B: TGF2Poly; G: TGF2N): Longint;
  var
    Z      : TGF2Poly;
    Max, i : Longword;
begin
  if G.FieldDegree > 16 then
    HITRaiseInvalidArg('SubfieldCurveOrder','Field degree > 16');

  Result := 0;
  with G do
  begin
    Max := Longword(1) shl FieldDegree;
    Z := TGF2Poly.Create;
    try
      for i := 1 to Max-1 do
      begin
        GF2Set(Z,i);
        Sqr(Z);
        Inv(Z);
        Mul(Z,B);
        GF2Add(Z,i);
        if Trace(Z) then Dec(Result) else Inc(Result);
      end;
    finally
      Z.Free;
    end;
    if Trace(A) then Result := -Result;
  end;
  Inc(Result,Max+1);
end;

//------------------------------------------------------------------------------
// Get the order of a curve Y^2 + XY = X^3 + ?X^2 + ? over the field GF(2^(de))
// being given the order of the curve over the field GF(2^d)
// -> s, order of the curve over GF(2^d)
// -> d, degree of the subfield
// -> e, degree of the extension field
// <- O := order of the curve
// See P1363 A.11.5
//------------------------------------------------------------------------------
procedure ExtfieldCurveOrder(O: TInteger; s, d, e: Longint);
  var K, P, Q : TInteger;
begin
  K := nil;
  P := nil;
  Q := nil;
  try
    K := TInteger.Create(e);
    P := TInteger.Create;
    Q := TInteger.Create;
    //----- Q := 2^d
    Q.Bit[d] := true;
    //----- P := 2^d + 1 - s
    ISet(P,Q);
    IInc(P);
    ISub(P,s);
    //----- O := V[e] element of the Lucas sequence (P,Q)
    ILucas(O,K,P,Q);
    //----- O := 2^(de) + 1 - V[e]
    K.Set1;
    K.Bit[e*d] := true;
    ISubr(O,K);
  finally
    K.Free;
    P.Free;
    Q.Free;
  end;
end;

//------------------------------------------------------------------------------
// Result := TRUE if and only if O = F * P with
//   1) F contains only factors <= L.LastPrime,
//   2) P > L.LastPrime and strong pseudoprime to the bases 2,3,5,7 and 11.
// <- F, small factors of O
// O is not modified
//------------------------------------------------------------------------------
function NearlyPrime(O, F: TInteger; L: TPrimeList): Boolean;
  const RATIO = 0.6; //Should be in [0.0, 1.0]
  var
    N, M     : TInteger;
    i, s, ns : Longint;
    d, n0    : Longword;
begin
  Result := false;
  N := nil;
  M := nil;
  try
    s := Trunc(RATIO * O.BitSize);
    N := TInteger.Create(O);
    //----- Divisibility by 2
    i := IShruo(N);
    if N.BitSize < s then Exit;
    F.Set1;
    if i > 0 then IShl(F,i);
    //----- Divisibility by other primes
    M := TInteger.Create;
    ns := N.LongSize;
    n0 := N.Long[0];
    for i := 2 to L.Count do
    begin
      d := L[i];
      if (ns = 1) and (n0 < d) then Exit;
      ISet(M,N);
      if IDiv(M,d) = 0 then
      begin
        repeat
          IMul(F,d);
          ISet(N,M);
        until IDiv(M,d) <> 0;
        if N.BitSize < s then Exit;
        ns := N.LongSize;
        n0 := N.Long[0];
      end;
    end;
    //----- Check size and primality of the remaining part
    Result := (N.BitSize >= s) and IProbablyPrime(N,0);
  finally
    N.Free;
    M.Free;
  end;
end;

//------------------------------------------------------------------------------
// Compute the coefficients of the curve over GF(2^de) from the coefficients of
// the curve over GF(2^d)
// Require: SA and SB (inputs) should be element of GF(2^d)
// See P1363 A.5.7
//------------------------------------------------------------------------------
procedure GetExtCoefs(A, B, SA, SB: TGF2Poly; gfm, gfd: TGF2N);
  var
    P, R   : TGF2Poly;
    i, Max : Longint;
begin
  P := nil;
  R := nil;
  try
    P := TGF2Poly.Create;
    R := TGF2Poly.Create;
    gfd.GetFieldPoly(P);
    gfm.FindRoot(R,P,false); //R := Root of P over gfm

    Max := SA.Degree;
    i := SB.Degree;
    if Max < i then Max := i;

    A.Set0;
    B.Set0;
    case gfd.BasisType of
      btStandard:
        with gfm do
        begin
          Set1(P);
          for i := 0 to Max do
          begin
            if SA[i] then Add(A,P);
            if SB[i] then Add(B,P);
            Mul(P,R);
          end;
        end;
      btNormal:
        with gfm do
        begin
          for i := 0 to Max do
          begin
            if SA[i] then Add(A,R);
            if SB[i] then Add(B,R);
            Sqr(R);
          end;
        end;
    end;
  finally
    P.Free;
    R.Free;
  end;
end;

//==============================================================================
// Compute the coefficients of a curve (if any) over the field GF(2^(de))
// -> gfm, Galois field GF(2^(de))
// -> gfd, Galois field GF(2^d)
// <- A,B, coefficients of a curve of nearly prime order O over GF(2^(de))
// <- SA,SB, coefficients of the subfield curve over GF(2^d)
// <- O, order of EC
// <- F, small prime factors of O
// <- Bound, upper limit for trial division
// Result := TRUE if a curve exists, FALSE otherwise
// See P1363 A.11.6
//==============================================================================
function GetSubfieldCurve(A, B, SA, SB : TGF2Poly;
                          O, K         : TInteger;
                          Bound        : Longword;
                          gfm, gfd     : TGF2N): Boolean;
  var
    T                  : TGF2Poly;
    U                  : TInteger;
    Max, i, s, m, d, e : Longint;
    Prime              : TPrimeList;
  //--------------------------------------------------------------------------
  procedure SetCoefficients(Lambda: Boolean);
  begin
    GetExtCoefs(A,B,SA,SB,gfm,gfd);
    if Lambda then
    begin
      if Odd(gfm.FieldDegree) then gfm.Set1(T)
      else
      repeat
        gfm.Rnd(T);
      until gfm.Trace(T);
      gfm.Add(A,T);
    end;
  end;
  //--------------------------------------------------------------------------
begin
  m := gfm.FieldDegree;
  d := gfd.FieldDegree;
  if (m mod d) <> 0 then
    HITRaiseInvalidArg('GetSubfieldCurve','d should divide m');
  e := m div d;
  if IGCD64(e,d) <> 1 then
    HITRaiseInvalidArg('GetSubfieldCurve','GCD(d, m div d) <> 1');

  //----- Set "TraceFlag" (if possible and if need be)
  if not gfd.CanSwapBasisType then gfd.SetTraceFlag;

  T := nil;
  U := nil;
  Prime := nil;
  try
    Prime := TPrimeList.Create;
    if Bound < 29 then Bound := 29;
    Prime.SetBound(Bound); //To do trial division with all primes <= Bound
    T := TGF2Poly.Create;
    U := TInteger.Create;
    U.Bit[m+1] := true;
    IAdd(U,2);
    gfd.Rnd(SA);
    gfd.Rnd(T);
    Max := Longint(1) shl d;
    for i := 1 to Max-1 do
    begin
      GF2Set(SB,T);
      GF2Add(SB,i);
      if not SB.Is0 then //SB should be different from 0
      begin
        s := SubfieldCurveOrder(SA,SB,gfd);
        ExtfieldCurveOrder(O,s,d,e);
        if NearlyPrime(O,K,Prime) then
        begin
          SetCoefficients(false);
          Result := true;
          Exit;
        end;
        ISubr(O,U);
        if NearlyPrime(O,K,Prime) then
        begin
          SetCoefficients(true);
          Result := true;
          Exit;
        end;
      end;
    end;
    Result := false;
  finally
    T.Free;
    U.Free;
    Prime.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

