////////////////////////////////////////////////////////////////////////////////
// Elliptic Curves over GF(2^N)
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit GF2N_EllipticCurves;

{.$DEFINE DEBUG} //Debugging code

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$DENYPACKAGEUNIT ON}
{$A+,B-,G-,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Classes, Integers, GF2;

//==============================================================================
// The elliptic curves over GF(2^N) used here are always
// y^2 + xy + x^3 + Ax^2 + B = 0
// Such a curve is non supersingular if and only if B <> 0.
// Since (0,0) cannot be a valid point, it is used as "point at infinity".
//------------------------------------------------------------------------------
// WARNING
// Conventionally, polynomials that are elements of a field with a normal basis 
// are displayed with a reversed bit order, i.e., starting with low indice on 
// the left to high indice on the right.
// If you want (or have to) follow this convention, you should reverse the bit 
// order of polynomials with the GF2Reverse routine before creating strings.
//==============================================================================

type
  TGF2NECurve = class;

  //===== Point of a curve y^2 + xy + x^3 + Ax^2 + B = 0 over GF(2^N)
  //      Affine coordinates
  //      Point at infinity O = (0,0)
  TGF2NEPoint = class(THITObject)
  private
    // Fields
    fX: TGF2Poly;
    fY: TGF2Poly;
  public
    // Constructors and destructor
    constructor Create; overload;
    constructor Create(P: TGF2NEPoint); overload;
    destructor  Destroy; override;
    // Methods
    function  Clone: TGF2NEPoint;
    procedure Copy(P: TGF2NEPoint);
    function  Equ(P: TGF2NEPoint): Boolean;
    function  IsO: Boolean;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    function  OnCurve(C: TGF2NECurve): Boolean;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure SetO;
    procedure SetXY(X, Y: TGF2Poly);
    // Properties
    property X: TGF2Poly read fX; //Don't use it to modify fX
    property Y: TGF2Poly read fY; //Don't use it to modify fY
  end;

  //===== Curve (Weierstrass form)
  //      y^2 + xy + x^3 + Ax^2 + B = 0 over GF(2^N)
  TGF2NECurveStatus = (csEmpty, csHasField, csHasCoefs);
  TGF2NECurve = class(THITObject)
  private
    // Fields
    fField: TGF2N; //Field GF(2^N)
    fA: TGF2Poly; //Coef, polynomial over GF(2)
    fB: TGF2Poly; //Coef, polynomial over GF(2)
    fC: TGF2Poly; //Aux, equal to fB^(2^(fField.FieldDegree-2)) = fB^(1/4)
    fStatus: TGF2NECurveStatus;
    // Methods
    procedure ErrorNoCoefs(const Caller: string);
    procedure ErrorNoField(const Caller: string);
    function  GetFieldBasisType: Integer;
    function  GetFieldCanSwapBasisType: Boolean;
    procedure JAdd(X0,Y0,Z0, X1,Y1,Z1: TGF2Poly);
    procedure JDouble(X,Y,Z: TGF2Poly);
    procedure JToAC(JX,JY,JZ: TGF2Poly; P: TGF2NEpoint);
  public
    // Constructors and destructor
    constructor Create; overload;
    constructor Create(C: TGF2NECurve); overload;
    destructor  Destroy; override;
    // Methods
    procedure Add(P, Q: TGF2NEPoint);
    function  Clone: TGF2NECurve;
    procedure Copy(C: TGF2NECurve);
    procedure Double(P: TGF2NEPoint);
    procedure GetCoefs(A, B: TGF2Poly);
    procedure GetField(F: TGF2N);
    procedure GetFieldPoly(P: TGF2Poly);
    procedure GetOrder2Point(P: TGF2NEPoint);
    procedure GetRndPoint(P: TGF2NEPoint);
    function  Halves(H1, H2, P: TGF2NEPoint): Boolean;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure Mul(P: TGF2NEPoint; K: TInteger; Jacobian: Boolean = false);
    procedure Negate(P: TGF2NEPoint); overload;
    function  OnCurve(P: TGF2NEPoint): Boolean;
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure SetCoefs(const A, B: array of Longint); overload;
    procedure SetCoefs(A, B: TGF2Poly); overload;
    procedure SetField(Deg: Byte; Basis: TGF2NBasisType); overload;
    procedure SetField(const Degs: array of Longint; Check: Boolean); overload;
    procedure SetField(F: TGF2N); overload;
    procedure SetField(P: TGF2Poly; Basis: TGF2NBasisType; Check: Boolean);
              overload;
    procedure SetRndCoefs(BOnly: Boolean = false);
    procedure Sub(P, Q: TGF2NEPoint);
    procedure SwapFieldBasisType;
    // Properties
    property  FieldBasisType: Integer read GetFieldBasisType;
    property  FieldCanSwapBasisType: Boolean read GetFieldCanSwapBasisType;
    property  Status: TGF2NECurveStatus read fStatus;
  end;

implementation
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// Resource strings
////////////////////////////////////////////////////////////////////////////////

resourcestring
  sECN0001 = 'Curve without coefs';
  sECN0002 = 'Curve without field';
  sECN0003 = 'Invalid degrees';
  sECN0004 = 'Invalid B coefficient';
  sECN0005 = 'Invalid A coefficient';
  sECN0006 = 'Operation impossible with the used field basis type';
  sECN0007 = 'Invalid Jacobian point';


////////////////////////////////////////////////////////////////////////////////
// Consts, Types
////////////////////////////////////////////////////////////////////////////////

type
  TEPHeaderFile = array[0..24] of AnsiChar;
  TECHeaderFile = array[0..24] of AnsiChar;

const
  //----- Header of "point" file
  EPHeaderFile : TEPHeaderFile = 'HIT EC over GF2N - Point ';
  //----- Header of "curve" file
  ECHeaderFile : TECHeaderFile = 'HIT EC over GF2N - Curve ';


////////////////////////////////////////////////////////////////////////////////
// TGF2NEPoint class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
//==============================================================================
function TGF2NEPoint.Clone: TGF2NEPoint;
begin
  Result := TGF2NEPoint.Create(Self);
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.Copy(P: TGF2NEPoint);
begin
  GF2Set(fX,P.fX);
  GF2Set(fY,P.fY);
end;

//==============================================================================
//==============================================================================
constructor TGF2NEPoint.Create;
begin
  inherited Create;
  fX := TGF2Poly.Create;
  fY := TGF2Poly.Create;
end;

//==============================================================================
//==============================================================================
constructor TGF2NEPoint.Create(P: TGF2NEPoint);
begin
  inherited Create;
  fX := P.fX.Clone;
  fY := P.fY.Clone;
end;

//==============================================================================
//==============================================================================
destructor TGF2NEPoint.Destroy;
begin
  fX.Free;
  fY.Free;
  inherited Destroy;
end;

//==============================================================================
// Result := TRUE iff Self = P
//==============================================================================
function TGF2NEPoint.Equ(P: TGF2NEPoint): Boolean;
begin
  Result := GF2Equ(fX,P.fX) and GF2Equ(fY,P.fY);
end;

//==============================================================================
// Return TRUE iff the point is the "point at infinity"
//==============================================================================
function TGF2NEPoint.IsO: Boolean;
begin
  Result := fX.Is0 and fY.Is0;
end;

//==============================================================================
// Load Self from a file
//==============================================================================
procedure TGF2NEPoint.LoadFromFile(const FileName: string);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------------------------
  var
    F : TFileStream;
    H : TEPHeaderFile;
begin
  F := TFileStream.Create(FileName,HITLOADFLAGS);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> EPHeaderFile then Error;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.LoadFromStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    fX.LoadFromStream(Stream);
    fY.LoadFromStream(Stream);
  except
    Stream.Position := pos;
    SetO;
    raise;
  end;
end;

//==============================================================================
// Result := TRUE iff Self in on curve C
//==============================================================================
function TGF2NEPoint.OnCurve(C: TGF2NECurve): Boolean;
begin
  Result := C.OnCurve(Self);
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,HITSAVEFLAGS);
  try
    F.WriteBuffer(EPHeaderFile,SizeOf(EPHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.SaveToStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    fX.SaveToStream(Stream);
    fY.SaveToStream(Stream);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.SetO;
begin
  fX.Set0;
  fY.Set0;
end;

//==============================================================================
//==============================================================================
procedure TGF2NEPoint.SetXY(X, Y: TGF2Poly);
begin
  GF2Set(fX,X);
  GF2Set(fY,Y);
end;

////////////////////////////////////////////////////////////////////////////////
// TGF2NECurve class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Add
// -> P, Q, points
// <- P = P + Q
//==============================================================================
procedure TGF2NECurve.Add(P, Q: TGF2NEPoint);
  var S, T : TGF2Poly;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Add');

  //----- If Q = O, there is nothing to do
  if Q.IsO then Exit;
  //----- If P = O then P := Q
  if P.IsO then
  begin
    P.Copy(Q);
    Exit;
  end;
  //----- PX = QX ?
  if GF2Equ(P.fX,Q.fX) then
  begin
    //----- If PY <> QY then return P = O
    if not GF2Equ(P.fY,Q.fY) then P.SetO
    else
    //----- If PX = 0 then return P = O (PX = 0 -> P is 2-order point)
    if P.fX.Is0 then P.SetO
    else Double(P);
    Exit;
  end;

  S := nil;
  T := nil;
  with fField do
  try
    S := P.fX.Clone;
    T := P.fY.Clone;
    //----- S := (PY + QY)/(PX + QX)
    Add(S,Q.fX);
    Inv(S);
    Add(T,Q.fY);
    Mul(S,T);
    //----- T = S + S + QX + A
    GF2Set(T,S);
    Sqr(T);
    Add(T,S);
    Add(T,Q.fX);
    Add(T,fA);
    //----- PX' = PX + T
    Add(P.fX,T);
    //----- PY' = PY + PX' + ST
    Mul(T,S);
    Add(P.fY,T);
    Add(P.fY,P.fX);
  finally
    S.Free;
    T.Free;
  end;
end;

//==============================================================================
// Create and set Result equal to Self
//==============================================================================
function TGF2NECurve.Clone: TGF2NECurve;
begin
  Result := TGF2NECurve.Create(Self);
end;

//==============================================================================
// Set Self equal to C
//==============================================================================
procedure TGF2NECurve.Copy(C: TGF2NECurve);
begin
  fField.Copy(C.fField);
  GF2Set(fA,C.fA);
  GF2Set(fB,C.fB);
  GF2Set(fC,C.fC);
  fStatus := C.fStatus;
end;

//==============================================================================
//==============================================================================
constructor TGF2NECurve.Create;
begin
  inherited Create;
  fField := TGF2N.Create;
  fA := TGF2Poly.Create;
  fB := TGF2Poly.Create;
  fC := TGF2Poly.Create;
end;

//==============================================================================
// Create Self equal to C
//==============================================================================
constructor TGF2NECurve.Create(C: TGF2NECurve);
begin
  inherited Create;
  fField := C.fField.Clone;
  fA := C.fA.Clone;
  fB := C.fB.Clone;
  fC := C.fC.Clone;
  fStatus := C.fStatus;
end;

//==============================================================================
//==============================================================================
destructor TGF2NECurve.Destroy;
begin
  fField.Free;
  fA.Free;
  fB.Free;
  fC.Free;
  inherited Destroy;
end;

//==============================================================================
// Double
// -> P, point
// <- P = P x 2
//==============================================================================
procedure TGF2NECurve.Double(P: TGF2NEPoint);
  var S, T : TGF2Poly;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Double');

  with fField, P do
  begin
    //----- If PX = O, there is nothing to do
    if fX.Is0 then //Handle the case P = O
    begin
      fY.Set0; //Here, equivalent to P.SetO
      Exit;
    end;
    S := nil;
    T := nil;
    try
      S := fX.Clone;
      T := fX.Clone;
      //----- T := PX + PY/PX
      Inv(T);
      Mul(T,fY);
      Add(T,fX);
      //----- PX' = T + T + A
      GF2Set(fX,T);
      Sqr(fX);
      Add(fX,T);
      Add(fX,fA);
      //----- PY' = PY + PX' + (PX + PX')T
      Add(fY,fX);
      Add(S,fX);
      Mul(S,T);
      Add(fY,S);
    finally
      S.Free;
      T.Free;
    end;
  end;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure TGF2NECurve.ErrorNoCoefs(const Caller: string);
begin
  HITRaiseInvalidCall(ClassName + '.' + Caller,LoadResString(@sECN0001));
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure TGF2NECurve.ErrorNoField(const Caller: string);
begin
  HITRaiseInvalidCall(ClassName + '.' + Caller,LoadResString(@sECN0002));
end;

//==============================================================================
// If you need only B, for instance, use "GetCoefs(nil,X);"
//==============================================================================
procedure TGF2NECurve.GetCoefs(A, B: TGF2Poly);
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('GetCoefs');

  if Assigned(A) then GF2Set(A,fA);
  if Assigned(B) then GF2Set(B,fB);
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.GetField(F: TGF2N);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('GetField');

  F.Copy(fField);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function TGF2NECurve.GetFieldBasisType: Integer;
begin
  if fStatus = csEmpty then Result := -1 else Result := Ord(fField.BasisType);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function TGF2NECurve.GetFieldCanSwapBasisType: Boolean;
begin
  Result := (fStatus <> csEmpty) and fField.CanSwapBasisType;
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.GetFieldPoly(P: TGF2Poly);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('GetFieldPoly');

  fField.GetFieldPoly(P);
end;

//==============================================================================
// Return the unique 2-torsion point of the curve (different from O)
// <- P = (0,B^(1/2))
//==============================================================================
procedure TGF2NECurve.GetOrder2Point(P: TGF2NEPoint);
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('GetOrder2Point');

  with P do
  begin
    fX.Set0;
    GF2Set(fY,fB);
    fField.SqrRoot(fY);
  end;
end;

//==============================================================================
// Generate a random point
// <- P = (X,Y), random point different from (0,0)
//==============================================================================
procedure TGF2NECurve.GetRndPoint(P: TGF2NEPoint);
  var U : TGF2Poly;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('GetRndPoint');

  U := TGF2Poly.Create;
  with fField, P do
  try
    repeat
      Rnd(fX);
      if fX.Is0 then
      begin
        GetOrder2Point(P);
        Exit;
      end;
      //----- U = (PX + A)PX + B
      GF2Set(U,fX);
      Add(U,fA);
      GF2Set(fY,fX);
      Sqr(fY);
      Mul(U,fY);
      Add(U,fB);
      if U.Is0 then
      begin
        fY.Set0;
        Break;
      end;
    until SolveQuad(fY,fX,U);
    //----- Toss up: P or -P ?
    Rnd(U);
    if U.Parity then Negate(P);
  finally
    U.Free;
  end;
end;

//==============================================================================
// Compute H1 and H2 such that P = 2*H1 = 2*H2 (if such halves exist)
// Return TRUE iff the halves exist.
//==============================================================================
function TGF2NECurve.Halves(H1, H2, P: TGF2NEPoint): Boolean;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Halves');

  //----- P = Identity ?
  if P.IsO then
  begin
    GetOrder2Point(H1);
    H2.SetO;
    Result := true;
    Exit;
  end;

  with fField do
  begin
    //----- Compute T = X + Y/X (X and Y are unknow, they are what we want)
    //      Solve f(T) = T + T + A + Px
    GF2Set(H2.fY,fA);
    Add(H2.fY,P.fX);
    Result := SolveQuad(H1.fX,H2.fY); //H1.fX is used for "T"
    if not Result then
    begin
      H1.SetO; //Cleaner
      H2.SetO; //"
      Exit;
    end;
    //----- The 2 roots T and T+1 lead to the 2 halves H1 and H2
    GF2Set(H2.fY,H1.fX);
    GF2Set(H1.fY,H1.fX);
    Add1(H1.fY); //!And not GF2Add1 because of possible Normal Basis
    GF2Set(H2.fX,H1.fY);
    //----- Compute H1.X
    //      X := (T Px + Py)^(1/2)
    Mul(H1.fX,P.fX);
    Add(H1.fX,P.fY);
    SqrRoot(H1.fX);
    //----- Compute H1.Y
    //      Y := X (X + T + 1)
    Add(H1.fY,H1.fX);
    Mul(H1.fY,H1.fX);
    //----- Compute H2.X
    //      X := ((T + 1) Px + Py)^(1/2)
    Mul(H2.fX,P.fX);
    Add(H2.fX,P.fY);
    SqrRoot(H2.fX);
    //----- Compute H2.Y
    //      Y := X (X + T)
    Add(H2.fY,H2.fX);
    Mul(H2.fY,H2.fX);
  end;
end;

//------------------------------------------------------------------------------
// Adding in Jacobian coordinates
// -> (X0,Y0,Z0) and (X1,Y1,Z1), points of the curve
// <- (X0,Y0,Z0) = (X0,Y0,Z0) + (X1,Y1,Z1)
//------------------------------------------------------------------------------
// The IEEE P1363 Annex A calls "projective" what is called "Jacobian" here.
// The Jacobian coordinates are projective ones, but the word "projective" is
// ordinary given to an other kind of projective coordinates. In most articles
// or books you will read, "projective" will not be equivalent to "projective"
// in the IEEE P1363.
//------------------------------------------------------------------------------
procedure TGF2NECurve.JAdd(X0,Y0,Z0, X1,Y1,Z1: TGF2Poly);
  var T1, T2, T3, T4 : TGF2Poly;
begin
  with fField do
  begin
    //----- (X1,Y1,Z1) = Idendity?
    if Z1.Is0 and Is1(X1) and Is1(Y1) then Exit;
    //----- (X0,Y0,Z0) = Idendity?
    if Z0.Is0 and Is1(X0) and Is1(Y0) then
    begin
      GF2Set(X0,X1);
      GF2Set(Y0,Y1);
      GF2Set(Z0,Z1);
      Exit;
    end;
  end;
  T1 := nil;
  T2 := nil;
  T3 := nil;
  T4 := nil;
  with fField do
  try
    T1 := X1.Clone;
    T2 := Y1.Clone;
    T3 := TGF2Poly.Create;
    T4 := X1.Clone;
    if not Is1(Z1) then //!Not Z1.Is1 because of possible Normal Basis
    begin
      GF2Set(T3,Z1);
      Sqr(T3);
      Mul(X0,T3);
      Mul(T3,Z1);
      Mul(Y0,T3);
    end;
    GF2Set(T3,Z0);
    Sqr(T3);
    Mul(T4,T3);
    Add(X0,T4);
    Mul(T3,Z0);
    Mul(T3,Y1);
    Add(Y0,T3);
    if X0.Is0 then
    begin
      if Y0.Is0 then
      begin
        GF2Set(X0,X1);
        GF2Set(Y0,Y1);
        GF2Set(Z0,Z1);
        JDouble(X0,Y0,Z0);
      end
      else
      begin
        Set1(X0); //!And not X0.Set1 because of possible Normal Basis
        Set1(Y0);
        Z0.Set0;
      end;
      Exit;
    end;
    Mul(T1,Y0);
    Mul(Z0,X0);
    Mul(T2,Z0);
    Add(T1,T2);
    GF2Set(T2,Z0);
    Sqr(T2);
    GF2Set(T3,T1);
    Mul(T3,T2);
    if not Is1(Z1) then Mul(Z0,Z1);
    GF2Set(T1,Y0);
    Add(T1,Z0);
    Mul(Y0,T1);
    GF2Set(T2,X0);
    Sqr(T2);
    Mul(X0,T2);
    if not fA.Is0 then
    begin
      GF2Set(T2,Z0);
      Sqr(T2);
      Mul(T2,fA);
      Add(X0,T2);
    end;
    Add(X0,Y0);
    Mul(T1,X0);
    GF2Swp(Y0,T1); //Faster than "GF2Set"
    Add(Y0,T3);
  finally
    T1.Free;
    T2.Free;
    T3.Free;
    T4.Free;
  end;
end;

//------------------------------------------------------------------------------
// Doubling in Jacobian coordinates
// -> (X,Y,Z) point of the curve
// <- (X,Y,Z) = 2(X,Y,Z)
//------------------------------------------------------------------------------
procedure TGF2NECurve.JDouble(X,Y,Z: TGF2Poly);
  var T : TGF2Poly;
begin
  //----- Handle special case
  if X.Is0 or Z.Is0 then
  begin
    fField.Set1(X); //!Not X.Set1 because of possible Normal Basis
    fField.Set1(Y);
    Z.Set0;
    Exit;
  end;

  T := fC.Clone;
  with fField do
  try
    Mul(Y,Z);
    Sqr(Z);
    Mul(T,Z);
    Mul(Z,X);
    Add(Y,Z);
    Add(T,X);
    Sqr(T);
    Sqr(T);
    Sqr(X);
    Add(Y,X);
    Mul(Y,T);
    Sqr(X);
    Mul(X,Z);
    Add(Y,X);
    GF2Swp(X,T); //Faster than "GF2Set"
  finally
    T.Free;
  end;
end;

//------------------------------------------------------------------------------
// Conversion: Jacobian coordinates to affine coordinates
//------------------------------------------------------------------------------
procedure TGF2NECurve.JToAC(JX,JY,JZ: TGF2Poly; P: TGF2NEpoint);
  var U, V : TGF2Poly;
begin
  if JZ.Is0 then
  begin
    if JX.Is0 and JY.Is0 then
       HITError(ClassName + '.JToAC'#13#10 + LoadResString(@sECN0007));

    P.fX.Set0;
    P.fY.Set0;
  end
  else
  begin
    V := nil;
    U := JZ.Clone;
    with fField, P do
    try
      Inv(U);
      V := U.Clone;
      Sqr(U);
      GF2Set(fX,JX);
      Mul(fX,U);  //AX := JX / JZ^2
      Mul(U,V);
      GF2Set(fY,JY);
      Mul(fY,U);  //AY := JY / JZ^3
    finally
      U.Free;
      V.Free;
    end;
  end;
end;

//==============================================================================
// Load Self from a file
//==============================================================================
procedure TGF2NECurve.LoadFromFile(const FileName: string);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------------------------
  var
    F : TFileStream;
    H : TECHeaderFile;
begin
  F := TFileStream.Create(FileName,HITLOADFLAGS);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> ECHeaderFile then Error;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.LoadFromStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    Stream.ReadBuffer(fStatus,SizeOf(fStatus));
    fField.LoadFromStream(Stream);
    fA.LoadFromStream(Stream);
    fB.LoadFromStream(Stream);
    fC.LoadFromStream(Stream);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
// Mul
// -> P, point
// -> K, integer
// -> Jacobian, if TRUE use Jacobian coordinates, if FALSE use affine ones
// <- P = P x K
//==============================================================================
procedure TGF2NECurve.Mul(P: TGF2NEPoint; K: TInteger; Jacobian: Boolean = false);
  const Selector : array[Boolean,Boolean] of Byte = ((0,1),(2,0));
  var
    PJX, PJY, PJZ, QpJX, QpJY, QpJZ, QmJY : TGF2Poly;
    Qp, Qm                                : TGF2NEPoint;
    H                                     : TInteger;
    i                                     : Longint;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Mul');

  //===== Handle special cases
  //----- If P = O then nothing to do
  if P.IsO then Exit;
  //----- If K = 0, return P.O
  if K.Is0 then
  begin
    P.SetO;
    Exit;
  end;
  //----- If K = 1, nothing to do
  if K.Is1 then Exit;
  //---- If K negative, (PX,PY)*(-K) = -(PX,PY)*K
  if K.Neg then
  begin
    Negate(P);
    if K.IsMinus1 then Exit;
  end;

  //===== Use algorithm described in IEEE-P1363 Annex A
  //----- Jacobian coordinates
  if Jacobian then
  begin
    PJX := nil;
    PJY := nil;
    PJZ := nil;
    QpJX := nil;
    QpJY := nil;
    QpJZ := nil;
    QmJY := nil;
    H := K.Clone;
    try
      //----- Convert from affine to Jacobian coordinates
      PJX := P.fX.Clone;
      PJY := P.fY.Clone;
      PJZ := TGF2Poly.Create;
      fField.Set1(PJZ); //!Not PJZ.Set1 because of possible Normal Basis
      QpJX := PJX.Clone;
      QpJY := PJY.Clone;
      QpJZ := PJZ.Clone;
      QmJY := PJY.Clone;
      fField.Add(QmJY,PJX);
      //----- Multiplication
      IMul(H,3);
      for i := H.BitSize-2 downto 1 do
      begin
        JDouble(PJX,PJY,PJZ);
        case Selector[H.Bit[i],K.Bit[i]] of
          1: JAdd(PJX,PJY,PJZ,QpJX,QmJY,QpJZ); //P := P - Q
          2: JAdd(PJX,PJY,PJZ,QpJX,QpJY,QpJZ); //P := P + Q
        end;
      end;
      //----- Convert to from Jacobian to affine coordinates
      JToAC(PJX,PJY,PJZ,P);
    finally
      H.Free;
      PJX.Free;
      PJY.Free;
      PJZ.Free;
      QpJX.Free;
      QpJY.Free;
      QpJZ.Free;
      QmJY.Free;
    end;
  end
  //----- Affine coordinates
  else
  begin
    Qp := nil;
    Qm := nil;
    H := K.Clone;
    try
      Qp := P.Clone;
      Qm := P.Clone;
      Negate(Qm);
      IMul(H,3);
      for i := H.BitSize-2 downto 1 do
      begin
        Double(P);
        case Selector[H.Bit[i],K.Bit[i]] of
          1: Add(P,Qm); //P := P - Q
          2: Add(P,Qp); //P := P + Q
        end;
      end;
    finally
      H.Free;
      Qp.Free;
      Qm.Free;
    end;
  end;
end;

//==============================================================================
// Negate
// <- P := -P  (if U = (X,Y) then -U = (X,Y+X))
//==============================================================================
procedure TGF2NECurve.Negate(P: TGF2NEPoint);
begin
  //----- If P = O or (PX = 0) then nothing to do
  //      (PX <> 0) => (P <> O), so no need to test PY
  with P do if not fX.Is0 then fField.Add(fY,fX);
end;

//==============================================================================
// Return TRUE iff P is on the curve
//==============================================================================
function TGF2NECurve.OnCurve(P: TGF2NEPoint): Boolean;
  var S, T : TGF2Poly;
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('OnCurve');

  if P.IsO then Result := true
  else
  begin
    S := nil;
    T := nil;
    with fField, P do
    try
      S := fX.Clone;
      T := fX.Clone;
      //----- S = (X + A)X^2 + B
      Sqr(S);
      Add(T,fA);
      Mul(S,T);
      Add(S,fB);
      //----- T := (Y + X)Y
      GF2Set(T,fX);
      Add(T,fY);
      Mul(T,fY);
      //----- S = T ?
      Result := GF2Equ(S,T);
    finally
      S.Free;
      T.Free;
    end;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,HITSAVEFLAGS);
  try
    F.WriteBuffer(ECHeaderFile,SizeOf(ECHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
//==============================================================================
procedure TGF2NECurve.SaveToStream(Stream: TStream);
  var pos : Longint;
begin
  pos := Stream.Position;
  try
    Stream.WriteBuffer(fStatus,SizeOf(fStatus));
    fField.SaveToStream(Stream);
    fA.SaveToStream(Stream);
    fB.SaveToStream(Stream);
    fC.SaveToStream(Stream);
  except
    Stream.Position := pos;
    raise;
  end;
end;

//==============================================================================
// Overloaded
// Set with "degree" array (See GF2SetDegs)
//==============================================================================
procedure TGF2NECurve.SetCoefs(const A, B: array of Longint);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseInvalidArg(ClassName + '.SetCoefs',LoadResString(@sECN0003));
  end;
  //--------------------------------------------------------------------------
  var AA, BB : TGF2Poly;
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('SetCoefs');

  AA := nil;
  BB := nil;
  try
    AA := TGF2Poly.Create;
    try
      GF2SetDegs(AA,A);
    except
      on EHITInvalidArg do Error;
    end;
    if AA.Degree >= fField.FieldDegree then Error;
    BB := TGF2Poly.Create;
    try
      GF2SetDegs(BB,B);
    except
      on EHITInvalidArg do Error;
    end;
    if BB.Is0 or (BB.Degree >= fField.FieldDegree) then Error;
    GF2Set(fA,AA);
    GF2Set(fB,BB);
  finally
    AA.Free;
    BB.Free;
  end;

  GF2Set(fC,fB);
  fField.SqrRoot(fC);
  fField.SqrRoot(fC);

  fStatus := csHasCoefs;
end;

//==============================================================================
// Overloaded
// Set A and B
// If new coefs are not valid, the old ones (if any) are not changed
//==============================================================================
procedure TGF2NECurve.SetCoefs(A, B: TGF2Poly);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('SetCoefs');

  //----- Check
  if B.Is0 or (B.Degree >= fField.FieldDegree) then
    HITRaiseInvalidArg(ClassName + '.SetCoefs',LoadResString(@sECN0004));
  if Assigned(A) then
    if A.Degree >= fField.FieldDegree then
      HITRaiseInvalidArg(ClassName + '.SetCoefs',LoadResString(@sECN0005));

  GF2Set(fB,B);
  if Assigned(A) then GF2Set(fA,A) else fA.Set0;
  GF2Set(fC,fB);
  fField.SqrRoot(fC);
  fField.SqrRoot(fC);

  fStatus := csHasCoefs;
end;

//==============================================================================
// Overloaded
// Deg must be greater than 1
//
// If the new field is not valid, the old one (if any) is not changed
//==============================================================================
procedure TGF2NECurve.SetField(Deg: Byte; Basis: TGF2NBasisType);
  var gf : TGF2N;
begin
  if Deg < 2 then
    HITRaiseInvalidArg(ClassName + '.SetField','Deg < 2');

  gf := TGF2N.Create;
  try
    gf.SetField(Deg,Basis);
  except
    gf.Free;
    raise;
  end;
  //----- All is right, free the old field (if any) and set the new one
  fField.Free;
  fField := gf;
  fField.CheckParam := false; //A little faster
  fStatus := csHasField;
end;

//==============================================================================
// Overloaded
// We must have Degs = [dmax, ..., 0] and Degs must be an irreducible
// polynomial over GF(2)
//
// If the new field is not valid, the old one (if any) is not changed
//==============================================================================
procedure TGF2NECurve.SetField(const Degs: array of Longint; Check: Boolean);
  var gf : TGF2N;
begin
  gf := TGF2N.Create;
  try
    gf.SetField(Degs,Check);
  except
    gf.Free;
    raise;
  end;
  //----- All is right, free the old field (if any) and set the new one
  fField.Free;
  fField := gf;
  fField.CheckParam := false;
  fStatus := csHasField;
end;

//==============================================================================
// Overloaded
//==============================================================================
procedure TGF2NECurve.SetField(F: TGF2N);
begin
  if Assigned(F) then
  begin
    fStatus := csEmpty;
    fField.Copy(F);
    fStatus := csHasField;
  end;
end;

//==============================================================================
// Overloaded
// P must be an irreducible polynomial over GF(2)
// If the new field is not valid, the old one (if any) is not changed
//==============================================================================
procedure TGF2NECurve.SetField(P: TGF2Poly; Basis: TGF2NBasisType; Check: Boolean);
  var gf : TGF2N;
begin
  gf := TGF2N.Create;
  try
    gf.SetField(P,Basis,Check);
  except
    gf.Free;
    raise;
  end;
  //----- All is right, free the old field (if any) and set the new one
  fField.Free;
  fField := gf;
  fField.CheckParam := false;
  fStatus := csHasField;
end;

//==============================================================================
// Generate random coefficients fA and fB
//==============================================================================
procedure TGF2NECurve.SetRndCoefs(BOnly: Boolean = false);
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('SetRndCoefs');

  fStatus := csHasField;
  if BOnly then fA.Set0 else fField.Rnd(fA);
  repeat
    fField.Rnd(fB);
  until not fB.Is0;
  fStatus := csHasCoefs;
end;

//==============================================================================
// Sub
// -> P, Q, points
// <- P = P - Q
//==============================================================================
procedure TGF2NECurve.Sub(P, Q: TGF2NEPoint);
begin
  //----- Check
  if fStatus <> csHasCoefs then ErrorNoCoefs('Sub');

  with P do
  if P = Q then SetO
  else
  begin
    fField.Add(fY,fX); //Negate P (if P = O then fX = 0, all is right)
    Add(P,Q);
    fField.Add(fY,fX);
  end;
end;

//==============================================================================
// Change the basis type (if possible) of the field GF(2^N)
//==============================================================================
procedure TGF2NECurve.SwapFieldBasisType;
begin
  //----- Check
  if fStatus = csEmpty then ErrorNoField('SwapFieldBasisType');
  if not fField.CanSwapBasisType then
    HITRaiseInvalidArg(ClassName + '.SwapFieldBasisType',LoadResString(@sECN0006));

  fField.SwapBasisType;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

