////////////////////////////////////////////////////////////////////////////////
// Computation of Gaussian normal field polynomials for GF(2^N)
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit GF2N_GNBFieldPolys;

interface
uses Integers, GF2;

{.$DEFINE DEBUG}

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

//==============================================================================
// In order to compute Gaussian sub-optimal normal field polynomials, the IEEE
// P1363 Standards (A.7.2) gives an algorithm that requires complex floating-
// point arithmetic.
// The method used here requires only finite field arithmetic. It is based on
// the facts:
// 1) We can construct the multiplication matrix for a field GF(2^N) in Gaussian
//    normal basis representation without knowing the field polynomial.
// 2) Over a field GF(2^N), the minimal polynomial of x is always equal to the
//    field polynomial.
//------------------------------------------------------------------------------
// WARNING
// Conventionally, polynomials that are elements of a field with a normal basis 
// are displayed with a reversed bit order, i.e., starting with low indice on 
// the left to high indice on the right.
// If you want (or have to) follow this convention, you should reverse the bit 
// order of polynomials with the GF2Reverse routine before creating strings.
//==============================================================================

function GaussianNormalBasisType(N: Word): Byte;
function GaussianFieldPoly(F: TGF2Poly; N: Word): Byte;

implementation
uses Integers_Primes;
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
// Resource strings
////////////////////////////////////////////////////////////////////////////////

resourcestring
  sGNB0001 = 'P should be prime';
  sGNB0002 = 'T should be greater than 1';
  sGNB0003 = 'T should divide P-1';

////////////////////////////////////////////////////////////////////////////////
// Unused
////////////////////////////////////////////////////////////////////////////////

//------------------------------------------------------------------------------
// Multiplication over GF(2^N) in normal basis
// -> A, B, elements of GF(2^N)
// -> M, multiplication matrix
// <- A := A * B over GF(2^N)
//------------------------------------------------------------------------------
// Was used to check multiplication matrices
//------------------------------------------------------------------------------
procedure NormalMul(A, B: TGF2Poly; M: TGF2SqrMatrix);
  var
    X    : TGF2Poly;
    i, N : Longint;
begin
  if A = B then
  begin
    GF2Rol(A,M.Dimension); //Square
    Exit;
  end;

  X := TGF2Poly.Create;
  try
    N := M.Dimension;
    for i := 0 to N-1 do
    begin
      if GF2ProductParity(A,M,B) then X[i] := true;
      GF2Ror(A,N);
      GF2Ror(B,N);
    end;
    GF2Swp(A,X);
  finally
    X.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// Routines
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Global
// -> N, field degree
// Result := Gaussian normal basis type for GF(2^N),
//           0 if doesn't exist or if type is greater than 255
//==============================================================================
function GaussianNormalBasisType(N: Word): Byte;
  //--------------------------------------------------------------------------
  function OrderOf2(P: Longword): Longword;
    var H, X : Longword;
  begin
    //----- Get the order of 2 modulo P
    H := P shr 1; //(P odd) -> (2H = P-1)
    X := 2;
    Result := 1;
    while X <> 1 do
    begin
      Inc(Result);
      if X <= H then Inc(X,X) else Inc(X,X-P);
    end;
  end;
  //--------------------------------------------------------------------------
  function Check(O, D, P: Longword): Boolean;
    var A, B : Longword;
  begin
    //----- Check whether GCD(D,(P-1)/O)) = 1
    A := (P-1) div O;
    while D > 0 do
    begin
      B := A mod D;
      A := D;
      D := B;
    end;
    Result := A = 1;
  end;
  //--------------------------------------------------------------------------
  var
    P : Longword;
    i : Byte;
begin
  Result := 0;
  if (N > 2) and ((N and 7) <> 0) then
  begin
    P := Longword(N) + 1;
    for i := 1 to 255 do
    begin
      if ITruePrime(P) and Check(OrderOf2(P),N,P) then
      begin
        Result := i;
        Exit;
      end;
      Inc(P,N);
    end;
  end;
end;

//------------------------------------------------------------------------------
// Local
// Result := (A * B) mod P
//------------------------------------------------------------------------------
function MulMod(A, B, P: Longword): Longword; register;
asm
  mul  edx
  div  ecx
  mov  eax, edx
end;

//------------------------------------------------------------------------------
// Local
// -> P, prime
// Result := Integer having order T modulo P or 0 if doesn't exist
//------------------------------------------------------------------------------
function GetIntOfOrderTModP(T, P: Longword): Longword;
  var K, G, H : Longword;
begin
  //----- Check
  if not ITruePrime(P) then
    HITRaiseInvalidArg('GetIntOfOrderTModP',LoadResString(@sGNB0001));
  if T < 2 then
    HITRaiseInvalidArg('GetIntOfOrderTModP',LoadResString(@sGNB0002));
  if ((P-1) mod T) <> 0 then
    HITRaiseInvalidArg('GetIntOfOrderTModP',LoadResString(@sGNB0003));

  G := 1;
  repeat
    Inc(G);
    H := G;
    K := 1;
    repeat
      H := MulMod(H,G,P);
      Inc(K);
    until H <= 1;
  until (K mod T) = 0;
  K := K div T;
  H := G;
  while K > 1 do
  begin
    H := MulMod(H,G,P);
    Dec(K);
  end;
  Result := H;
end;

//------------------------------------------------------------------------------
// Local
// Build a multiplication matrix for the field GF(2^N) in Gaussian normal basis
// See IEEE P1363, A.3.7 and A.3.8
// -> N, field degree
// -> T, Gaussian normal basis type for GF(2^N)
// -> P, prime, equal to N * T + 1
// -> U, integer having order T modulo P
// <- M, multiplication matrix for GF(2^N)
//------------------------------------------------------------------------------
procedure GetMultiplicationMatrix(M: TGF2SqrMatrix; N, T, P, U: Longword);
  var
    i, j, k, v, w : Longword;
    S             : TDynLongArray;
begin
  //===== Build the S sequence
  SetLength(S,P);
  FillChar(S[0],P*4,0);
  w := 1;
  for i := 0 to T-1 do
  begin
    v := w;
    for j := 0 to N-1 do
    begin
      S[v] := j;
      Inc(v,v);
      if v >= P then Dec(v,P);
    end;
    w := MulMod(w,U,P);
  end;

  //===== Build the M matrix
  with M do
  begin
    Dimension := N;
    SetNull;
    if Odd(T) then
    begin
      u := N shr 1;
      for k := 0 to u-1 do
      begin
        w := k + u;
        FlipCell(k,w);
        FlipCell(w,k);
      end;
    end;
    for i := 1 to P-2 do FlipCell(S[i+1],S[P-i]);
  end;
end;

//------------------------------------------------------------------------------
// Local
// Build the field polynomial of a Galois field having M as multiplication
// matrix.
// This procedure computes the minimal polynomial of x, i.e., it computes a
// polynomial over GF(2^N) equal to the product of (X + x^(2^i)), i running
// through 0..N-1.
// -> M, multiplication matrix for GF(2^N)
// <- F, field polynomial
//------------------------------------------------------------------------------
procedure GetFieldPoly(F: TGF2Poly; M: TGF2SqrMatrix);
  var
    T          : TGF2PolyVector;
    X, U, V    : TGF2Poly;
    N, i, j, k : Longint;
begin
  N := M.Dimension;
  T := nil;
  X := nil;
  try
    //----- T is used as a polynomial over GF(2^N), i.e., as a polynomial
    //      of which the coefficients are elements of GF(2^N)
    T := TGF2PolyVector.Create;
    T.Size := N;
    T[N-1].Set1; //(000..01) with normal basis is equivalent to P(x) = x
    X := TGF2Poly.Create(1); //X := x
    for j := N-2 downto 0 do
    begin
      //----- X := X
      GF2Rol(X,N);
      //----- T[i] := T[i] + T[i+1] * X
      for i := j to N-2 do
      begin
        U := T[i]; //Just copy the pointer
        V := T[i+1];
        for k := 0 to N-1 do
        begin
          if GF2ProductParity(X,M,V) then U.Flip(k);
          if GF2DivX(X) then X[N-1] := true; //Rotation
          if GF2DivX(V) then V[N-1] := true;
        end;
      end;
      //----- T[N-1] := T[N-1] + T[N] * X with T[N] equal to "1"
      GF2Add(T[N-1],X);
    end;

    //===== Set F (casts the poly over GF(2^N) to a poly over GF(2))
    //      Here, the coefs of T should be equal to either 0 or 1, i.e.,
    //      for all i, either T[i] = (00..00) or T[i] = (11..11)
    F.SetMonomial(N);
    X.Fill(N);
    for i := N-1 downto 0 do
      if not T[i].Is0 then
        if GF2Equ(T[i],X) then F[i] := true
        //----- Should never occur...
        else HITError('Internal error in GetFieldPoly !?');
  finally
    T.Free;
    X.Free;
  end;
end;

//------------------------------------------------------------------------------
// Local
// Build a Gaussian optimal normal field polynomial of type II and of degree N
// See IEEE P1363, A.7.2
//------------------------------------------------------------------------------
procedure GetFieldPolyOfTypeII(F: TGF2Poly; N: Longint);
  var
    X, Y : TGF2Poly;
    i    : Longint;
begin
  X := nil;
  Y := nil;
  try
    X := TGF2Poly.Create(1);
    Y := TGF2Poly.Create;
    GF2Set(F,3); //F := x + 1
    for i := 1 to N-1 do
    begin
      GF2Set(Y,X);
      GF2Set(X,F);
      GF2MulX(F);
      GF2Add(F,Y);
    end;
  finally
    X.Free;
    Y.Free;
  end;
end;

//==============================================================================
// Global
// -> N, field degree
// <- F, Gaussian normal field polynomial for GF(2^N) or 0 if doesn't exist
// Result := Type of the Gaussian normal base or 0 if doesn't exist
//==============================================================================
function GaussianFieldPoly(F: TGF2Poly; N: Word): Byte;
  var
    M : TGF2SqrMatrix;
    P : Longword;
begin
  Result := GaussianNormalBasisType(N);
  case Result of
    0: F.Set0;
    1: F.Fill(N+1);
    2: GetFieldPolyOfTypeII(F,N);
    else
      P := Longword(Result) * N + 1;
      M := TGF2SqrMatrix.Create;
      try
        GetMultiplicationMatrix(M,N,Result,P,GetIntOfOrderTModP(Result,P));
        GetFieldPoly(F,M);
      finally
        M.Free;
      end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

