////////////////////////////////////////////////////////////////////////////////
// Galois Fields GF(2^N) with Optimal Normal Basis (type I and II)
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit GF2N_ONBFields;

{.$DEFINE DEBUG}

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Classes, Integers, GF2;

//==============================================================================
// The following TGF2NO class, that allows to work with Optimal Normal Basis
// in a conventional way, was implemented only to show the efficiency of the
// TGF2N class of GF2 when working with Normal Basis.
// For instance,
//
//   onb := TGF2NO.Create;
//   gfn := TGF2N.Create;
//   onb.SetField(254); //ONB of type II
//   onb.GetFieldPoly(F);
//   gfn.SetField(F,btNormal,false); //No check, we are sure that F is normal
//
// and compare the running times of multiplications
//
//   onb.Rnd(A);
//   onb.Rnd(M);
//   GF2Set(B,A);
//   onb.Mul(A,M);
//   gfn.Mul(B,M);
//
// or the running times of inversions
//
//   onb.Rnd(A);
//   GF2Set(B,A);
//   onb.Inv(A);
//   gfn.Inv(B);
//------------------------------------------------------------------------------
// WARNING
// Conventionally, polynomials that are elements of a field with a normal basis 
// are displayed with a reversed bit order, i.e., starting with low indice on 
// the left to high indice on the right.
// If you want (or have to) follow this convention, you should reverse the bit 
// order of polynomials with the GF2Reverse routine before creating strings.
//==============================================================================

type
  //----- Galois Field GF(2^N) with Optimal Basis (type I or II)
  TONBType = (btNone, btType1, btType2);
  TGF2NO = class(THITObject)
  private
    // Fields
    fFieldPoly: TGF2Poly;   //Field polynomial
    fDegree: Longint;       //Equal to fFieldPoly.Degree
    fPoly1: TGF2Poly;       //Polynomial equal to "1"
    fMatrix: TGF2SqrMatrix; //Multiplication matrix
    fONBType: TONBType;     //Optimal Normal Basis type
    fCheckParam: Boolean;   //Check parameter validity (default = TRUE)
  public
    // Class function
    class function GetONBType(Deg: Longint): TONBType;
    // Constructors and destructor
    constructor Create; overload;
    constructor Create(G: TGF2NO); overload;
    destructor  Destroy; override;
    // Methods
    procedure Add(P, Q: TGF2Poly);
    procedure Add1(P: TGF2Poly);
    function  Clone: TGF2NO;
    procedure Copy(G: TGF2NO);
    procedure Divide(P, Q: TGF2Poly);
    procedure GetFieldPoly(P: TGF2Poly);
    procedure HalfTrace(P: TGF2Poly);
    function  Inv(P: TGF2Poly): Boolean;
    function  Is1(P: TGF2Poly): Boolean;
    procedure LoadFromFile(const FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure Mul(P, Q: TGF2Poly);
    procedure Order(N: TInteger);
    function  Pow(P: TGF2Poly; E: Int64): Boolean; overload;
    function  Pow(P: TGF2Poly; E: TInteger): Boolean; overload;
    procedure Rnd(P: TGF2Poly);
    procedure SaveToFile(const FileName: string); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure Set1(P: TGF2Poly);
    procedure SetField(Deg: Longint);
    function  SolveQuad(R, B: TGF2Poly): Boolean; overload;
    function  SolveQuad(R, A, B: TGF2Poly): Boolean; overload;
    procedure Sqr(P: TGF2Poly); overload;
    procedure Sqr(P: TGF2Poly; Times: Longint); overload;
    procedure SqrRoot(P: TGF2Poly);
    function  Trace(P: TGF2Poly): Boolean;
    // Properties
    property CheckParam: Boolean read fCheckParam write fCheckParam;
    property FieldDegree: Longint read fDegree;
    property ONBType: TONBType read fONBType;
  end;


implementation
//==============================================================================
uses Integers_Primes;

////////////////////////////////////////////////////////////////////////////////
// Resource strings
////////////////////////////////////////////////////////////////////////////////

resourcestring
  sONB0001 = 'A field element should have a degree less than the field degree';
  sONB0002 = 'No optimal basis';
  sONB0003 = 'Field degree should be odd';

////////////////////////////////////////////////////////////////////////////////
// Consts, Types
////////////////////////////////////////////////////////////////////////////////

type
  TL = record Lo, Hi: Longword; end;
  TGF2NOHeaderFile = array[0..25] of AnsiChar;

const
  //----- Header of TGF2NO file
  GF2NOHeaderFile : TGF2NOHeaderFile = 'HIT Galois Field GF2N ONB ';

////////////////////////////////////////////////////////////////////////////////
// TGF2NO class
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Public
// P := (P + Q) over GF(2^N)
//==============================================================================
procedure TGF2NO.Add(P, Q: TGF2Poly);
begin
  //----- Check
  if CheckParam then
    if (P.Degree >= fDegree) or (Q.Degree >= fDegree) then
      HITRaiseInvalidArg(ClassName + '.Add',LoadResString(@sONB0001));

  GF2Add(P,Q);
end;

//==============================================================================
// Public
// P := (P + "1") over GF(2^N)
//==============================================================================
procedure TGF2NO.Add1(P: TGF2Poly);
begin
  //----- Check
  if CheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.Add1',LoadResString(@sONB0001));

  GF2Add(P,fPoly1); //P = "1" is (11..111) with Normal Basis
end;

//==============================================================================
// Public
// Create and set Result equal to Self
//==============================================================================
function TGF2NO.Clone: TGF2NO;
begin
  Result := TGF2NO.Create(Self);
end;

//==============================================================================
// Public
// Set Self equal to G
//==============================================================================
procedure TGF2NO.Copy(G: TGF2NO);
begin
  GF2Set(fFieldPoly,G.fFieldPoly);
  GF2Set(fPoly1,G.fPoly1);
  fMatrix.Copy(G.fMatrix);
  fONBType := G.fONBType;
  fCheckParam := G.fCheckParam;
  fDegree := G.fDegree;
end;

//==============================================================================
// Public
// Create Self
//==============================================================================
constructor TGF2NO.Create;
begin
  inherited Create;
  fFieldPoly := TGF2Poly.Create;
  fPoly1 := TGF2Poly.Create;
  fMatrix := TGF2SqrMatrix.Create;
  fONBType := btNone;
  fCheckParam := true;
  fDegree := -1;
  SetField(2);
end;

//==============================================================================
// Public
// Create Self equal to G
//==============================================================================
constructor TGF2NO.Create(G: TGF2NO);
begin
  inherited Create;
  fFieldPoly := G.fFieldPoly.Clone;
  fPoly1 := G.fPoly1.Clone;
  fMatrix := G.fMatrix.Clone;
  fONBType := G.fONBType;
  fCheckParam := G.fCheckParam;
  fDegree := G.fDegree;
end;

//==============================================================================
// Public
// Destroy Self
//==============================================================================
destructor TGF2NO.Destroy;
begin
  fFieldPoly.Free;
  fPoly1.Free;
  fMatrix.Free;
  inherited Destroy;
end;

//==============================================================================
// Public
// P := (P / Q) over GF(2^N)
//==============================================================================
procedure TGF2NO.Divide(P, Q: TGF2Poly);
  var T : TGF2Poly;
begin
  //----- Check
  if Q.Is0 then HITRaiseDivByZero(ClassName + '.Divide');
  if CheckParam then
    if (P.Degree >= fDegree) or (Q.Degree >= fDegree) then
      HITRaiseInvalidArg(ClassName + '.Divide',LoadResString(@sONB0001));

  if P = Q then GF2Set(P,fPoly1) //Obviously, since P <> 0, P div P = 1
  else
  begin
    T := Q.Clone;
    try
      Inv(T);
      Mul(P,T);
    finally
      T.Free;
    end;
  end;
end;

//==============================================================================
// Public
// P := Field polynomial
//==============================================================================
procedure TGF2NO.GetFieldPoly(P: TGF2Poly);
begin
  GF2Set(P,fFieldPoly);
end;

//==============================================================================
// Public
// Result := Optimal Normal Basis type (I or II) if any
//==============================================================================
class function TGF2NO.GetONBType(Deg: Longint): TONBType;
  //--------------------------------------------------------------------------
  function OrderOf2(P: Longword): Longword; register;
  asm
       push ebx
       push edi
       mov  ebx, 1        //Result := 1
       mov  edi, 2        //edi := 2
       mov  ecx, eax      //ecx := P
       mov  edx, edi
  @00: inc  ebx           //Inc(Result)
       mov  eax, edx
       mul  edi
       div  ecx
       cmp  edx, 1        //(2^Result mod P) = 1 ?
       ja   @00
       mov  @Result, ebx
       pop  edi
       pop  ebx
  end;
  //--------------------------------------------------------------------------
  function Check(D, P: Longword): Boolean;
    var a, b : Longword;
  begin
    a := (P-1) div OrderOf2(P);
    while D > 0 do
    begin
      b := a mod D;
      a := D;
      D := b;
    end;
    Result := a = 1;
  end;
  //--------------------------------------------------------------------------
begin
  if (Deg < 2) or ((Deg and 7) = 0) then Result := btNone
  else
  //----- Search first for type 1, then for type 2
  //      Thus, if both types 1 and 2 are ok, return type 1
  if ITruePrime(Deg+1) and Check(Deg,Deg+1) then Result := btType1
  else
  if ITruePrime(Deg+Deg+1) and Check(Deg,Deg+Deg+1) then Result := btType2
  else Result := btNone;
end;

//==============================================================================
// Public
// P := Half-trace of P over GF(2^N)
// Exception if the degree of Self is not odd
//==============================================================================
procedure TGF2NO.HalfTrace(P: TGF2Poly);
  var
    H : TGF2Poly;
    i : Longint;
begin
  //----- Check
  if not Odd(fDegree) then
    HITRaiseInvalidCall(ClassName + '.HalfTrace',LoadResString(@sONB0003));
  if fCheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.HalfTrace',LoadResString(@sONB0001));

  //----- Break ?
  HITBreak;

  H := P.Clone;
  try
    for i := 1 to Pred(fDegree) shr 1 do //"fDegree shr 1" would be ok
    begin
      GF2Rol(P,fDegree,2);
      GF2Add(P,H);
    end;
  finally
    H.Free;
  end;
end;

//==============================================================================
// Public
// P := P^(-1) over GF(2^N)
//==============================================================================
function TGF2NO.Inv(P: TGF2Poly): Boolean;
  var
    A, B : TGF2Poly;
    e    : Longword;
    i, k : Longint;
begin
  //----- Check
  if CheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.Inv',LoadResString(@sONB0001));

  Result := not P.Is0;
  if not Result then Exit;
  A := nil;
  B := nil;
  try
    A := TGF2Poly.Create;
    B := P.Clone;
    e := fDegree - 1;
    k := 1;
    for i := IBitSize32(e)-2 downto 0 do
    begin
      GF2Set(A,B);
      GF2Rol(A,fDegree,k); //Square k times
      Mul(B,A);
      Inc(k,k);
      if IBit32(e,i) then
      begin
        GF2Rol(B,fDegree); //Square
        Mul(B,P);
        Inc(k);
      end;
    end;
    GF2Rol(B,fDegree);
    GF2Swp(P,B);
  finally
    A.Free;
    B.Free;
  end;
end;

//==============================================================================
// Public
// Result := P = "1" over GF(2^N)
//==============================================================================
function TGF2NO.Is1(P: TGF2Poly): Boolean;
begin
  //----- Check
  if CheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.Is1',LoadResString(@sONB0001));

  Result := GF2Equ(P,fPoly1);
end;

//==============================================================================
// Public/Virtual
//==============================================================================
procedure TGF2NO.LoadFromFile(const FileName: string);
  //--------------------------------------------------------------------------
  procedure Error;
  begin
    HITRaiseFileError(ClassName + '.LoadFromFile',FileName);
  end;
  //--------------------------------------------------------------------------
  var
    F : TFileStream;
    H : TGF2NOHeaderFile;
begin
  F := TFileStream.Create(FileName,HITLoadFlags);
  try
    F.ReadBuffer(H,SizeOf(H));
    if H <> GF2NOHeaderFile then Error;
    try
      LoadFromStream(F);
    except
      on EHITStreamError do Error;
    end;
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public/Virtual
//==============================================================================
procedure TGF2NO.LoadFromStream(Stream: TStream);
  var pos : Longint;
begin
  with Stream do
  begin
    pos := Position;
    try
      //----- Field type
      ReadBuffer(fONBType,SizeOf(fONBType));
      //----- Field polynomial
      fFieldPoly.LoadFromStream(Stream);
      //----- Multiplication matrix
      fMatrix.LoadFromStream(Stream);
      //----- Check parameter status
      ReadBuffer(fCheckParam,SizeOf(fCheckParam));
      //----- Update
      fDegree := fFieldPoly.Degree;
      fPoly1.Fill(fDegree);
    except
      Position := pos;
      SetField(2); //Reset
      raise;
    end;
  end;
end;

//==============================================================================
// Public
// P := (P * Q) over GF(2^N)
//==============================================================================
procedure TGF2NO.Mul(P, Q: TGF2Poly);
  var
    R       : TGF2Poly;
    i, dmax : Longint;
    BreakOn : Boolean;
begin
  //----- This is not only faster, this is indispensable!
  if P = Q then
  begin
    Sqr(P);
    Exit;
  end;

  //----- Check
  if CheckParam then
    if (P.Degree >= fDegree) or (Q.Degree >= fDegree) then
      HITRaiseInvalidArg(ClassName + '.Mul',LoadResString(@sONB0001));

  //----- Break ?
  BreakOn := HITBreakOn;
  if BreakOn then HITBreak;

  R := TGF2Poly.Create;
  try
    dmax := fDegree - 1;
    for i := 0 to dmax do
    begin

      // The method using the so-called "Lambda" array is slower than
      // the GF2 "product parity" operation using the multiplication
      // matrix (but it is a little more compact)

      //----- R[i] := Parity(P * M * Q)
      R[i] := GF2ProductParity(P,fMatrix,Q);
      //----- Rotations by 1 bit to the right (on dmax+1 bits)
      if GF2DivX(P) then P[dmax] := true;
      if GF2DivX(Q) then Q[dmax] := true;

      //----- Break ?
      if BreakOn then HITBreak;
    end;
    GF2Swp(R,P);
  finally
    R.Free;
  end;
end;

//==============================================================================
// Public
// Result := Number of elements of Self
//==============================================================================
procedure TGF2NO.Order(N: TInteger);
begin
  N.Set0;
  N.Bit[fDegree] := true;
end;

//==============================================================================
// Public/Overloaded
// P := P^E over GF(2^N)
//==============================================================================
function TGF2NO.Pow(P: TGF2Poly; E: Int64): Boolean;
  var
    Z : TGF2Poly;
    i : Longint;
begin
  //----- Check
  if CheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.Pow',LoadResString(@sONB0001));

  //----- Break ?
  HITBreak;

  Result := true;
  //----- P = 1 ?
  if GF2Equ(P,fPoly1) then Exit;
  //----- E = 0 ?
  if (TL(E).Hi or TL(E).Lo) = 0 then
  begin
    GF2Set(P,fPoly1);
    Exit;
  end;
  //----- P = 0 ?
  if P.Is0 then
  begin
    //----- E < 0 ?
    if (TL(E).Hi and $80000000) <> 0 then Result := false;
    Exit;
  end;
  //
  // Here, P > 1 and |E| <> 0
  //
  //----- E < 0 ?
  if (TL(E).Hi and $80000000) <> 0 then
  begin
    Result := Inv(P);
    if not Result then Exit;
    E := -E;
  end;
  //
  // Here, E cannot be negative except if equal to MinInt64
  //
  //----- (E > 1) or (E = MinInt64) ?
  if (TL(E).Hi <> 0) or (TL(E).Lo > 1) then
  begin
    Z := P.Clone;
    try
      for i := IBitSize64(E)-2 downto 0 do
      begin
        Sqr(P);
        if IBit64(E,i) then Mul(P,Z);
      end;
    finally
      Z.Free;
    end;
  end;
end;

//==============================================================================
// Public/Overloaded
// P := P^E over GF(2^N)
//==============================================================================
function TGF2NO.Pow(P: TGF2Poly; E: TInteger): Boolean;
  var
    Z : TGF2Poly;
    i : Longint;
begin
  //----- Check
  if fCheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.Pow',LoadResString(@sONB0001));

  //----- Break ?
  HITBreak;

  Result := true;
  //----- E = 0 ?
  if E.Is0 then
  begin
    GF2Set(P,fPoly1);
    Exit;
  end;
  //----- P = 0 ?
  if P.Is0 then
  begin
    //----- E < 0 ?
    if E.Neg then Result := false;
    Exit;
  end;
  if GF2Equ(P,fPoly1) then Exit; //if P = 1
  //
  // Here, P > 1 and |E| <> 0
  //
  //----- E < 0 ?
  if E.Neg then
  begin
    Result := Inv(P);
    if not Result then Exit; //Impossible...
  end;
  //----- |E| > 1 ?
  if ICmpAbs(E,1) > 0 then
  begin
    Z := P.Clone;
    try
      for i := E.BitSize-2 downto 0 do
      begin
        Sqr(P);
        if E.Bit[i] then Mul(P,Z);
      end;
    finally
      Z.Free;
    end;
  end;
end;

//==============================================================================
// Public
// P := Random element of which the degree is at most fDegree-1
//==============================================================================
procedure TGF2NO.Rnd(P: TGF2Poly);
begin
  GF2Rnd(P,fDegree-1,false);
end;

//==============================================================================
// Public/Virtual
//==============================================================================
procedure TGF2NO.SaveToFile(const FileName: string);
  var F : TFileStream;
begin
  F := TFileStream.Create(FileName,HITSaveFlags);
  try
    F.WriteBuffer(GF2NOHeaderFile,SizeOf(GF2NOHeaderFile));
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

//==============================================================================
// Public/Virtual
//==============================================================================
procedure TGF2NO.SaveToStream(Stream: TStream);
  var pos : Longint;
begin
  with Stream do
  begin
    pos := Position;
    try
      //----- Field type
      WriteBuffer(fONBType,SizeOf(fONBType));
      //----- Field polynomial
      fFieldPoly.SaveToStream(Stream);
      //----- Multiplication matrix
      fMatrix.SaveToStream(Stream);
      //----- Check parameter status
      WriteBuffer(fCheckParam,SizeOf(fCheckParam));
    except
      Position := pos;
      raise;
    end;
  end;
end;

//==============================================================================
// Public
// P := "1" over GF(2^N)
//==============================================================================
procedure TGF2NO.Set1(P: TGF2Poly);
begin
  //----- Check
  if CheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.Set1',LoadResString(@sONB0001));

  GF2Set(P,fPoly1);
end;

//==============================================================================
// Public
// Set Self as a Galois field of degree Deg (if any with ONB of type I or II)
//==============================================================================
procedure TGF2NO.SetField(Deg: Longint);
  var
    X, Y    : TGF2Poly;
    A       : TGF2SqrMatrix;
    i, j, k : Longint;
begin
  //----- Set field polynomial
  case GetONBType(Deg) of
    btType1:
      begin
        fFieldPoly.Fill(Deg+1);
        fONBType := btType1;
      end;
    btType2:
      begin
        X := nil;
        Y := nil;
        try
          X := TGF2Poly.Create;
          Y := TGF2Poly.Create(1);
          GF2Set(fFieldPoly,3); //P := t + 1
          for i := 1 to Deg-1 do
          begin
            GF2Set(X,Y);
            GF2Set(Y,fFieldPoly);
            GF2MulX(fFieldPoly);
            GF2Add(fFieldPoly,X);
          end;
        finally
          X.Free;
          Y.Free;
        end;
        fONBType := btType2;
      end;
    else HITRaiseInvalidCall(ClassName + '.SetField',LoadResString(@sONB0002));
  end;
  fDegree := Deg;
  //----- Set the polynomial equal to "1" in ONB representations
  fPoly1.Fill(Deg);

  //----- Compute the multiplication matrix
  A := nil;
  X := TGF2Poly.Create;
  try
    fMatrix.Dimension := Deg;
    A := TGF2SqrMatrix.Create;
    A.Dimension := Deg;
    //----- M[i] := x^(2^i) mod P
    //      A[i] := (M[i] * x) mod P
    X.SetX;
    fMatrix.SetRow(0,X);
    GF2MulX(X);
    if Deg = 2 then GF2Add(X,fFieldPoly);
    A.SetRow(0,X);
    for i := 1 to Deg-1 do
    begin
      fMatrix.GetRow(i-1,X);
      GF2Sqr(X);
      GF2Mod(X,fFieldPoly);
      fMatrix.SetRow(i,X);
      GF2MulX(X);
      if X[Deg] then GF2Add(X,fFieldPoly);
      A.SetRow(i,X);
    end;
    //----- A := A * M^(-1)
    GF2Inv(fMatrix);
    GF2MulByRight(A,fMatrix);
    //----- M[i,j] := A[(j-i) mod Deg, (-i) mod Deg]
    for j := 0 to Deg-1 do fMatrix[0,j] := A[j,0]; //Case i = 0
    for i := 1 to Deg-1 do
    begin
      k := Deg - i;
      for j := 0 to i-1   do fMatrix[i,j] := A[j + k,k];
      for j := i to Deg-1 do fMatrix[i,j] := A[j - i,k];
    end;
  finally
    X.Free;
    A.Free;
  end;
end;

//==============================================================================
// Public/Overloaded
// R := root of R^2 + R + B = 0 (if exists) over GF(2^N)
// Result := TRUE if solutions exist, FALSE otherwise
//==============================================================================
function TGF2NO.SolveQuad(R, B: TGF2Poly): Boolean;
  var i : Longint;
begin
  //----- Check
  if R = B then HITRaiseDuplicatePtr(ClassName + '.SolveQuad','R,B');
  if CheckParam then
    if (R.Degree >= fDegree) or (B.Degree >= fDegree) then
      HITRaiseInvalidArg(ClassName + '.SolveQuad',LoadResString(@sONB0001));

  //----- Break ?
  HITBreak;

  R.Set0;
  //----- Solutions exist if and only if Trace(B) = 0
  if Trace(B) then Result := false
  else
  begin
    Result := true;
    if not B.Is0 then //Otherwise R = 0 is solution
      for i := 1 to fDegree-1 do R[i] := R[i-1] xor B[i];
  end;
end;

//==============================================================================
// Public/Overloaded
// R := root of R^2 + AR + B = 0 (if exists) over GF(2^N)
// Result := TRUE if solutions exist, FALSE otherwise
//==============================================================================
function TGF2NO.SolveQuad(R, A, B: TGF2Poly): Boolean;
  var U : TGF2Poly;
begin
  //----- Check
  if (R = A) or (R = B) then
    HITRaiseDuplicatePtr(ClassName + '.SolveQuad','R,A,B');
  if CheckParam then
    if (R.Degree >= fDegree) or
       (A.Degree >= fDegree) or
       (B.Degree >= fDegree) then
      HITRaiseInvalidArg(ClassName + '.SolveQuad',LoadResString(@sONB0001));

  //----- Break ?
  HITBreak;

  //----- Handle trivial cases
  if A.Is0 then
  begin
    GF2Set(R,B);
    SqrRoot(R);
    Result := true;
    Exit;
  end;
  if B.Is0 then
  begin
    R.Set0;
    Result := true;
    Exit;
  end;

  //----- Solve R^2 + R + U = 0 for U = B/A^2
  //      R := R*A is solution for original equation
  U := A.Clone;
  try
    //----- Reduce
    Inv(U);
    Sqr(U);
    Mul(U,B);
    //----- Solve
    Result := SolveQuad(R,U);
    //----- Set result
    if Result then Mul(R,A);
  finally
    U.Free;
  end;
end;

//==============================================================================
// Public/Overloaded
// P := P over GF(2^N)
//==============================================================================
procedure TGF2NO.Sqr(P: TGF2Poly);
begin
  //----- Check
  if CheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.Sqr',LoadResString(@sONB0001));

  GF2MulX(P);
  if P[fDegree] then
  begin
    P[0] := true;
    P[fDegree] := false;
  end;
end;

//==============================================================================
// Public/Overloaded
// P := P^(2^Times) over GF(2^N)
//==============================================================================
procedure TGF2NO.Sqr(P: TGF2Poly; Times: Longint);
begin
  //----- Check
  if CheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.Sqr',LoadResString(@sONB0001));

  //----- Reduce exponent (if need be)
  if Times >= fDegree then Times := Times mod fDegree
  else
  if Times < 0 then
  begin
    Times := Times mod fDegree;
    Inc(Times,fDegree);
  end;

  //----- Square Times times
  GF2Rol(P,fDegree,Times);
end;

//==============================================================================
// Public
// P := P^(1/2) over GF(2^N)
//==============================================================================
procedure TGF2NO.SqrRoot(P: TGF2Poly);
begin
  //----- Check
  if CheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.SqrRoot',LoadResString(@sONB0001));

  if GF2DivX(P) then P[fDegree-1] := true;
end;

//==============================================================================
// Public
// Result := Trace of P over GF(2^N)
//==============================================================================
function TGF2NO.Trace(P: TGF2Poly): Boolean;
begin
  //----- Check
  if CheckParam then
    if P.Degree >= fDegree then
      HITRaiseInvalidArg(ClassName + '.Trace',LoadResString(@sONB0001));

  Result := P.Parity;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////
