////////////////////////////////////////////////////////////////////////////////
// Computation of Elliptic Curves over GF(P) of "Nearly" Prime Order
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit GFP_ECBuilder;

{.$DEFINE DEBUG}

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers;

//==============================================================================
// The following procedure generates
// 1) a PSize-bit prime P so that it exists a CM Discriminant in {1,2,3,7,11,
//    19,43,67,163}
// 2) a curve of order K*R with K smooth and R prime (K is at most KMaxSize-bit
//    sized)
// 3) a point (X,Y) of order R over the curve
//==============================================================================

procedure BuildCurve(P,K,R,A,B,X,Y: TInteger; PSize,KMaxSize: Longint);


implementation
//==============================================================================
uses Integers_Primes, GFP_EllipticCurves;

//------------------------------------------------------------------------------
// Check whether R is "nearly" prime
// -> R, any positive integer
// <- R, prime of at least RMinSize bits (if Result = true)
// <- K, integer of at most KMaxSize bits (constituted of factors less than
//       L.Last
// Result := TRUE if and only if R is nearly prime
//------------------------------------------------------------------------------
function NearlyPrime(R, K: TInteger; RMinSize, KMaxSize: Longint;
         L: TPrimeList): Boolean;
  var
    M     : TInteger;
    i, RS : Longint;
    d, R0 : Longword;
begin
  //----- Divisibility by 2
  i := IShruo(R);
  if ((i+1) > KMaxSize) or (R.BitSize < RMinSize) then
  begin
    Result := false;
    Exit;
  end;
  K.Set1;
  if i > 0 then IShl(K,i);

  //----- Divisibility by other primes
  M := TInteger.Create;
  try
    RS := R.LongSize;
    R0 := R.Long[0];
    for i := 2 to L.Count do
    begin
      d := L[i];
      if (RS = 1) and (R0 <= d) then Break;

      ISet(M,R);
      if IDiv(M,d) = 0 then
      begin
        repeat
          IMul(K,d);
          ISet(R,M);
        until IDiv(M,d) <> 0;
        if (K.BitSize > KMaxSize) or (R.BitSize < RMinSize) then
        begin
          Result := false;
          Exit;
        end;
        RS := R.LongSize;
        R0 := R.Long[0];
      end;
    end;
  finally
    M.Free;
  end;

  //----- Check size and primality of the remaining part
  Result := (K.BitSize <= KMaxSize) and
            (R.BitSize >= RMinSize) and
            IProbablyPrime(R,0);
end;

//------------------------------------------------------------------------------
// Check whether D is a CM discriminant for P
// Solve 4P = W + DV
// ! The returned V is any (not taken in account) if D <> 1 or D <> 3
// See IEEE P1363 A.14.2.2
//------------------------------------------------------------------------------
function CMDiscriminant(D: Longint; P, W, V: TInteger): Boolean;
  var
    S, T, A, B, C : TInteger;
    SNeg          : Boolean;
begin
  S := nil;
  T := nil;
  A := nil;
  B := nil;
  C := nil;
  try
    S := TInteger.Create;
    T := TInteger.Create;
    A := P.Clone;
    B := TInteger.Create(D);
    B.Negate;
    ISqrRootMod(B,P);
    C := B.Clone;
    ISqr(C);
    IAdd(C,D);
    IDiv(C,P);
    W.Set1;
    V.Set0;
    repeat
      ISet(S,B);
      IShl(S);
      //----- Check whether |2B| <= A <= C
      if (ICmpAbs(S,A) <= 0) and (ICmpAbs(A,C) <= 0) then Break;
      //----- S := |_ (2B + C)/(2C) _|
      IAdd(S,C);
      ISet(T,C);
      IShl(T);
      SNeg := S.Neg;
      if not IDiv(S,T) then if SNeg then IDec(S);
      //----- A := C', B := SC' - B', C := (SC' - 2B')S + A'
      ISwp(A,C);
      ISet(T,A);
      IMul(T,S);
      ISub(T,B);
      ISwp(T,B);
      T.Negate;
      IAdd(T,B);
      IMul(T,S);
      IAdd(C,T);
      //----- W := SW' + V', V := -W'
      ISet(T,W);
      IMul(T,S);
      IAdd(V,T);
      ISwp(V,W);
      V.Negate;
    until false;

    if (D = 11) and (ICmp(A,3) = 0) then
    begin
      W.Negate;
      ISwp(W,V);
      B.Negate;
      ISwp(A,C);
    end;

    if (D = 1) or (D = 3) then
    begin
      //----- W := 2W, V := 2V
      IShl(W);
      IShl(V);
      Result := true;
    end
    else
    if A.Is1 then
    begin
      //----- W := 2W
      IShl(W);
      Result := true;
    end
    else
    if ICmp(A,4) = 0 then
    begin
      //----- W := 4W + BV
      IShl(W,2);
      IMul(V,B);
      IAdd(W,V);
      Result := true;
    end
    else Result := false;
  finally
    S.Free;
    T.Free;
    A.Free;
    B.Free;
    C.Free;
  end;
end;

//------------------------------------------------------------------------------
// Search a CM Discriminant for P and an order K*R with K smooth and R prime
// Result := TRUE if succeded, FALSE if failed
// See IEEE P1363 A.14.2.1
//------------------------------------------------------------------------------
function GetDKR(var D: Longint; K, R, P: TInteger; RMinSize, KMaxSize: Longint;
         L: TPrimeList): Boolean;
  var
    V, W       : TInteger;
    H          : Longword;
    PRes, DRes : Longint;
begin
  Result := false;
  V := nil;
  W := nil;
  try
    V := TInteger.Create;
    W := P.Clone;
    ISqrRoot(W);
    IInc(W);
    ISqr(W);
    IShr(W,RMinSize-1); //"RMin" is equal to 2^(RMinSize-1)
    if W.LongSize < 2 then H := W.Long[0] else H := 4;
    W.Set0;

    PRes := P.Long[0] and 7;
    if H = 1 then D := 3
    else
    case PRes of
      1,5: D := 1;
      3:   D := 2;
      7:   D := 3;
    end;

    repeat
      if (IJacobi(-D,P) = 1) and CMDiscriminant(D,P,W,V) then
      begin
        //----- R := P + 1 + W
        ISet(R,P);
        IInc(R);
        IAdd(R,W);
        Result := NearlyPrime(R,K,RMinSize,KMaxSize,L);
        if Result then Exit;
        //----- R := P + 1 - W
        ISet(R,P);
        IInc(R);
        ISub(R,W);
        Result := NearlyPrime(R,K,RMinSize,KMaxSize,L);
        if Result then Exit;
        if D = 1 then
        begin
          //----- R := P + 1 + V
          ISet(R,P);
          IInc(R);
          IAdd(R,V);
          Result := NearlyPrime(R,K,RMinSize,KMaxSize,L);
          if Result then Exit;
          //----- R := P + 1 - V
          ISet(R,P);
          IInc(R);
          ISub(R,V);
          Result := NearlyPrime(R,K,RMinSize,KMaxSize,L);
          if Result then Exit;
        end
        else
        if (D = 3) then
        begin
          //----- R := P + 1 + (W + 3V)/2
          ISet(R,V);
          IMul(R,3);
          IAdd(R,W);
          IShr(R);
          IAdd(R,P);
          IInc(R);
          Result := NearlyPrime(R,K,RMinSize,KMaxSize,L);
          if Result then Exit;
          //----- R := P + 1 + (W - 3V)/2
          ISet(R,V);
          IMul(R,3);
          R.Negate;
          IAdd(R,W);
          IShr(R);
          IAdd(R,P);
          IInc(R);
          Result := NearlyPrime(R,K,RMinSize,KMaxSize,L);
          if Result then Exit;
          //----- R := P + 1 - (W - 3V)/2
          ISet(R,V);
          IMul(R,3);
          R.Negate;
          IAdd(R,W);
          IShr(R);
          R.Negate;
          IAdd(R,P);
          IInc(R);
          Result := NearlyPrime(R,K,RMinSize,KMaxSize,L);
          if Result then Exit;
          //----- R := P + 1 - (W + 3V)/2
          ISet(R,V);
          IMul(R,3);
          IAdd(R,W);
          IShr(R);
          R.Negate;
          IAdd(R,P);
          IInc(R);
          Result := NearlyPrime(R,K,RMinSize,KMaxSize,L);
          if Result then Exit;
        end;
      end;

      repeat
        if H = 1 then Inc(D,8)
        else
        if H < 4 then
        case PRes of
          1: repeat
               Inc(D);
             until (D and 7) <> 7;
          3: repeat
               Inc(D);
               DRes := D and 7;
             until (DRes = 2) or (DRes = 3);
          5: repeat
               Inc(D);
               DRes := D and 7;
             until Odd(D) and (DRes <> 7);
          7: repeat
               Inc(D);
               DRes := D and 7;
             until (DRes = 3) or (DRes = 6)
        end
        else
        case PRes of
          1: Inc(D);
          3: repeat
               Inc(D);
               DRes := D and 7;
             until (DRes = 2) or (DRes = 3) or (DRes = 7);
          5: repeat
               Inc(D);
             until Odd(D);
          7: repeat
               Inc(D);
               DRes := D and 7;
             until (DRes = 3) or (DRes = 6) or (DRes = 7);
        end;
        if D > 163 then Exit;
      until D in [1,2,3,7,11,19,43,67,163];
    until false;

  finally
    V.Free;
    W.Free;
  end;
end;

//------------------------------------------------------------------------------
// Search for a curve modulo P having order KR and for a point (X,Y) having
// order R
// See IEEE P1363 A.14.4.1 & A.14.4.2
//------------------------------------------------------------------------------
function GetCurveAndPoint(D: Longint; P, K, R, A, B, X, Y: TInteger): Boolean;
  var
    E, AA, BB : TInteger;
    ec        : TGFPECurve;
    ep        : TGFPEPoint;
    i, s      : Longint;
begin
  Result := false;
  E := nil;
  AA := nil;
  BB := nil;
  ec := nil;
  ep := nil;
  try
    case D of
      1: begin
           A.Set1;
           B.Set0;
         end;
      2: begin
           ISet(A,-30);
           ISet(B,56);
         end;
      3: begin
           A.Set0;
           B.Set1;
         end;
      7: begin
           ISet(A,-35);
           ISet(B,98);
         end;
     11: begin
           ISet(A,-264);
           ISet(B,1694);
         end;
     19: begin
           ISet(A,-152);
           ISet(B,722);
         end;
     43: begin
           ISet(A,-3440);
           ISet(B,77658);
         end;
     67: begin
           ISet(A,-29480);
           ISet(B,1948226);
         end;
    163: begin
           ISet(A,-8697680);
           ISet(B,9873093538);
         end;
    end;
    IMod(A,P);
    IMod(B,P);

    AA := TInteger.Create;
    BB := TInteger.Create;
    E := TInteger.Create;
    ec := TGFPECurve.Create;
    ep := TGFPEPoint.Create;
    ec.SetModulus(P);
    s := P.BitSize shl 1;
    i := 0;
    repeat
      Inc(i);
      if i > 50 then Exit; //It never occurs!
      //----- 0 < E < P
      repeat
        IRnd(E,s);
        IMod(E,P);
      until not E.Is0;
      ISet(AA,A);
      ISet(BB,B);
      if D = 1 then
      begin
        IMul(AA,E);
        IMod(AA,P);
      end
      else
      if D = 3 then
      begin
        IMul(BB,E);
        IMod(BB,P);
      end
      else
      begin
        IMul(BB,E);
        IMod(BB,P);
        ISqr(E);
        IMod(E,P);
        IMul(AA,E);
        IMod(AA,P);
        IMul(BB,E);
        IMod(BB,P);
      end;
      //----- Search for a point of order R
      ec.SetCoefs(AA,BB);
      repeat
        ec.GetRndPoint(ep);
        ec.Mul(ep,K,true);
      until not ep.IsO;
      ISet(X,ep.X);
      ISet(Y,ep.Y);
      //----- Check whether the order is R
      ec.Mul(ep,R,true);
    until ep.IsO;
    ISet(A,AA);
    ISet(B,BB);
    Result := true;
  finally
    AA.Free;
    BB.Free;
    E.Free;
    ec.Free;
    ep.Free;
  end;
end;

//==============================================================================
// Generate a prime P, an elliptic curve over GF(P) and a point of order R
// The generated curve has order K*R with K smooth and R prime
// The generated point (X,Y) has order R
// -> PSize, binary size of P
// <- P, prime
// <- K, small integer
// <- R, prime
// <- A,B, coefficients of an elliptic curve over GF(P) having order K*R
// <- X,Y, point on E(A,B)/P having order R
//==============================================================================
procedure BuildCurve(P,K,R,A,B,X,Y: TInteger; PSize,KMaxSize: Longint);
  var
    L           : TPrimeList;
    D, RMinSize : Longint;
begin
  //----- Check
  if PSize < 16 then HITRaiseInvalidArg('BuildCurve','PSize < 16');
  if KMaxSize < 1 then HITRaiseInvalidArg('BuildCurve','KMaxSize < 1');

  L := TPrimeList.Create;
  try
    if KMaxSize >= (PSize shr 1) then KMaxSize := PSize shr 1 - 1;
    RMinSize := PSize - KMaxSize;

    //----- Resize the prime list according to KMaxSize
    if KMaxSize < 18 then L.Count := Pi2Pow[KMaxSize] else L.Count := Pi2Pow18;

    repeat
      repeat
        IRnd(P,PSize);
        IMakePrime(P,[2,3]);
      until GetDKR(D,K,R,P,RMinSize,KMaxSize,L);
    until GetCurveAndPoint(D,P,K,R,A,B,X,Y);
  finally
    L.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

