////////////////////////////////////////////////////////////////////////////////
// Computation of Gordon primes (also called "strong" primes)
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit GordonPrimes;

{.$DEFINE DEBUG}

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers;

//==============================================================================
// A Gordon prime (also called a "strong" prime) is a prime P such that
//  P-1 is divisible by a large prime R
//  P+1 is divisible by a large prime S
//  R-1 is divisible by a large prime T
//==============================================================================

function IMakeGordonPrime(P: TInteger; BitCount: Longint): Boolean; overload;
function IMakeGordonPrime(P, R, S, T: TInteger; BitCount: Longint): Boolean;
         overload;

implementation
uses Integers_Primes, HIT_Rand;
//==============================================================================

//==============================================================================
// Global/Overloaded
// Compute a Gordon Prime of BitCount bits
// -> BitCount, whished binary size of P, should be greater than 63
// <- P, Gordon prime of BitCount bits (or, sometimes, a little less)
// <- R, large prime which divides P-1
// <- S, large prime which divides P+1
// <- T, large prime which divides R-1
// Result := TRUE iff resulting P has its bitsize equal to BitCount
// P.BitSize is almost always equal to BitCount
//==============================================================================
function IMakeGordonPrime(P, R, S, T: TInteger; BitCount: Longint): Boolean;
  var
    rg            : TIntCSRG; //Secure random generator (See HIT_Rand)
    A, B          : TInteger;
    ss, si, sj, L : Longint;
    i, j, Mask    : Longword;
begin
  //------- Check
  if BitCount < 64 then HITRaiseInvalidArg('IMakeGordonPrime','BitCount < 64');
  if (P = R) or (P = S) or (P = T) or
     (R = S) or (R = T) or (S = T)
    then HITRaiseDuplicatePtr('IMakeGordonPrime','P,R,S,T');     

  rg := nil;
  A := nil;
  B := nil;
  try
    //----- Secure random generator
    rg := TIntCSRG.Create;

    //----- Random i and j
    case BitCount of
      64..127:  Mask := $fff;
      128..200: Mask := $ffff;
      else      Mask := $ffffff;
    end;
    i := rg.RndLong and Mask;
    j := rg.RndLong and Mask;
    si := IBitSize32(i);
    sj := IBitSize32(j);

    //----- Compute T
    L := (BitCount - si - sj - 1) shr 1;
    if L < 16 then L := 16;
    rg.Rnd(T,L);
    if T.BitSize < 16 then T.Bit[L-1] := true;
    IMakePrime(T,[2,3,5,7,11],0,0);

    //----- Compute R = 2iT + 1
    A := T.Clone;
    IShl(A);
    ISet(R,A);
    IMul(R,i);
    IInc(R);
    while not IProbablyPrime(R,0) do IAdd(R,A); //Inc i

    //----- Compute S
    L := BitCount - R.BitSize - sj + 3; //+3 to oversize S
    rg.Rnd(S,L,true);
    B := S.Clone;
    repeat
      S.Bit[L-1] := true;
      IMul(S,j);
      IMul(S,R);
      ss := S.BitSize;
      if ss >= BitCount then Dec(L);
      ISet(S,B);
      ICut(S,L);
    until ss < BitCount;
    S.Bit[L-1] := true;
    IMakePrime(S,[2,3,5,7,11],0,0);

    //----- Compute P = 2(S^(R-2) mod R)S - 1
    ISet(P,S);
    ISet(B,R);
    ISub(B,2);
    IPowMod(P,B,R);
    IShl(S);
    IMul(P,S);
    IDec(P);
    //----- B := 2RS
    ISet(B,R);
    IMul(B,S);
    IShr(S);

    //----- Compute P = (2(S^(R-2) mod R)S - 1) + 2jRS
    ISet(A,B);
    IMul(A,j);
    IAdd(P,A);
    while not IProbablyPrime(P,0) do IAdd(P,B); //Inc j
    Result := P.BitSize = BitCount;
  finally
    rg.Free;
    A.Free;
    B.Free;
  end;
end;

//==============================================================================
// Global/Overloaded
// Compute a Gordon prime of BitCount bits
// -> BitCount, whished binary size of P
// <- P, Gordon prime of BitCount bits (or, sometimes, a little less)
// Result := TRUE iff resulting P has its bitsize equal to BitCount
//==============================================================================
function IMakeGordonPrime(P: TInteger; BitCount: Longint): Boolean;
  var R, S, T : TInteger;
begin
  R := nil;
  S := nil;
  T := nil;
  try
    R := TInteger.Create;
    S := TInteger.Create;
    T := TInteger.Create;
    Result := IMakeGordonPrime(P,R,S,T,BitCount);
  finally
    R.Free;
    S.Free;
    T.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

