////////////////////////////////////////////////////////////////////////////////
// Lucas Primality Test
// Unit of the HIT package for DELPHI 4/5
//------------------------------------------------------------------------------
// Copyright (c) 2000, Marcel Martin. All rights reserved.
////////////////////////////////////////////////////////////////////////////////
unit LucasTest;

{.$DEFINE DEBUG} //Debugging code

//==============================================================================
{$IFDEF DEBUG}
{$C+,D+,L+,O-,YD,W+}
{$ELSE}
{$C-,D-,L-,O+,Y-,W-}
{$ENDIF}
//------------------------------------------------------------------------------
{$A+,B-,G+,H+,I-,J-,Q-,R-}
//==============================================================================

interface
uses Integers;

//------------------------------------------------------------------------------
// Lucas primality test
// N must be odd
// F must contain all the prime odd factors (without exponent) of N+1
//------------------------------------------------------------------------------

function LTest(N: TInteger; const F: array of Longword): Integer; overload;
function LTest(N: TInteger; const F: array of TInteger): Integer; overload;

implementation
//==============================================================================
uses Integers_Primes;

////////////////////////////////////////////////////////////////////////////////
// Lucas tests
////////////////////////////////////////////////////////////////////////////////

//==============================================================================
// Generate random P and Q such that
// Jacobi(Q,N) = -1, Jacobi(P-4Q,N) = -1 and GCD(2Q(P-4Q),N) = 1
// N must be odd
//==============================================================================
procedure GetRandomPQ(P, Q, N: TInteger);
  var
    D, E : TInteger;
    bsz  : Longint;
begin
  D := nil;
  E := nil;
  try
    D := TInteger.Create;
    E := TInteger.Create;
    P.Set0;
    Q.Set0;
    //----- Q, random such that Jacobi(Q,N) = -1 and GCD(Q,N) = 1
    bsz := N.BitSize;
    IRnd(Q,Succ(bsz));
    repeat
      IInc(Q);
      IMod(Q,N);
    until IJacobi(Q,N) = -1; //if TRUE then IGCD1(Q,N) = TRUE
    //----- P, random such that Jacobi(P-4Q,N) = -1 and GCD((P-4Q),N) = 1
    IRnd(P,Succ(bsz));
    IMod(P,N);
    ISet(E,Q);
    IShl(E,2);
    IMod(E,N); //E = 4Q mod N
    ISet(D,P);
    ISqr(D);
    IMod(D,N);
    ISub(D,E);
    if D.Neg then IAdd(D,N); //D := (P-4Q) mod N
    if IJacobi(D,N) = -1 then Exit; //if TRUE then IGCD1(D,N) = TRUE
    ISet(E,P);
    IShl(E);
    IDec(E);
    IMod(E,N); //E := 2P+1 mod N
    repeat
      IInc(P);
      IMod(P,N); //P := (P+1) mod N
      IAdd(E,2); //E = 2P
      IAdd(D,E);
      IMod(D,N);
    until IJacobi(D,N) = -1; //if TRUE then IGCD1(D,N) = TRUE
  finally
    D.Free;
    E.Free;
  end;
end;

//==============================================================================
// Lucas primality test
// -> N must be odd
// -> F must contain all the prime odd factors (without exponent) of N+1
// <- -1  Exception, error, etc.
// <-  0  N is composite
// <-  1  N is prime
//==============================================================================
function LTest(N: TInteger; const F: array of TInteger): Integer;
  label NEWTRY;
  var
    K, P, Q, V : TInteger;
    i          : Longint;
begin
  Result := -1;
  if High(F) < 0 then Exit;
  //----- Test the primality of N
  K := nil;
  P := nil;
  Q := nil;
  V := nil;
  try
    K := TInteger.Create;
    P := TInteger.Create;
    Q := TInteger.Create;
    V := TInteger.Create;
  NEWTRY:
    GetRandomPQ(P,Q,N);
    //----- Calculate V[K] with K = (N+1)/2
    ISet(K,N);
    IInc(K);
    IShr(K);
    //----- Calculate the V[k] element of a Lucas sequence (modulo N)
    ILucasMod(V,K,P,Q,N);
    //----- If V[(N+1)/2] = 0, N might be prime
    if V.Is0 then
    begin
      for i := 0 to High(F) do
      begin
        //----- Calculate V[K] with K = (N+1)/(2*F[i])
        ISet(K,N);
        IInc(K);
        IShr(K);
        IDiv(K,F[i]);
        ILucasMod(V,K,P,Q,N);
        //----- If V[(N+1)/(2*F[i])] = 0, we cannot conclude.
        //      We have to retry with other values P and Q
        if V.Is0 then goto NEWTRY;
      end;
      //----- N is prime
      Result := 1;
    end
    //----- If V[(N+1)/2] <> 0 then N is composite
    else Result := 0;
  finally
    K.Free;
    P.Free;
    Q.Free;
    V.Free;
  end;
end;

//==============================================================================
//==============================================================================
function LTest(N: TInteger; const F: array of Longword): Integer;
  label NEXT;
  var
    K, P, Q, V : TInteger;
    i          : Longint;
begin
  Result := -1;
  if High(F) < 0 then Exit;
  K := nil;
  P := nil;
  Q := nil;
  V := nil;
  try
    K := TInteger.Create;
    P := TInteger.Create;
    Q := TInteger.Create;
    V := TInteger.Create;
  NEXT:
    GetRandomPQ(P,Q,N);
    ISet(K,N);
    IInc(K);
    IShr(K);
    //----- Calculate the V[k] element of a Lucas sequence (modulo N)
    ILucasMod(V,K,P,Q,N);
    if V.Is0 then
    begin
      for i := 0 to High(F) do
      begin
        ISet(K,N);
        IInc(K);
        IShr(K);
        IDiv(K,F[i]);
        ILucasMod(V,K,P,Q,N);
        if V.Is0 then goto NEXT;
      end;
      Result := 1;
    end
    else Result := 0;
  finally
    K.Free;
    P.Free;
    Q.Free;
    V.Free;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
end.
////////////////////////////////////////////////////////////////////////////////

